"""Entrypoint of the ``sequoia`` CLI command."""

import argparse
import importlib
import inspect
import logging
import sys

from pyramid.config import Configurator
from pyramid.exceptions import ConfigurationError
from pyramid.paster import setup_logging
from pyramid.settings import aslist

from .config import export_api, get_appsettings, includeme

log = logging.getLogger(__name__)


def shell(settings, **kwargs):
    try:
        includes = aslist(settings.pop("pyramid.includes", ""), flatten=False)
        config = Configurator(settings=settings)
        config.include(includeme)
        for conf in includes:
            if conf == "sequoia.base3":
                log.warning(
                    "Ignoring sequoia.base3 from `pyramid.includes` "
                    "you should remove it"
                )
                continue
            try:
                config.include(conf)
            except ConfigurationError:
                log.exception(
                    "An error occurred when trying " "to configure: {}".format(conf)
                )
                sys.exit(1)

        config.include(export_api, route_prefix=settings.get("sequoia.route_prefix"))
        # push it back for the monitoring view
        config.registry.settings["pyramid.includes"] = includes
        config.end()
    except Exception:
        log.exception("Unexpected exception while configuring app")

    try:
        from IPython import embed
        from traitlets.config import Config

        cfg = Config()
        cfg.InteractiveShellEmbed.confirm_exit = False
        embed(config=cfg, header="Sequoia Shell")
    except ImportError:
        # try ~IPython-0.10 API
        try:
            from IPython.Shell import IPShellEmbed as embed

            ipshell = embed(banner="Sequoia Shell")
            ipshell()
        except ImportError:
            import code

            code.interact("Sequoia Shell", local=locals())


def main(args=sys.argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", dest="conffile", default="development.ini")

    kwargs, _ = parser.parse_known_args(args[1:])
    kwargs = vars(kwargs)

    config_uri = kwargs.pop("conffile")
    setup_logging(config_uri)
    settings = get_appsettings(config_uri, "main")

    if "sequoia.export" not in settings:
        settings = get_appsettings(config_uri, "sequoia")

    kwargs["settings"] = settings

    subparsers = parser.add_subparsers(title="action")

    sp_setup = subparsers.add_parser("shell")
    sp_setup.set_defaults(func=shell)

    services = aslist(settings["sequoia.export"], flatten=False)
    for service in services:
        if service.startswith("#"):
            continue

        try:
            module_name = "%s.bin" % service
            binscript = importlib.import_module(module_name)
        except ImportError as err:
            # Our specific module wasn't found
            if repr(module_name) in repr(err):
                log.info("%s does not handle the sequoia CLI" % service)
                continue

            # Another module wasn't found, probably because the script
            # did a bad import
            raise

        else:
            binscript.bind_subparser(subparsers)

    kwargs = parser.parse_args(args[1:])
    kwargs = vars(kwargs)
    kwargs["settings"] = settings
    kwargs.pop("conffile")

    func = kwargs.pop("func", shell)
    log.info("Running %s:%s" % (inspect.getmodule(func).__name__, func.__name__))
    func(**kwargs)


if __name__ == "__main__":
    main()
