# -*- coding: utf-8 -*-
import logging
from functools import wraps
from inspect import isclass

from cornice.resource import view
from pyramid.decorator import reify
from pyramid.response import Response

log = logging.getLogger(__name__)


def paginated_view(**kw):
    def wrapper(func):
        original = func

        @wraps(func)
        def wrapped(self, *args, **kwargs):
            param_get = self.request.validated.get("querystring", {})

            page = param_get.pop("page", 1)
            per_page = param_get.pop("per_page", self.per_page) or 1

            limit = param_get.pop("limit", per_page)
            offset = param_get.pop("offset", (page - 1) * limit)

            sort_by = param_get.pop("sort_by", "id")

            kwargs.update({"limit": limit, "offset": offset, "sort_by": sort_by})

            ret = func(self, *args, **kwargs)
            if isinstance(ret, (Response, str)):
                return ret

            if isinstance(ret, dict):
                xtra = self.build_list_headers(
                    page,
                    per_page,
                    sort_by,
                    ret.get("X-Total-Count"),
                    ret.get("X-Filtered-Count"),
                )
                ret.update(xtra)
                return ret
            else:
                # (count, list) tuple assumed or (total_count, list,
                # filtered_count)
                if len(ret) == 2:
                    len_, list_ = ret
                    self.set_list_headers(page, per_page, sort_by, len_)
                elif len(ret) == 3:
                    len_, list_, filtered_len = ret
                    self.set_list_headers(page, per_page, sort_by, len_, filtered_len)
                else:
                    raise RuntimeError("Invalid return: %r" % (ret,))
                return list_

        wrapped.unwrapped = original

        schema = kw.get("schema")
        if not schema:
            raise RuntimeError("A schema is mandatory to use @paginated_view")

        # XXX bound_schema_validator won't have been called yet
        if isclass(schema):
            schema = schema()

        # To help migration from sequoia.base to here
        if not schema.get("querystring"):
            raise RuntimeError("querystring Node is mandatory in your schema")

        return view(**kw)(wrapped)

    return wrapper


class APIBase(object):
    per_page = 50

    @reify
    def session(self):
        raise NotImplementedError

    def __init__(self, context, request):
        self.request = request
        self.context = context

    def build_list_headers(self, page, per_page, sort_by, len_, filtered_len=None):
        if not per_page or len_ is None:
            log.warning(
                "Cannot build pagination header with given length %s "
                "and per page %s",
                len_,
                per_page,
            )
            return {}
        if filtered_len:
            len_ = filtered_len
        # Minimum to 1 to avoid page 0 if len_ is 0
        if len_ > 0:
            last_page = int(len_ / per_page) + (1 if len_ % per_page else 0)
        else:
            last_page = 1

        if page >= last_page:
            # In case there is only one page, you are on the last page,
            # don't say to the consumer there is another page
            # which is different of the current one.
            return {}

        params = self.request.GET.copy()
        params.update({"per_page": per_page, "page": page + 1})
        if sort_by:
            if isinstance(sort_by, (list, tuple)):
                sort_by = ",".join(sort_by)
            # this is for restela api
            elif isinstance(sort_by, str) and sort_by.endswith(" desc"):
                sort_by = "-%s" % (sort_by[:-5],)
            params["sort_by"] = sort_by

        url_next = self.request.current_route_url(_query=params)

        params["page"] = last_page
        url_last = self.request.current_route_url(_query=params)

        # http://tools.ietf.org/html/rfc5988
        link = '<%s>; rel="next", ' % url_next
        link += '<%s>; rel="last"' % url_last
        return {"link": link}

    def set_list_headers(self, page, per_page, sort_by, len_, filtered_len=None):
        if not per_page:
            return

        headers_ = self.build_list_headers(page, per_page, sort_by, len_, filtered_len)
        headers = []
        headers.append(
            ("Link", headers_["link"]),
        )
        headers.append(
            ("X-Total-Count", str(len_)),
        )
        if filtered_len:
            headers.append(
                ("X-Filtered-Count", str(filtered_len)),
            )

        # if we don't explicitly ask for latin1, waitress.task.WSGITask
        # tests k.__class__ is str and will fail...
        self.request.response.headerlist.extend(
            [(key.encode("latin1"), value.encode("latin1")) for key, value in headers]
        )

    def get_internal_url(self, route_name, **param):
        """
        Retrieve an URL for a given route, with optional parameters

        Ignores the HATEOAS variables that will otherwise generate
        the URL for the outside world

        :param route_name: usually the name of the @resource
        :type route_name: str

        :return: the URL as string

        """
        removed = {}
        try:
            for key in list(self.request.environ):
                if key.lower().startswith("http_hateoas"):
                    removed[key] = self.request.environ.pop(key)

            return self.request.route_url(route_name, **param)
        finally:
            self.request.environ.update(removed)

    def get_url(self, route_name, **param):
        """
        Retrieve an URL for a given route, with optional parameters

        :param route_name: usually the name of the @resource
        :type route_name: str

        :return: the URL as string

        """
        return self.request.route_url(route_name, **param)
