# -*- coding: utf-8 -*-
"""

Handle Sequoia API Dry Run Mechanism.

A dry run is about validating that sending parameters are correct.

To do a dry run call, a variable ``Dry-Run: 1`` should be added to the http
query header.


In Dry run. The http response code should be 200 event if the sent parameters
are incorrect in the normal call, it means, the query successfully build the
errors list. Errors cases are reserve to things like the query is not parsable.


.. todo::

    Dry Run should explain what is going to be done, like:

     * product price that are going to be charge in case the client do the
     request without dry run

     * contracts that are implicitly accepted by doing the call

"""


import colander
from cornice.resource import view as cornice_view
from pyramid.settings import asbool

from ..helpers.colander import LocalizedString
from ..helpers.renderer import JsonRenderer


class DryRunReturn(colander.MappingSchema):
    warning = colander.SchemaNode(
        colander.String(), missing=colander.drop, location="header"
    )
    status = colander.SchemaNode(colander.String())

    @colander.instantiate(name="errors", missing=colander.drop)
    class Errors(colander.SequenceSchema):
        @colander.instantiate()
        class Error(colander.MappingSchema):
            name = colander.SchemaNode(colander.String(), missing=colander.drop)
            location = colander.SchemaNode(colander.String())
            description = colander.SchemaNode(LocalizedString())

    @colander.instantiate(name="infos", missing=colander.drop)
    class Infos(colander.SequenceSchema):
        @colander.instantiate()
        class Info(colander.MappingSchema):
            field = colander.SchemaNode(colander.String())
            type = colander.SchemaNode(colander.String())
            required = colander.SchemaNode(colander.Boolean())

    # TODO add call charges and currency
    # @colander.instantiate(name='cost', missing=colander.drop)
    # class Cost(colander.MappingSchema):
    #     amount = colander.SchemaNode(colander.Decimal())
    #     currency = colander.SchemaNode(colander.String())

    @colander.instantiate(name="contracts", missing=colander.drop)
    class Contracts(colander.SequenceSchema):
        @colander.instantiate()
        class Contract(colander.MappingSchema):
            id = colander.SchemaNode(colander.String())
            name = colander.SchemaNode(colander.String())
            href = colander.SchemaNode(colander.String(), missing=colander.drop)
            registry_contract_href = colander.SchemaNode(
                colander.String(), missing=colander.drop
            )


def is_dry_run(request):
    """Return `True` if the query is in dry run mode, `False` otherwise"""
    return asbool(request.headers.get("Dry-Run", "0"))


def dry_run_handler(request):
    """Default handler for the dry run method"""
    ret = {"status": "success"}
    if request.errors:
        ret = {"status": "error", "errors": request.errors}

    if not hasattr(request, "validated"):
        return ret

    # list of dicts, extra has the priority if both are present
    # This will check for request.validated[..]
    # and request.validated['extra'][..] to return those
    # values in dry-run responses
    to_check = request.validated, request.validated.get("extra")

    for validated in filter(None, to_check):
        if "contracts" in validated:
            ret["contracts"] = validated["contracts"]

        if "warning_contract" in validated:
            # warning-value = warn-code SP warn-agent SP warn-text
            #                                       [SP warn-date]
            # from https://tools.ietf.org/html/rfc2616#page-150
            # 199 Miscellaneous warning
            # The warning text MAY include arbitrary information to be
            # presented to a human user, or logged. A system receiving this
            # warning MUST NOT take any automated action, besides presenting
            # the warning to the user.
            #
            # `-` is the warn-agent
            ret["warning"] = "199 - %s" % (validated["warning_contract"],)

        if "infos" in validated:
            ret["infos"] = validated["infos"]

    return ret


def dry_run_decorator(view_config):
    """
    This decorator shunt the API Call implementation to ensure a dry run,
    do nothing else except the validation.
    """

    def wrapper(view):
        """In case of dry run, ensure the implementation method is not called,
        and return the validation informations"""

        def wrapped(context, request):
            if is_dry_run(request):
                request.context.return_schema = DryRunReturn
                for validator in view_config.get("validators", []):
                    validator(request, **view_config)
                response = view_config.get("dry_run_handler", dry_run_handler)(request)
                # The JsonRenderer also append correct header for the response
                request.response.text = JsonRenderer(None)(
                    response, {"request": request}
                )
                return request.response
            else:
                return view(context, request)

        return wrapped

    return wrapper


def dry_run_view(**kwargs):
    """
    A decorator that replace the @view decorator to add dry run support
    to the view.
    """

    if "decorator" not in kwargs:
        kwargs["decorator"] = []

    # XXX cornice will raise an error on unknown value in kwargs
    #     so we need to copy it for us and drop it for cornice
    kwargs["decorator"].append(dry_run_decorator(kwargs.copy()))
    kwargs.pop("dry_run_handler", None)
    return cornice_view(**kwargs)
