# -*- coding: utf-8 -*-

import logging
import re
import sys
from typing import Any, Mapping

from pyramid.config import aslist
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import HTTPException
from pyramid.paster import get_appsettings as get_paster_settings

from sequoia.base3.helpers.settings import patch_settings_from_env

from .helpers.pyramid import (
    AcceptPredicate,
    DocumentationPredicate,
    ExtractCStructPredicate,
    HttpVaryPredicate,
)
from .helpers.renderer import ReturnSchemaPredicate
from .helpers.sqla import create_engine

log = logging.getLogger(__name__)


def includeme(config):
    HTTPException.mapping = {}  # add a mapping attribute to the class
    HTTPException.errors = {}
    # in order map values in error explanation used as translated string.

    # Configure Sentry client
    config.include("sequoia.base3.sentry")

    settings = config.registry.settings
    config.add_translation_dirs("locales/")
    try:
        config.add_translation_dirs("colander:locale")
    except ConfigurationError:
        log.error("Colander locales not found (Fix the colander package)")

    if "sequoia.config" in settings:
        from gandiconf import Configuration

        for file_ in aslist(settings["sequoia.config"]):
            name, path = file_.split(":", 1)
            Configuration.load(path, name)
            log.info('Configuration "%s" loaded', file_)


def export_api(config):
    # XXX This part should be in the includeme function
    # But as it is include many times it makes problem
    # that should be fixed before that
    config.include("cornice")
    config.include("pyramid_tm")

    # XXX: we must add the predicate before scanning the routes
    # otherwise cornice will raise an unknown predicate
    log.info("add the return_schema view predicate")
    config.add_view_predicate("return_schema", ReturnSchemaPredicate)
    config.add_view_predicate("http_vary", HttpVaryPredicate)
    config.add_view_predicate("documentation", DocumentationPredicate)
    config.add_view_predicate("accept_only", AcceptPredicate)
    config.add_view_predicate("extract_cstruct", ExtractCStructPredicate)
    # End of part

    settings = config.registry.settings

    apis = aslist(settings["sequoia.export"], flatten=False)

    for api in apis:
        api = api.strip()

        if not api.startswith("#"):
            log.info("Expose API {api}".format(api=api))

            api_name = re.sub(r"sequoia[._]api[._]", "", api)
            sql_url = "sql.%s.url" % api_name

            if sql_url in settings:
                config.scan("%s.models" % api)

                if settings[sql_url] != "sqlite://":
                    log.info("Creating SQLA engine for %s", api)
                    create_engine(api_name, settings)

            try:
                config.scan(
                    api,
                    # XXX   MUST BE unicode in PY3, bytes in PY2
                    # => use str() to cast the correct type
                    ignore=[str("%s.tests" % api)],
                )
            except ConfigurationError:
                log.exception(
                    "An error occurred when trying " "to export api: {}".format(api)
                )
                sys.exit(1)
            try:
                config.add_translation_dirs("%s:locales" % api)
            except ConfigurationError:
                log.info("No locales found for api %s" % api)

    config.add_route("index", "/", request_method="GET")
    config.add_route("monitoring", "/monitoring", request_method="GET")

    config.scan("sequoia.base3.api")
    config.scan("sequoia.base3.errors")

    # Setup zipkin is the collector has been set
    if settings.get("zipkin.collector"):
        zipkin_apis = filter_config_comments(apis)
        zipkin_default_name = "sequoia.base3-api"
        if len(zipkin_apis) > 0:
            zipkin_default_name = apis[0]
        zipkin_name = settings.get("zipkin.service_name", zipkin_default_name)
        settings["zipkin.service_name"] = zipkin_name
        config.include("zipkin.binding.pyramid")


def filter_config_comments(settings):
    """Filter out comments from pyramid config lists"""
    return [line for line in settings if not line.startswith("#")]


def get_appsettings(config_uri: str, section: str = "main") -> Mapping[str, Any]:
    settings = get_paster_settings(config_uri, section)
    patch_settings_from_env(settings)
    return settings
