"""
Display http exceptions in json format
"""
import json
import logging

from colander import Invalid
from pyramid.exceptions import Forbidden
from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPClientError,
    HTTPConflict,
    HTTPGatewayTimeout,
    HTTPInternalServerError,
    HTTPMethodNotAllowed,
    HTTPNotFound,
    HTTPNotImplemented,
    HTTPRequestTimeout,
    HTTPServiceUnavailable,
    HTTPUnauthorized,
    HTTPUnprocessableEntity,
)
from pyramid.response import Response
from pyramid.view import view_config

from sequoia.base3.helpers.i18n import gettext as _

try:
    from celery.exceptions import TimeoutError as CeleryTimeoutError
except ImportError:
    CeleryTimeoutError = None


log = logging.getLogger(__name__)


# XXX We can't define our own JSONEncoder as the `TranslationString` object
#     will be see as a "Str" one thus using "local" encode_basestring and not
#     calling 'JSONEncoder.default'
def localized_errors(lczr, errors):
    for error in errors:
        if "description" in error:
            error["description"] = lczr(error["description"])
    return errors


@view_config(context=Forbidden)
def http_forbidden(exc, request):
    # If the user is not authenticated, then the unauthorized error shall
    # be raised
    if request.unauthenticated_userid is None:
        exc = HTTPUnauthorized(
            explanation=_(
                "The server could not verify that you authorized to access"
                " the document you requested. Either you supplied the wrong"
                " credentials (e.g., bad api key), or your access token has"
                " expired"
            ),
        )
    return http_exception(exc, request)


@view_config(context=HTTPConflict)
@view_config(context=HTTPRequestTimeout)
@view_config(context=HTTPInternalServerError)
@view_config(context=HTTPNotFound)
@view_config(context=HTTPMethodNotAllowed)
@view_config(context=HTTPServiceUnavailable)
@view_config(context=HTTPGatewayTimeout)
@view_config(context=HTTPClientError)
@view_config(context=HTTPNotImplemented)
def http_exception(exc, request):
    # If the view has two formal arguments, the first is the context.
    # The context is always available as ``request.context`` too.
    translate = request.localizer.translate
    mapping = getattr(exc, "mapping", {})
    object_name = getattr(exc, "object_name", exc.__class__.__name__)
    error = {
        "object": object_name,
        "cause": translate(exc.title),
        "code": exc.code,
        "message": translate(exc.explanation, mapping=mapping),
    }
    if exc.errors:
        error["errors"] = localized_errors(translate, exc.errors)
    response = Response(json.dumps(error))
    response.content_type = "application/json"
    response.status_int = exc.code
    return response


@view_config(context=Invalid)
def colander_invalid(exc, request):
    translate = request.localizer.translate
    try:
        errors = exc.asdict(translate=translate)
    except TypeError:
        errors = exc.asdict()
    error = {
        "object": (
            request.matched_route.name
            if request.matched_route
            else translate(_("Unknown"))
        ),
        "cause": translate(_("Bad Request")),
        "code": 400,
        "errors": errors,
    }
    response = Response(json.dumps(error))
    response.content_type = "application/json"
    response.status_int = 400
    return response


@view_config(context=Exception)
def internal_server_error(exc, request):
    # If internal server errors are disabled, directly re-raise the exception
    # This is used so tests will actually show the correct error.
    if internal_server_error.disabled:
        raise

    translate = request.localizer.translate

    if isinstance(exc, Response):
        return exc

    error = {
        "object": (
            request.matched_route.name
            if request.matched_route
            else translate(_("Unknown"))
        ),
        "cause": translate(_("Internal Server Error")),
        "code": 500,
        "message": translate(
            _(
                "The server has either erred or is incapable "
                "of performing the requested operation."
            )
        ),
    }
    response = Response(json.dumps(error))
    response.content_type = "application/json"
    response.status_int = 500

    log.exception("Unexpected error, return status 500")

    return response


internal_server_error.disabled = False


def colander_error_handler(errors):
    raise HTTPUnprocessableEntity()


@view_config(context=HTTPUnprocessableEntity)
@view_config(context=HTTPBadRequest)
def http_unprocessable_entity(exc, request):
    # If the view has two formal arguments, the first is the context.
    # The context is always available as ``request.context`` too.
    translate = request.localizer.translate
    mapping = getattr(exc, "mapping", {})
    object_name = getattr(exc, "object_name", exc.__class__.__name__)
    error = {
        "object": object_name,
        "cause": translate(exc.title),
        "code": exc.code,
        "message": translate(exc.explanation, mapping=mapping),
    }
    if getattr(request, "errors", None):
        # XXX should be a list of dict with keys:
        #     - description
        #     - location
        #     - name
        error["errors"] = localized_errors(translate, request.errors)
    response = Response(json.dumps(error))
    response.content_type = "application/json"
    response.status_int = exc.code
    return response


if CeleryTimeoutError:

    @view_config(context=CeleryTimeoutError)
    def celery_timeout_error(exc, request):
        raise HTTPRequestTimeout()  # might need to be localizer ?
