# -*- coding: utf-8 -*-
import datetime
import logging
import re
import uuid
import warnings

import colander
import idna
import IPy
import isodate
from colander import ContainsOnly
from colander import Date as Date_
from colander import DateTime as DateTime_
from colander import Integer, Invalid, List
from colander import MappingSchema as MappingSchema_
from colander import (
    OneOf,
    Range,
    Regex,
    SchemaNode,
    SchemaType,
    SequenceSchema,
    String,
    drop,
    null,
)
from pyramid.httpexceptions import HTTPAccepted, HTTPCreated, HTTPNoContent

from .i18n import gettext as _
from .sqla import _Base as ModelBase

log = logging.getLogger(__name__)


class Mapping(colander.Mapping):
    """
    Custom mapping type for SQLAlchemy objects: validation will be done
    on the SQLAlchemy object itself, and not a dict created from it.

    Otherwise, creating a dict from the SQLAlchemy object will evaluate all its
    properties, including all relations. Triggering unwanted SQL selects.

    """

    def _validate(self, node, value):
        if isinstance(value, ModelBase):
            return value
        return super(Mapping, self)._validate(node, value)


###
# Colander preparers
# http://docs.pylonsproject.org/projects/colander/en/latest/basics.html#preparing-deserialized-data-for-validation


def sql_like_preparer(value):
    """Replace url wilcard by sql wildcard.

      * '*' becomes '%'
      * '?' becomes '_'
      * backslash is used to escape wildcard

    You can use it with:

    _username = colander.SchemaNode(colander.String(),
                                    name='~username',
                                    preparer=sql_like_preparer,
                                    validator=colander.Length(min=4, max=24),
                                    missing=colander.drop)
    """
    if not value:
        return value
    value = value.replace("%", "\\%").replace("_", "\\_")
    value = re.sub(r"([^\\])\*", "\\1%", value)
    value = re.sub(r"([^\\])\?", "\\1_", value)
    value = re.sub(r"(^\*)", "%", value)
    value = re.sub(r"(^\?)", "_", value)
    return value.replace(r"\*", "*").replace(r"\?", "?")


def sort_by_preparer(value):
    """
    Split the GET parameter sort=field,field,-field
    to ['field' , 'field', '-field']
    """
    if not value:
        return value
    return [val.strip() for val in value.split(",") if val]


@colander.deferred
def nested_missing(node, kw):
    """
    To be used on an optional 'SchemaNode' will generate a 'default' dict
    based on the 'missing' value of all available sub nodes
    """
    if not isinstance(node, colander.MappingSchema):
        raise RuntimeError("node has to be a MappingSchema")
    missing = {}
    for child in node.children:
        if child.missing == colander.drop:
            continue
        missing[child.name] = child.missing
    return missing if missing else colander.drop


@colander.deferred
def missing_country(node, kw):
    request = node.bindings["request"]
    return request.headers.get("X-Country-Code", colander.drop)


class MappingSchema(MappingSchema_):
    schema_type = Mapping

    def __init__(self, *arg, **kw):
        super(MappingSchema, self).__init__(*arg, **kw)
        for prop in getattr(self, "_properties", []):
            if not self.get(prop):
                self.add(getattr(self, prop))

    def __getitem__(self, name):
        """Get a subnode by name."""
        val = self.get(name)
        return val


class UUID(colander.SchemaType):
    err_template = _("Invalid UUID")

    @staticmethod
    def _uuid(node, value):
        try:
            # UUID() does not check type, do it here
            if not isinstance(value, str):
                raise ValueError

            return uuid.UUID(value)
        except ValueError:
            raise colander.Invalid(
                node, _('"${val}" is not an UUID', mapping={"val": value})
            )

    def serialize(self, node, appstruct):
        if not appstruct:
            return colander.null

        if not isinstance(appstruct, uuid.UUID):
            # Will raise an Invalid error if badly formatted / bad type
            appstruct = self._uuid(node, appstruct)

        return str(appstruct)

    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null

        if isinstance(cstruct, uuid.UUID):
            return str(cstruct)

        # Will raise a Invalid error if badly formatted / bad type
        self._uuid(node, cstruct)

        return cstruct


class UUIDList(SequenceSchema):
    _ = SchemaNode(UUID())


class ListOf(List):
    def _validate(self, node, value):
        return List.deserialize(self, node, value)

    def _impl(self, node, value, callback):
        value = self._validate(node, value)
        error = None
        result = []

        for num, subval in enumerate(value):
            try:
                result.append(callback(node, subval))
            except Invalid as e:
                if error is None:
                    error = Invalid(node)
                error.add(e, num)

        if error is not None:
            raise error

        return list(result)

    def deserialize(self, node, cstruct):
        if cstruct is colander.null:
            return colander.null

        def callback(subnode, subval):
            if isinstance(subnode, MappingSchema):
                return subnode.deserialize(subval)
            else:
                return subnode.deserialize(subnode, subval)

        return self._impl(node, cstruct, callback)


class StrictMappingSchema(MappingSchema):
    """Strictly check GET and POST params"""

    @staticmethod
    def schema_type():
        """override default schema_type to set unknown to raise"""
        return MappingSchema.schema_type(unknown="raise")


class NoParamSchema(StrictMappingSchema):
    """A route that take no parameter"""


class FilteredPaginationSchema(MappingSchema):
    """
    Pagination parameters

    @key page: the wanted page (>0)
    @key per_page: the number of element per page
    """

    # TODO when we don't need to pass session to models anymore, we have to
    # limit the possibles fields to YYY.list
    page = SchemaNode(Integer(), validator=Range(min=1), missing=1)
    per_page = SchemaNode(Integer(), validator=Range(min=1), missing=drop)
    sort_by = SchemaNode(String(), missing="id")
    sharing_id = SchemaNode(UUID(), missing=drop)


class PaginationSchema(MappingSchema):
    querystring = FilteredPaginationSchema(missing=nested_missing)


class CreatedAt(Regex):
    """Validate a creation date comparison"""

    def __init__(self, msg=None):
        regexp = "^[<>][12][0-2][0-9]{2}-[01][0-9]-[0-3][0-9]$"
        if msg is None:
            msg = "Invalid date comparison"
        super(CreatedAt, self).__init__(regexp, msg=msg)


class SortBy(String):
    def serialize(self, node, appstruct):
        return super(SortBy, self).serialize(node, appstruct)

    def deserialize(self, node, cstruct):
        warnings.warn("Use the sort_by_preparer instead", DeprecationWarning)
        result = super(SortBy, self).deserialize(node, cstruct)
        if result:
            result = sort_by_preparer(result)
        return result


class SortByFields(ContainsOnly):
    # for the ContainsOnly
    err_template = _(
        "One or more of the field is not acceptable for sort_by: "
        "available elements ${choices}"
    )
    err_template_len = _(
        "Too many elements in the sort_by field, " "${max_choices} max"
    )
    err_template_content = _("Duplicate element found in sort_by field")

    def __init__(self, choices, max_choices=3):
        self.max_choices = max_choices
        super(SortByFields, self).__init__(
            choices + ["-%s" % choice for choice in choices]
        )

    def __call__(self, node, value):
        if len(value) > self.max_choices:
            err = _(
                self.err_template_len,
                mapping={"max_choices": self.max_choices, "len": len(value)},
            )
            raise Invalid(node, err)

        # we don't authorize neither field,-field nor field,field
        filtered_val = set([val[1:] if val.startswith("-") else val for val in value])
        if len(filtered_val) != len(value):
            err = _(self.err_template_content)
            raise Invalid(node, err)
        return super(SortByFields, self).__call__(node, value)


class ListableMappingSchema(MappingSchema):
    """should be used to have extended returns when making a return for
    GET and list"""

    _list = False
    _listable = []

    def __init__(self, *args, **kwargs):
        MappingSchema.__init__(self, *args, **kwargs)

        if kwargs.get("is_list"):
            self._list = True

        for name, node_list, node in self._listable:
            item = node_list if self._list else node
            if item is None:
                continue

            self.add(item)
            self.__setitem__(name, item)
            setattr(self.__class__, name, item)


class ListReturn(MappingSchema):

    """
    Has to be used when a view is decorated with the paginated_view decorator
    """

    length = SchemaNode(Integer(), name="X-Total-Count", location="header")
    filtered_length = SchemaNode(
        Integer(), name="X-Filtered-Count", location="header", missing=drop
    )
    link = SchemaNode(String(encoding="utf-8"), location="header", missing=drop)
    body = None  # override me


# XXX Date And Datetime override should be consistent


class Date(Date_):
    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null
        if isinstance(cstruct, (datetime.datetime, datetime.date)):
            cstruct = cstruct.isoformat()
        return super(Date, self).deserialize(node, cstruct)


class DateTime(DateTime_):
    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null
        if isinstance(cstruct, datetime.datetime):
            return cstruct
        return super(DateTime, self).deserialize(node, cstruct)


class Duration(String):
    err_template = _("${value} is not a valid ISO8601 duration")

    def _validate(self, val):
        try:
            return isodate.parse_duration(val)
        except isodate.isoerror.ISO8601Error as exc:
            raise ValueError(exc)

    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null
        ret = super().deserialize(node, cstruct)

        try:
            ret = self._validate(ret)
        except ValueError:
            raise Invalid(
                node,
                _(self.err_template, mapping={"value": ret}),
            )
        return ret


class DurationReturn(colander.SchemaType):
    err_template = _("${value} is not a valid datetime.timedelta")

    def _validate(self, val):
        if not isinstance(val, datetime.timedelta):
            raise ValueError()

        return isodate.duration_isoformat(val)

    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null
        try:
            cstruct = self._validate(cstruct)
        except ValueError:
            raise Invalid(
                node,
                _(self.err_template, mapping={"value": cstruct}),
            )
        return cstruct


class CIDR(String):
    err_template = _("${value} is not a valid CIDR")

    def _validate(self, val):
        return IPy.IP(val)  # raise the ValueError

    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null

        ret = super(CIDR, self).deserialize(node, cstruct)

        try:
            self._validate(ret)
        except ValueError:
            raise Invalid(
                node,
                _(self.err_template, mapping={"value": ret}),
            )
        return ret


class IP(CIDR):
    err_template = _("${value} is not a valid IP")

    def _validate(self, val):
        ip = super(IP, self)._validate(val)
        try:
            if len(ip) > 1:
                raise ValueError("Not an IP Address")
        except OverflowError:
            # http://stackoverflow.com/questions/15650878/returning-from-len-when-64-bits
            raise ValueError("Not an IP Address")


class MissingDrop(MappingSchema):
    _kls = None

    def __init__(self, kls):
        self._kls = kls

    def deserialize(self, cstruct):
        if not cstruct:
            return drop
        return self._kls.deserialize(cstruct)


class ScalarSequenceSchema(SequenceSchema):
    """
    SequenceSchema which accept scalar value and turn them into a list.
    Specifically useful for sequence in querystring.

    >>> class RustyStatusQuerystring(colander.SequenceSchema):
    >>>     _ = colander.SchemaNode(colander.String)
    >>>
    >>> class ShinyStatusQuerystring(ScalarSequenceSchema):
    >>>     _ = colander.SchemaNode(colander.String)
    >>>
    >>> #/entity?param=a&param=b
    >>> RustyStatusQuerystring().deserialize(['a', 'b'])
    >>> ['a', 'b']
    >>> ShinyStatusQuerystring().deserialize(['a', 'b'])
    >>> ['a', 'b']
    >>>
    >>> #/entity?param=a
    >>> RustyStatusQuerystring().deserialize('a')
    >>> colander.Invalid
    >>> ShinyStatusQuerystring().deserialize('a')
    >>> ['a']
    """

    @staticmethod
    def schema_type():
        return colander.Sequence(accept_scalar=True)


class StringList(SequenceSchema):
    _ = SchemaNode(String())


class RestelaSortBy(String):
    """sequoia API wants '-' to sort desc, this will convert from one to
    the other
    """

    def deserialize(self, node, cstruct):
        ret = super(RestelaSortBy, self).deserialize(node, cstruct)
        if ret and ret.startswith("-"):
            ret = ret[1:] + " desc"
        return ret


class SplittedString(String):
    def deserialize(self, node, cstruct):
        ret = super(SplittedString, self).deserialize(node, cstruct)
        return ret.split(",") if ret else ret


class ManyOf(OneOf):
    """This will validate that values are all in a list"""

    def serialize(self, value):
        return value

    def deserialize(self, value):
        return value

    def __call__(self, node, values):
        ret = []
        errors = []
        for val in values:
            value = self.deserialize(val)

            if val in ret or value in ret:
                err = Invalid(node, "Can not have twice the same sort_by '%s'" % value)
                errors.append(err)

            try:
                super(ManyOf, self).__call__(node, value)
            except Invalid as exp:
                errors.append(exp)

            ret.append(self.serialize(value))

        if errors:
            error = Invalid(node)
            for err in errors:
                error.add(err)
            raise error

        return ret


class ManyOfSortBy(ManyOf):
    """This is a many of with '-' support"""

    sign = ""

    def serialize(self, value):
        return "%s%s" % (self.sign, value)

    def deserialize(self, value):
        self.sign = ""
        if value.startswith("-"):
            self.sign = "-"
            return value[1:]
        return value


class HTTPCreatedReturn(MappingSchema):
    status_code = SchemaNode(
        Integer(), location="status_code", missing=HTTPCreated.code
    )
    location = SchemaNode(String(encoding="utf-8"), location="header")
    content_location = SchemaNode(
        String(encoding="utf-8"), location="header", missing=drop  # Deprecated
    )
    content_location_ = SchemaNode(
        String(encoding="utf-8"),
        location="header",
        name="Content-Location",
        missing=drop,
    )
    message = SchemaNode(String(), location="body", missing=_("Resource Created"))


class HTTPAcceptedReturn(MappingSchema):
    status_code = SchemaNode(
        Integer(), location="status_code", missing=HTTPAccepted.code
    )
    location = SchemaNode(String(encoding="utf-8"), location="header", missing=drop)
    content_location = SchemaNode(
        String(encoding="utf-8"), location="header", missing=drop  # Deprecated
    )
    content_location_ = SchemaNode(
        String(encoding="utf-8"),
        location="header",
        name="Content-Location",
        missing=drop,
    )
    message = SchemaNode(
        String(),
        location="body",
        missing=_("Request Accepted"),
    )


class HTTPNoContentReturn(MappingSchema):
    status_code = SchemaNode(
        Integer(), location="status_code", missing=HTTPNoContent.code
    )


class UUIDHTTPCreatedReturn(HTTPCreatedReturn):
    uuid = SchemaNode(UUID(), location="body")


class IDHTTPCreatedReturn(HTTPCreatedReturn):
    id = SchemaNode(UUID(), location="body")
    href = SchemaNode(String(), location="body", missing=colander.drop)


class IDHTTPAcceptedReturn(HTTPAcceptedReturn):
    id = SchemaNode(UUID(), location="body")
    href = SchemaNode(String(), location="body", missing=colander.drop)


class Multiple(object):
    """This will validate that the node is a multiple of _multiple"""

    def __init__(self, _multiple):
        self._multiple = _multiple

    def __call__(self, node, value):
        if value % self._multiple != 0:
            raise Invalid(node, "Is not a multiple of %d" % self._multiple)
        return


def cast_idna(node, value, allow_numbers=False):
    prepend = ""
    # client is sending an IP address instead of a domain name ..
    if not allow_numbers and re.match("^[.0-9]+$", value):
        raise Invalid(
            node, "fqdn must contain at least one " "letter (param: '%s')" % value
        )

    # If the name starts with *., the idna module will refuse to encode it,
    # remove the '*.' and re-append it afterwards
    # It's currently only accepted by WildcardFQDN
    if value.startswith("*."):
        value = value[2:]
        prepend = "*."

    if re.search("[^-a-z0-9.@]", value, re.ASCII | re.I):
        try:
            value = idna.encode(value, uts46=True)
            value = value.decode("utf-8")
        except (idna.InvalidCodepoint, idna.core.IDNAError):
            try:
                value = value.encode("idna")
                value = value.decode("utf-8")
            except UnicodeError as e:
                value = prepend + value
                raise Invalid(node, "%s '%s'" % (e, value))

    value = prepend + value
    return value.lower()


class FqdnType(String):
    allow_numbers = False

    def deserialize(self, node, value):
        value = super(FqdnType, self).deserialize(node, value)
        if value:
            value = cast_idna(node, value, allow_numbers=self.allow_numbers)
        return value


class Fqdn(SchemaNode):
    schema_type = FqdnType
    _rxp = (
        # Leading label(s)
        "^("
        "(?!-)"  # no leading dash # noqa
        "[-a-z0-9]{1,63}"  # max 63 characters
        "(?<!-)"  # no trailing dash
        r"\."  # ends with a dot
        ")+"  # indefinite number of labels
        # TLD
        "(?=.{1,63}$)"  # No more than 63 characters
        "(?!-)"  # No leading dash
        "("
        "[-a-z0-9]*"
        "[a-z]"  # at LEAST one letter
        "[-a-z0-9]*"
        r"\.?"  # and maybe a final dot
        ")"
        "(?<!-)$"  # and no trailing dash
    )
    _rxpc = re.compile(_rxp, re.I)

    def validator(self, node, cstruct):
        if len(cstruct) > 255:
            raise Invalid(
                node, _('"${val}" is longer than 255 chars.', mapping={"val": cstruct})
            )
        if not self._rxpc.match(cstruct):
            raise Invalid(
                node, _('"${val}" is not a valid FQDN.', mapping={"val": cstruct})
            )
        return cstruct


class WildcardFqdn(Fqdn):
    # FQDN that may start with *.

    def validator(self, node, cstruct):
        if not cstruct.startswith("*."):
            return super(WildcardFqdn, self).validator(node, cstruct)

        # super() check will be truncated, make it here too
        if len(cstruct) > 255:
            raise Invalid(node, "%s is longer than 255 chars." % cstruct)

        return "*." + super().validator(node, cstruct[2:])


class LocalizedString(String):

    """
    To be used for all 'field_label'
    the 'field_label' need to be set with the right value (gettext one),
    the localizer will be done here
    """

    def deserialize(self, node, cstruct):
        if cstruct:
            # Raise Invalid in case cstruct is not a string object
            super(LocalizedString, self).deserialize(node, cstruct)

            request = node.bindings["request"]
            lczr = request.localizer.translate
            # cstruct should be a TranslationString
            # Localize BEFORE the super() call, to ensure cstruct have
            # a translation domain
            cstruct = lczr(cstruct)

        # In case the string have to be encoded in a specified charset
        return super(LocalizedString, self).deserialize(node, cstruct)


def email_validator():
    return colander.Function(validate_email, _("Invalid email format"))


def validate_email(email):
    """Validate that the email a valid email"""

    if not isinstance(email, str):
        return False

    splitted = email.split("@")
    if len(splitted) != 2:
        return False

    localpart, hostpart = splitted
    localpart = localpart.lower()
    hostpart = hostpart.lower()

    try:
        if hostpart.startswith("xn--"):
            hostpart.encode("utf-8").decode("idna")
        else:
            hostpart = hostpart.encode("idna").decode("ascii")
    except UnicodeError:
        return False

    if len(hostpart) > 255:
        return False

    _local_rgx = r"^[-a-z0-9._+&!#$%\'=?^`{|}~]+$"

    if not re.match(_local_rgx, localpart, re.UNICODE) or not re.match(
        Fqdn._rxp, hostpart, re.UNICODE
    ):
        return False

    return True


class StrippedString(String):
    """
    A classic String, with 2 alterations

    - leading/trailing spaces are removed
    - non-ASCII whitespaces are replaced with an ASCII whitespace (0x20)

    """

    # source https://www.fileformat.info/info/unicode/category/Zs/list.htm
    _re_space = re.compile(
        r"[\x00\x04\x09\x0b"
        r"\u00A0\u1680\u2000\u2001\u2002\u2003\u2004\u205F"
        r"\u2005\u2006\u2007\u2008\u2009\u200A\u202F\u3000"
        r"\u2028\u2029\u202A\u202B\u202C\u202D\u202E\u202F]",
        re.UNICODE,
    )

    _re_trim = re.compile(r"^[\u200B\s]+|[\u200B\s]+$", re.UNICODE)

    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null
        ret = super(StrippedString, self).deserialize(node, cstruct)
        ret = self._re_space.sub(" ", ret)
        # https://phabricator.corp.gandi.net/T82645
        return self._re_trim.sub("", ret)


class NullableSchemaNode(colander.SchemaNode):
    """
    Generic SchemaNode, allowing 'null' usage, deserialized as 'None'.

    Replace usage of `NullableMappingSchema` and `keep_null`.
    Unlike those 2, this SchemaNode is self-sufficient and is compatible with
     any missing you might (or not) set.

    >>> class SampleSchema(colander.MappingSchema):
    >>>     @colander.instantiate(name='body')
    >>>     class AwesomeSchema(colander.MappingSchema):
    >>>         normal_node = colander.SchemaNode(colander.Integer(),
    >>>                                           missing=12)
    >>>         nullable_node = NullableSchemaNode(colander.Integer())
    >>>         nullable_node_with_missing = NullableSchemaNode(
    >>>             colander.Integer(),
    >>>             missing=13,
    >>>         )
    >>>
    >>> # JSON: '{"normal_node": 46, "nullable_node": null}'
    >>> print(request.validated['body'])
    >>> {
    >>>     'normal_node': 46,
    >>>     'nullable_node': None,
    >>>     'nullable_node_with_missing': 13,
    >>> }
    """

    def deserialize(self, cstruct=colander.null):
        """
        `cstruct`, before 'deserialize', comes as:
            - `colander.null` is missing (by default)
            - `None` if json was "null"
        """
        if cstruct is None:
            return None
        return super().deserialize(cstruct)


class ContactType(SchemaType):
    """contact type accept both integer and string for v5 only"""

    def serialize(self, node, appstruct):
        if appstruct is null:
            return null

        return appstruct

    def deserialize(self, node, cstruct):
        if cstruct is null:
            return null

        if not isinstance(cstruct, (int, str)):
            raise Invalid(node, "%s do not validate contact type" % cstruct)

        if cstruct in (4, "reseller"):
            raise Invalid(node, 'Contact type "reseller" is not allowed')

        return cstruct
