# coding=utf-8
"""
This module provides helpers to automatically get data from a valid request.

In cornice ony may add validators to a view that are to be called with
the request as the first and only parameter and we access to the validated
parameters using request's ``validated`` attribute

    >>> class Request(object):
    ...     def __init__(self, foo='foo', bar='bar'):
    ...         self.validated = {
    ...             'extra': {},
    ...             'path': {},
    ...             'headers': {},
    ...             'querystring': {},
    ...             'body': {
    ...                 'foo': foo,
    ...                 'bar': bar,
    ...             }
    ...     def _reset(self):
    ...         self.validated['body'].pop('someid')
    >>> request = Request('pl', 'op')
    >>> def shorten(request, **kwargs):
    ...     body = request.validated['body']
    ...     nick = body['foo'] + body['bar']
    ...     request.validated['extra']['nick'] = nick
    >>> shorten(request)
    >>> request.validated['extra']['nick']
    'plop'


This is how we used to compute an id and pass it through


    >>> import base64
    >>> base64.b64encode('plop')
    'cGxvcA=='
    >>> def load_id(request, **kwargs):
    ...     '''Base64 encode the nick '''
    ...
    ...     params =  request.validated['body']
    ...     if 'nick' not in params:
    ...         return # There is no nick, another previous
    ...                #validator raised and error
    ...     nick = params['nick']
    ...     request.validated['extra']['someid'] = base64.b64encode(nick)
    ...
    >>> load_id(request)
    >>> request.validated['extra']['someid']
    'cGxvcA=='
    >>> request._reset()

The load_id validator takes a lot of lines for something trivial.

    >>> @required_params('nick', location='body')
    ... def compute_id(request, **kwargs):
    ...     nick = request.validated['body']['nick']
    ...     request.validated['extra']['someid'] = base64.b64encode(nick)
    ...
    >>> compute_id(request)
    >>> request.validated['extra']['someid']
    'cGxvcA=='
    >>> request._reset()

Cool we saved the checking lines.

    >>> @fetch_required()
    ... def fetch_and_compute_id(request, nick, **kwargs):
    ...     request.validated['extra']['someid'] = base64.b64encode(nick)
    ...
    >>> fetch_and_compute_id(request)
    >>> request.validated['extra']['someid']
    'cGxvcA=='
    >>> request._reset()


    >>> request.validated['body']['nick'] = None
    >>> fetch_and_compute_id(request)
    {}
    >>> 'someid' in request.validated['extra']
    False



One more line won
"""
import inspect
import logging
from collections import abc
from functools import wraps

import colander
import translationstring
from cornice.validators import extract_cstruct
from pyramid.httpexceptions import HTTPBadRequest

log = logging.getLogger(__name__)


def format_error(request, invalid, location, idx=None):
    lczr = request.localizer.translate
    error_dict = invalid.asdict(translate=lczr)
    for name, msg in error_dict.items():
        if idx is not None:
            # Rename the name of the inner schema by its index
            #
            # class Body(colander.SequenceSchema)
            #     @colander.instantiate()
            #     class Inner(colander.MappingSchema):
            #         name = colander.SchemaNode(...)
            #
            # Inner.name => 0.name
            name = ".".join([str(idx)] + name.split(".")[1:])
        request.errors.add(location, name, msg)


def bound_schema_validator(request, **kwargs):
    schema = kwargs["schema"]
    if inspect.isclass(schema):
        schema = schema()
    schema = schema.bind(request=request)
    my_extract_cstruct = kwargs.get("extract_cstruct", extract_cstruct)
    cstruct = my_extract_cstruct(request)
    # if body couldn't be extract cstruct will be empty and an
    # error will already be set no need to go further
    if request.errors:
        raise HTTPBadRequest()

    translate = translationstring.TranslationStringFactory("colander")

    locations = ["method", "url", "path", "querystring", "body", "header"]
    request.validated["extra"] = {}
    for location in locations:
        if location not in schema:
            continue

        if isinstance(schema[location], colander.MappingSchema):
            request.validated[location] = {}
        else:
            request.validated[location] = []

        if isinstance(request.validated[location], dict):
            if not isinstance(cstruct[location], abc.Mapping):
                # message is only partially translated
                err = "Does not implement dict-like functionality."
                msg = translate(
                    '"${val}" is not a mapping type: ${err}',
                    mapping={
                        "val": cstruct[location],
                        "err": err,
                    },
                )
                request.errors.add(location, "", msg)
                continue

            for child in schema[location].children:
                if location != "body" and isinstance(child.typ, colander.Mapping):
                    # nested dict in querystring such as
                    # `meta_data.dc_code` in the organization /search API
                    prefix = child.name + "."
                    value = {
                        key: val
                        for key, val in cstruct[location].items()
                        if key.startswith(prefix)
                    }
                    value = child.unflatten(value)
                    if not value:
                        if child.missing == colander.drop:
                            continue
                        if child.missing != colander.required:
                            value = child.missing
                else:
                    if (
                        child.name not in cstruct[location]
                        and child.missing == colander.drop
                    ):
                        continue
                    value = cstruct[location].get(child.name, colander.null)
                try:
                    value = child.deserialize(value)
                except colander.Invalid as invalid:
                    format_error(request, invalid, location)
                else:
                    if value != colander.drop:
                        request.validated[location][child.name] = value

            # In case the Mapping is set to preserve, we need to manually
            # set the values in the validated dict
            unknown = schema[location].schema_type().unknown
            if unknown == "preserve":
                for ckey, cvalue in cstruct[location].items():
                    request.validated[location].setdefault(ckey, cvalue)
            elif unknown == "raise":
                additional = set(cstruct[location]) - set(request.validated[location])

                if additional:
                    evalue = {key: cstruct[location][key] for key in additional}

                    msg = translate(
                        'Unrecognized keys in mapping: "${val}"',
                        mapping={"val": evalue},
                    )
                    request.errors.add(location, "", msg)

        else:
            if not isinstance(cstruct[location], abc.Sequence):
                msg = translate(
                    '"${val}" is not iterable',
                    mapping={"val": cstruct[location]},
                )
                request.errors.add(location, "", msg)
                continue

            sub_schema = schema[location].children[0]
            for idx, elem in enumerate(cstruct[location]):
                try:
                    value = sub_schema.deserialize(elem)
                except colander.Invalid as invalid:
                    format_error(request, invalid, location, idx)
                    # XXX need to keep the same index as valid entries
                    #     might failed due to a cornice validator
                    request.validated[location].append(colander.null)
                else:
                    # XXX what if colander.drop in value ?
                    request.validated[location].append(value)

        if schema[location].validator:
            try:
                schema[location].validator(
                    schema[location], request.validated[location]
                )
            except colander.Invalid as invalid:
                request.errors.add(location, "", invalid.msg)

    # XXX We might need to check errors here to:
    #     200 OK id dry_run ?
    #     422 HTTPUnprocessableEntity otherwise ?


def required_params(*args, location="body"):
    """Checks that a request has given cornice validated attribues.
    This is used when we need to check that a previous loader has worked fine.
    this avoid to write early returns since the needed results are not
    available.
    """
    args = frozenset(args)

    def call_with_check(request_function):
        @wraps(request_function)
        def called(request, *cargs, **ckwargs):
            if args.issubset(request.validated.get(location, {})):
                return request_function(request, *cargs, **ckwargs)

        return called

    return call_with_check


def fetch_request_params(location="body"):
    def _(request_validator):
        names = list(inspect.signature(request_validator).parameters.keys())[1:]

        @wraps(request_validator)
        def fetch_and_call(request, **kwargs):
            if location in request.validated:
                kwargs = {name: request.validated[location].get(name) for name in names}
            else:
                kwargs = {name: None for name in names}
            return request_validator(request, **kwargs)

        return fetch_and_call

    return _


def fetch_required(location="body"):
    def wrapper(request_validator):
        names = []
        defaults = {}
        params = inspect.signature(request_validator).parameters
        # drop 'request' (first) and '**kwargs' (last)
        itr = list(params)[1:-1]
        for name in itr:
            if params[name].default != inspect._empty:
                defaults[name] = params[name].default
            else:
                names.append(name)

        @wraps(request_validator)
        def fetch_and_call(request, **kwargs):
            if location not in request.validated:
                return {}

            extra_args = {name: request.validated[location].get(name) for name in names}

            if None in extra_args.values():
                return {}

            extra_with_defaults_args = {
                key: request.validated[location].get(key, val)
                for key, val in defaults.items()
            }
            kwargs.update(extra_with_defaults_args)
            kwargs.update(extra_args)
            return request_validator(request, **kwargs)

        return fetch_and_call

    return wrapper


def fetch_location(request_validator):
    names = []
    params = inspect.signature(request_validator).parameters
    itr = list(params)[1:-1]  # drop 'request' (first) and '**kwargs' (last)
    for name in itr:
        if params[name].default != inspect._empty:
            raise AttributeError(
                "Request validator required parameters "
                "should not have default values"
            )
        names.append(name)

    @wraps(request_validator)
    def fetch_and_call(request, **kwargs):
        params = [request.validated.get(name, {}) for name in names]
        return request_validator(request, *params, **kwargs)

    return fetch_and_call
