import logging
from functools import wraps
from inspect import signature

from cornice.resource import add_view
from cornice.resource import resource as cornice_resource
from cornice.resource import view as cornice_view
from pydantic import ValidationError
from pyramid.httpexceptions import HTTPInternalServerError
from pyramid.response import Response

from sequoia.base3.context import DefaultContext
from sequoia.base3.errors import colander_error_handler as error_handler
from sequoia.base3.helpers.pydantic.pagination import PaginationSchema
from sequoia.base3.helpers.pydantic.renderer import JsonRenderer
from sequoia.base3.helpers.pydantic.request_validator import (
    build_pydantic_resource,
    validate_pydantic_schema,
)
from sequoia.base3.helpers.pydantic.response_schema import (
    CollectionResponseSchema,
    ResponseSchema,
)

log = logging.getLogger(__name__)


def resource(**kwargs):
    kwargs["factory"] = DefaultContext
    kwargs["error_handler"] = error_handler
    return cornice_resource(**kwargs)


def view(**kwargs):
    renderer = JsonRenderer(None)

    def wrapper(func):
        sig = signature(func)
        params = sig.parameters.values()
        annotations = [p.annotation for p in params if p.annotation != p.empty]
        try:
            schema = annotations[0]
        except IndexError:
            raise ValueError(f"{func} is not a valid api handler")
        return_schema = sig.return_annotation
        if not isinstance(return_schema, ResponseSchema):
            return_schema = ResponseSchema[return_schema]
        kwargs["return_schema"] = return_schema

        def my_view(api):
            resource = build_pydantic_resource(api.request, schema)
            try:
                ret = func(api, resource)
            except ValidationError:
                log.exception(
                    "Unexpected validation error in api call %s %s",
                    api.request.method,
                    api.request.matched_route.path,
                )

                raise HTTPInternalServerError

            response = api.request.response
            response.text = renderer(ret, {"request": api.request})
            return api.request.response

        return add_view(my_view, **kwargs)

    return wrapper


def paginated_view(**kw):
    def wrapper(func):
        original = func

        @wraps(func)
        def wrapped(self, page_info: PaginationSchema, *args, **kwargs):
            ret = func(self, *args, **kwargs)
            if isinstance(ret, (Response, str)):
                return ret

            elif isinstance(ret, CollectionResponseSchema):
                xtra = self.build_list_headers(
                    page_info.page,
                    page_info.per_page,
                    page_info.sort_by,
                    ret.headers.get("X-Total-Count"),
                    ret.headers.get("X-Filtered-Count"),
                )
                ret.headers.update(xtra)
            elif isinstance(ret, dict):
                xtra = self.build_list_headers(
                    page_info.page,
                    page_info.per_page,
                    page_info.sort_by,
                    ret.get("headers", {}).get("X-Total-Count"),
                    ret.get("headers", {}).get("X-Filtered-Count"),
                )
                ret.setdefault("headers", {}).update(xtra)
            else:
                raise RuntimeError("Invalid return")
            return ret

        wrapped.unwrapped = original  # type: ignore

        return view(**kw)(wrapped)

    return wrapper


def legacy_view(**kwargs):
    if "renderer" not in kwargs:
        kwargs["renderer"] = "pydantic"
    validators = kwargs.get("validators", [])
    kwargs["validators"] = [validate_pydantic_schema] + validators
    return cornice_view(**kwargs)


def legacy_paginated_view(**kw):
    def wrapper(func):
        original = func

        @wraps(func)
        def wrapped(self, *args, **kwargs):
            param_get = self.request.validated.get("querystring", {})

            page = param_get.pop("page", 1)
            per_page = param_get.pop("per_page", self.per_page) or 1

            limit = param_get.pop("limit", per_page)
            offset = param_get.pop("offset", (page - 1) * limit)

            sort_by = param_get.pop("sort_by", "id")

            kwargs.update({"limit": limit, "offset": offset, "sort_by": sort_by})

            ret = func(self, *args, **kwargs)
            if isinstance(ret, (Response, str)):
                return ret

            elif isinstance(ret, CollectionResponseSchema):
                xtra = self.build_list_headers(
                    page,
                    per_page,
                    sort_by,
                    ret.headers.get("X-Total-Count"),
                    ret.headers.get("X-Filtered-Count"),
                )
                ret.headers.update(xtra)
            elif isinstance(ret, dict):
                xtra = self.build_list_headers(
                    page,
                    per_page,
                    sort_by,
                    ret.get("headers", {}).get("X-Total-Count"),
                    ret.get("headers", {}).get("X-Filtered-Count"),
                )
                ret.setdefault("headers", {}).update(xtra)
            else:
                raise RuntimeError("Invalid return")
            return ret

        wrapped.unwrapped = original  # type: ignore

        if "schema" not in kw:
            kw["schema"] = PaginationSchema
        return legacy_view(**kw)(wrapped)

    return wrapper
