import datetime
import json
import logging
from typing import Any, Type, Union

from pydantic import ValidationError
from pydantic.json import ENCODERS_BY_TYPE
from pyramid.config import Configurator
from pyramid.httpexceptions import HTTPInternalServerError, HTTPNoContent
from pyramid.response import Response

from sequoia.base3.helpers.pydantic.response_schema import (
    ResponseSchema,
    ResponseTypedDict,
)
from sequoia.base3.helpers.renderer.encoder import set_content_type

try:
    from freezegun.api import FakeDatetime
except ImportError:
    FakeDatetime = None


log = logging.getLogger(__name__)


class JSONEncoder(json.JSONEncoder):
    _datetypes = (datetime.date, datetime.datetime)

    # During tests, we will need to serialize faked datetime
    # object generated by freezegun, handle them correctly in decoder,
    # without breaking if freezegun isn't installed
    if FakeDatetime:
        _datetypes += (FakeDatetime,)

    def default(self, o: Any) -> Any:
        # override pydantic datetime serialization to keep Z instead of +00:00
        if isinstance(o, self._datetypes):
            return o.strftime("%Y-%m-%dT%H:%M:%SZ")

        for typ, serializer in ENCODERS_BY_TYPE.items():
            if isinstance(o, typ):
                return serializer(o)
        return super(JSONEncoder, self).default(o)


def render_response(
    schema: Type[ResponseSchema[Any]],
    data: Union[ResponseSchema[Any], ResponseTypedDict],
    response: Response,
) -> str:
    if not schema:
        raise RuntimeError("Missing return_schema predicate")
    if not issubclass(schema, ResponseSchema):
        raise TypeError(f"{schema.__module__}.{schema.__name__}: not a ResponseSchema")

    if isinstance(data, schema):
        validated_data = data
    elif isinstance(data, schema.__fields__["body"].type_):
        validated_data = schema(body=data)
    elif isinstance(data, dict):
        validated_data = schema(**data)
    else:
        raise TypeError(f"{data}: unmanaged type")
    resp = validated_data.dict(include={"status_code", "headers"})

    for key, val in resp["headers"].items():
        response.headers[key] = str(val)

    response.status_code = resp["status_code"]
    if resp["status_code"] == HTTPNoContent.code:
        response.content_type = None
        return ""

    # We don't use the pydantic helper model.json() that dump
    # because we render onlyt the body
    resp = validated_data.dict(exclude_unset=True, exclude={"status_code", "headers"})
    body = resp["body"]
    return json.dumps(body, cls=JSONEncoder)


class JsonRenderer:
    """
    Template Factory to render json using a Pydandic Schema.
    """

    def __init__(self, _renderer_helper):
        ...

    def __call__(self, data, context):
        request = context["request"]
        try:
            resp = render_response(
                request.context.return_schema, data, request.response
            )
        except ValidationError:
            log.exception(
                "Unexpected exception raised while rendering response of %s %s",
                request.method,
                request.matched_route.path,
            )
            raise HTTPInternalServerError()
        set_content_type(request)
        return resp


def includeme(config: Configurator):
    config.add_renderer("pydantic", JsonRenderer)
