"""
Use pydandic for request and response validation.
"""
from typing import Any, Tuple

from pydantic.error_wrappers import ErrorWrapper, get_exc_type
from pydantic.errors import MissingError
from pyramid.request import Request

from .translations import PYDANTIC_ERRORS
from .typing import HttpLocation


def extract_val_from_request(
    request: Request, location: HttpLocation, name: str, is_scalar: bool
) -> Tuple[Any, bool]:
    if location == "path":
        part = request.matchdict
    elif location == "header":
        part = request.headers
    elif location == "querystring":
        part = request.GET
    elif location == "body":
        try:
            part = request.json_body  # should be cached
        except Exception:
            part = None
    else:
        raise ValueError(f"Invalid location: {location}")

    if part is None:
        return None, False

    if is_scalar:
        val = part.getall(name)
    else:
        val = part.get(name)
    return val, name not in part


def translate_err(localizer, err: ErrorWrapper) -> str:
    ctx = err.exc.__dict__
    typ = get_exc_type(err.exc.__class__)
    if typ in PYDANTIC_ERRORS:
        return localizer.translate(PYDANTIC_ERRORS[typ], mapping=ctx)

    msg_template = getattr(err.exc, "msg_template", None)
    if msg_template and msg_template in PYDANTIC_ERRORS:
        return localizer.translate(PYDANTIC_ERRORS[msg_template], mapping=ctx)
    return localizer.translate(str(err.exc))  # fallback


def validate_pydantic_schema(request, schema, **kwargs):
    """Validate the request schema, using the pydantic library."""
    request.validated.update({"path": {}, "header": {}, "querystring": {}, "body": {}})
    data = {}
    # validate field by field
    for field in schema.__fields__.values():
        loc: HttpLocation = field.field_info.extra["location"]
        is_scalar = loc == "querystring" and field.is_complex()
        val, missing = extract_val_from_request(request, loc, field.alias, is_scalar)
        if missing and field.required:
            val, err = None, ErrorWrapper(MissingError(), (loc, field.alias))
        else:
            if missing:
                if field.default_factory is not None:
                    val = field.default_factory()
                else:
                    val = field.default
            val, err = field.validate(val, {}, loc=(loc, field.alias))
        if err:
            request.errors.add(loc, field.alias, translate_err(request.localizer, err))
        else:
            if not missing:
                data[field.name] = request.validated[loc][field.name] = val
            elif field.default:
                data[field.name] = request.validated[loc][field.name] = field.default
            elif field.default_factory:
                data[field.name] = request.validated[loc][
                    field.name
                ] = field.default_factory()
    return data


def build_pydantic_resource(request, schema, **kwargs):
    data = validate_pydantic_schema(request, schema)
    if request.errors:
        return None
    return schema(**data)
