from typing import Any, Generic, Mapping, MutableMapping, Sequence, TypeVar, Union

from pydantic import BaseModel, Field, validator
from pydantic.generics import GenericModel
from typing_extensions import NotRequired, TypedDict

IntStr = Union[int, str]


T_contra = TypeVar("T_contra", bound=BaseModel, contravariant=True)


class ResponseSchema(GenericModel, Generic[T_contra]):
    status_code: int = Field(default=200)
    headers: MutableMapping[str, IntStr] = Field(default_factory=dict)
    body: T_contra = Field(...)


class ResponseTypedDict(TypedDict):
    status_code: NotRequired[int]
    headers: NotRequired[MutableMapping[str, IntStr]]
    body: Mapping[str, Any]


class NoContentResponseSchema(ResponseSchema[BaseModel]):
    status_code: int = Field(default=204)
    body: str = Field("")

    @validator("body")
    def validate_body(cls, value: Any, values: Any) -> str:
        # we may log a warning here
        return ""


class CollectionResponseSchema(ResponseSchema[T_contra], Generic[T_contra]):
    body: Sequence[T_contra] = Field(default_factory=list)

    @validator("headers")
    def validate_x_total_count(cls, value: Any, values: Any):
        if "X-Total-Count" not in value:
            raise ValueError("Missing header X-Total-Count")
        try:
            int(value["X-Total-Count"])
        except ValueError:
            raise ValueError("X-Total-Count is not a number")
        return value
