from urllib.parse import urlparse

from pyramid.exceptions import PredicateMismatch
from pyramid.interfaces import IRequestFactory, IRoutePregenerator, IRoutesMapper
from pyramid.request import Request as BaseRequest
from pyramid.urldispatch import RoutesMapper
from zope.interface import implementer


class Request(BaseRequest):
    @property
    def client_addr(self):
        """
        The effective client IP address as a string:

        Uses, in order:
        - HTTP HEADER X-Real-IP if set (set by Gandi proxies)

        - Pyramid implementation which is currently:
          - First value of the X-Forwarded-For HTTP header
          - Remote address of the TCP connection

        """
        real_ip = self.environ.get("HTTP_X_REAL_IP")
        if real_ip:
            return real_ip

        return super().client_addr


@implementer(IRoutesMapper)
class HateoasRoutesMapper(RoutesMapper):
    """Override pyramid default routes mapper in order to add some url tweak,
    due to some Gandi convention.

    This make the pyramid ``route_url`` take into consideration the http header
    'Hateoa-Prefix' to let the client overrides its url.

    Something like::

        GET /hosting/v5/hosting/some/random/url
        Host: pivoine

    Which would translate like::

        > GET /v5/some/random/url
        > Host: api.sequoia-hosting.svc.0x35.net
        > Hateoas-Prefix: https://pivoine/hosting/
        < 200 OK
        < Content-Type: application/json
        <
        < {
        <   "blah_url": "https://pivoine/hosting/v5/some/random/url
        < }


    If no Hateoas-Prefix is given, then the setting
    'sequoia.hosts.{route_name}' can be used to override the generated hostname
    and sequoia.hosts.{route_name}.scheme to override the generated scheme.

    Otherwise defaults value come from `HTTP/1.1` headers `Host` and
    `X-Forwarded-Proto` headers.

    """

    def get_route(self, name):
        route = super(HateoasRoutesMapper, self).get_route(name)
        if not route:
            return None

        if route.pregenerator:  # The route is already tweaked
            return route
        else:
            return TweakedRoute(name, route)


class TweakedRoute(object):
    """A Route that adapt hostname, scheme and path prefix depending the
    situation"""

    def __init__(self, name, route):
        self.route = route
        self.pregenerator = HateosPrefix(name)

    def generate(self, kw):
        path = self.route.generate(kw)

        if path.startswith(self.pregenerator.drop_prefix):
            path = path[len(self.pregenerator.drop_prefix):]  # fmt: skip

        return self.pregenerator.path_prefix.rstrip("/") + path


@implementer(IRoutePregenerator)
class HateosPrefix(object):
    """
    Generate the {scheme}://{host}:{port}/ part of the url and
    save a path prefix if necessary.
    The ``path_prefix`` is used by the :class:`.TweakedRoute` while generating
    the route_path.
    """

    def __init__(self, route_name):
        self.route_name = route_name
        self.path_prefix = ""
        self.drop_prefix = ""

    def __call__(self, request, elements, kw):
        if "Hateoas-Prefix" in request.headers:
            self.drop_prefix = request.headers.get("Hateoas-Replace", "")

            # As there is many places where url is sanitized,
            # we loose some separators due to `drop_prefix` containing slashes.
            # this ensure we still keep a slash separator whatever the
            # parameters are.
            self.drop_prefix = self.drop_prefix.rstrip("/")

            # Override hostname using the request
            url = urlparse(request.headers["Hateoas-Prefix"])
            kw["_scheme"] = url.scheme
            kw["_host"] = url.netloc

            self.path_prefix = url.path

        else:
            settings = request.registry.settings
            key = "sequoia.hosts.%s" % self.route_name

            if key in settings:
                # Override hostname using the registry
                kw["_host"] = settings[key]

                if "%s.scheme" % key in settings:
                    kw["_scheme"] = settings["%s.scheme" % key]

        if "_host" not in kw:
            if "Host" in request.headers:
                # If not host given or no host compute, use the HTTP/1.1 header
                kw["_host"] = request.headers["Host"]
            else:
                kw["_host"] = request.environ.get("SERVER_NAME", "localhost")
                if "SERVER_PORT" in request.environ:
                    kw["_port"] = request.environ["SERVER_PORT"]

        x_forwarded_proto = request.headers.get("X-Forwarded-Proto", "")
        if "_scheme" not in kw and x_forwarded_proto in ("http", "https"):
            # for the scheme
            kw["_scheme"] = request.headers["X-Forwarded-Proto"]

        if "_port" not in kw and "_host" in kw and ":" in kw["_host"]:
            kw["_host"], kw["_port"] = kw["_host"].rsplit(":", 1)

        return elements, kw


@implementer(IRequestFactory)
def request_factory(environ):
    """
    Try working around clients encoding their requests in latin1,
    otherwise request will systematically crash with an internal server error
    """
    request = Request(environ)

    try:
        request.path  # noqa
    except UnicodeDecodeError:
        pathinfo = environ["PATH_INFO"].decode("latin1").encode("utf-8")
        environ["PATH_INFO"] = pathinfo
        request = Request(environ)

    return request


request_factory.blank = lambda path: Request.blank(path)


def http_cache_factory(handler, registry):
    """A Pyramid tween view that enforce HTTP/1.0 and HTTP/1.1 caching
    headers. Ensure nothing is cached by default.

    Every view can use the `http_cache` attribute to overrice caching headers.

    if `http_cache` is not present, or if an exception is raised, `http_cache`
    is not applyed to the response, and default caching header are used.
    """

    def http_cache_handler(request):
        response = handler(request)

        if str("Cache-Control") not in response.headers:
            # There is no cache directive, ensure no one cache response
            response.cache_expires(0)

        elif request.context.http_vary:
            # There is cache directive, we may add a vary header
            http_vary = request.context.http_vary
            if isinstance(http_vary, (list, tuple)):
                http_vary = ", ".join(http_vary)
            response.headers[str("Vary")] = str(http_vary)

        return response

    return http_cache_handler


class HttpVaryPredicate(object):
    """
    Store the configuration to the request for views predicate `http_vary`
    to the @view_config decorator.

    """

    def __init__(self, http_vary, config):
        self.http_vary = http_vary

    def text(self):
        return "http_vary = %s" % (self.http_vary,)

    phash = text

    def __call__(self, context, request):
        context.http_vary = self.http_vary
        return True


class DocumentationPredicate(object):
    """
    Stores a documentation argument on a view's definition
    """

    def __init__(self, schema, *args, **kwargs):
        self.schema = schema

    def text(self):
        return "documentation = {!r}".format(self.schema)

    phash = text

    def __call__(self, context, request):
        return True


class AcceptPredicate(object):
    def __init__(self, val, *args, **kwargs):
        self.val = val
        self.priority = kwargs.get("priority", "application/json")

    def text(self):
        return "documentation = {!r}".format(self.val)

    phash = text

    def __call__(self, context, request):
        if not request.accept:
            raise PredicateMismatch(
                "This view require a accept header "
                "in the incomming request in order to match"
            )

        # Try to get an offer from received accept header
        offers = [self.priority, self.val]
        acceptable = request.accept.acceptable_offers(offers)
        if not acceptable:
            return False

        request.acceptable_offer = acceptable[0][0]

        return request.acceptable_offer == self.val


class ExtractCStructPredicate:
    def __init__(self, func, config):
        pass

    def text(self):
        return "predicate used by bound_schema_validator"

    phash = text

    def __call__(self, context, request):
        return True
