"""Render python object to json."""

import json
import logging
import warnings

import colander
from pyramid.httpexceptions import HTTPInternalServerError, HTTPNoContent

from sequoia.base3.helpers.colander import ListReturn
from sequoia.base3.helpers.i18n import gettext as _

from .encoder import JSONEncoder, set_content_type

log = logging.getLogger(__name__)


def _is_overriding(subschema, schema):
    """Check if a schema has a specific deserialization method"""
    return subschema.__class__.deserialize is not schema.deserialize


class ResponseBuilder:
    """
    Build a response dictionary that validate the view response dictionary-like
    object
    """

    def __init__(self, schema):
        self.schema = schema
        self.errors = []
        self._raise = False

    def fill_errors(self, exc, error_prefix):
        if self._raise:
            raise exc
        for key, val in exc.asdict().items():
            self.errors.append((error_prefix + key, val))

    def wrap_location(self, data):
        filldict = {"status_code": 200, "body": {}, "header": {}}
        for attr in self.schema.children:
            location = getattr(attr, "location", "body")
            name = attr.name
            if name not in data:
                continue
            value = data[name]

            if location == "header":
                # ensure header is accepted by WebOb
                # header must be string or a unicode on python 3
                value = str(value)

            if location == "body" and isinstance(self.schema, ListReturn):
                # XXX in that case,
                # the whole content is a list
                # this is the case for collection_get returns
                filldict[location] = value
                continue

            if location == "status_code":
                filldict[location] = value
                continue

            if location == "header" and name == "content_location":
                warnings.warn(
                    "Deprecated Use the name='Content-Location'", DeprecationWarning
                )
                name = "Content-Location"

            filldict[location][name] = value
        return filldict

    def build_mappingschema(self, data, schema, error_prefix):
        # log.debug('Building {} with {}'.format(data,
        #                                        schema.__class__.__name__))
        result = {}
        for attr in schema.children:
            if attr.name in data:
                value = self._validate(
                    data[attr.name],
                    attr,
                    error_prefix="{}{}.".format(error_prefix, attr.name),
                )
                if value == colander.drop:
                    if attr.default not in (colander.null, colander.drop):
                        warnings.warn(
                            "Put default value in the missing value", DeprecationWarning
                        )
                        value = attr.default
                    else:
                        continue
                result[attr.name] = value
            else:
                if attr.missing not in (
                    colander.required,
                    colander.null,
                    colander.drop,
                ):
                    # attr.missing contains the default value
                    result[attr.name] = attr.missing

                if attr.missing is colander.null:
                    warnings.warn(
                        "Put None instead of colander.null for " "missing attribute",
                        DeprecationWarning,
                    )
                    result[attr.name] = None
                    continue

                if attr.missing is colander.required:
                    # expect to be in the dict
                    if attr.default not in (colander.null, colander.drop):
                        warnings.warn(
                            "Put the default value in the "
                            "missing argument instead. "
                            "It SHOULD RAISE an ERROR",
                            DeprecationWarning,
                        )
                        result[attr.name] = attr.default
                        continue
                    if self._raise:
                        raise colander.Invalid(attr, _("Required"))
                    self.errors.append((error_prefix + attr.name, _("Required")))

                elif attr.missing is colander.drop:
                    if attr.default not in (colander.null, colander.drop):
                        warnings.warn(
                            "Put your default value in the " "missing value",
                            DeprecationWarning,
                        )
                        result[attr.name] = attr.default
                        continue

        if schema.validator or _is_overriding(schema, colander.MappingSchema):
            # Revalidate to play constraints on the sequence
            # itself, unused sql relationship already dropped
            try:
                result = schema.deserialize(result)
            except colander.Invalid as exc:
                self.fill_errors(exc, error_prefix)
                result = {}

        return result

    def build_sequenceschema(self, data, attr, error_prefix):
        # log.debug('Building {}{} with {}'.format(error_prefix, attr.name,
        #                                         data))
        if data is None and attr.missing == colander.drop:
            return colander.drop

        result = []
        schema = attr.children[0]
        for idx, item in enumerate(data):
            try:
                self._raise = True
                value = self._validate(item, schema, error_prefix=error_prefix)
            except colander.Invalid as exc:
                self._raise = False
                self.fill_errors(exc, error_prefix + "{}.".format(idx))
            else:
                self._raise = False
                if value == colander.drop:
                    continue
                result.append(value)

        if attr.validator or _is_overriding(attr, colander.SequenceSchema):
            # Revalidate to play constraints on the sequence
            # itself, unused sql relationship already dropped
            try:
                result = attr.deserialize(result)
            except colander.Invalid as exc:
                self.fill_errors(exc, error_prefix)
                result = []

        return result

    def _is_json_type(self, attr):
        # the list should contain colander.Date, colander.DateTime
        # but we cannot add them without breaking everything.
        # datetime is actually used in two distinct way:
        # our .helpers.colander.{DateTime, Date}
        # which receive datetime.datetime python object and are ignored
        # in the serialization
        # and colander.DateTime
        # which receive strings from restela and are serialized to datetime
        # then we let the json renderer cast them back to string...

        return not isinstance(
            attr.typ,
            # Those types are not serializable in json,
            # they have to be serialized to string during
            # the deserialization of the appscruct to the cstruct (json)
            (
                colander.Enum,
                # TODO: Add colander.{DateTime, Date} for the breaking change
            ),
        )

    def _build_simpletype(self, data, attr, error_prefix, meth):
        # log.debug('Building {}{} with {}'.format(error_prefix, attr.name,
        #                                          data))
        try:
            if data is None:
                # deserialize will generate the correct type for json
                # but, null values must be `colander.null` to be treated
                # properly, otherwise `None` may be casted to `'None'`
                data = colander.null

            return getattr(attr, meth)(data)
        except colander.Invalid as exc:
            log.info("Invalid result build, dropping")
            if self._raise:
                raise
            self.fill_errors(exc, error_prefix)
            return colander.drop

    def build_jsontype(self, data, attr, error_prefix):
        """
        Serialize data using the attr schema.

        data is a json simple type.
        """
        return self._build_simpletype(data, attr, error_prefix, "deserialize")

    def build_nonjsontype(self, data, attr, error_prefix):
        """
        Serialize data using the attr schema.

        data is a simple type, but not a json type.
        """
        if data is None:
            return attr.missing
        return self._build_simpletype(data, attr, error_prefix, "serialize")

    def _validate(self, data, schema, error_prefix=""):
        if not schema.children:
            # deserializing a simple type here in a sequence schema
            if self._is_json_type(schema):
                return self.build_jsontype(data, schema, error_prefix)
            else:
                return self.build_nonjsontype(data, schema, error_prefix)

        if isinstance(schema, colander.MappingSchema):
            return self.build_mappingschema(data, schema, error_prefix)

        if isinstance(schema, colander.SequenceSchema):
            return self.build_sequenceschema(data, schema, error_prefix)

        log.warning("Cannot serialize the given data because of the schema")
        self.errors.append((error_prefix + schema.name, _("Invalid Schema")))
        return colander.drop

    def build(self, data):
        data = self._validate(data, self.schema)
        if data is colander.drop:
            data = {}
        return self.wrap_location(data)


class JsonRenderer:
    """
    Template Factory for render json that accept datetime.
    """

    def __init__(self, _renderer_helper):
        pass

    def __call__(self, data, context):
        request = context["request"]
        if request.context.return_schema:
            schema = request.context.return_schema().bind(request=request)
            builder = ResponseBuilder(schema)
            resp = builder.build(data)

            if builder.errors:
                try:
                    log.error(
                        "Return Schema {0}.{1} does not "
                        "validate on {2} {3}:\n{4}"
                        "".format(
                            schema.__class__.__module__,
                            schema.__class__.__name__,
                            request.method,
                            request.current_route_url(),
                            "\n".join(
                                [
                                    "{}:{}".format(err[0], err[1])
                                    for err in builder.errors
                                ]
                            ),
                        )
                    )
                except Exception:
                    log.exception(
                        "Unexpected exception while logging error "
                        "in return schema validation"
                    )
                raise HTTPInternalServerError()
        else:
            log.warning(
                "Request {} has no return schema" "".format(request.matched_route.name)
            )
            resp = {"status_code": 200, "body": data, "header": {}}

        response = context["request"].response
        response.status_code = resp["status_code"]

        for key, val in resp["header"].items():
            response.headers[key.title()] = val

        if response.status_code == HTTPNoContent.code:
            response.content_type = None
            if resp["body"]:
                log.warning(
                    "A body as been build but cannot be sent because "
                    "of the HTTPNoContent status code"
                )
            return str("")

        set_content_type(request)
        return json.dumps(resp["body"], cls=JSONEncoder)
