import json
import logging
import re
import uuid
from datetime import datetime

try:
    # zope.sqlalchemy < 1.2
    from zope.sqlalchemy import ZopeTransactionExtension

    zqsla_register = None
except ImportError:
    from zope.sqlalchemy import register as zqsla_register

from sqlalchemy import DateTime, String, engine_from_config
from sqlalchemy.dialects.postgresql import ARRAY as PgARRAY
from sqlalchemy.dialects.postgresql import INET as PgINET
from sqlalchemy.dialects.postgresql import JSON as PgJSON
from sqlalchemy.dialects.postgresql import JSONB as PgJSONB
from sqlalchemy.dialects.postgresql import UUID as PgUUID
from sqlalchemy.dialects.postgresql.base import ischema_names as pg_ischema_names
from sqlalchemy.dialects.sqlite.base import ischema_names as sqlite_ischema_names
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.attributes import QueryableAttribute
from sqlalchemy.sql.expression import FunctionElement, func
from sqlalchemy.types import (
    CHAR,
    Concatenable,
    TypeDecorator,
    UnicodeText,
    UserDefinedType,
)

# Diconnect Handling
# See: http://docs.sqlalchemy.org/en/latest/core/pooling.html


try:
    from _mysql_exceptions import OperationalError as MySQLOperationalError  # noqa?
except ImportError:

    class MySQLOperationalError(Exception):
        """Dummy drop-in Exception in case MySQL driver is not installed"""


try:
    from uwsgidecorators import postfork as uwsgi_postfork
except ImportError:

    def uwsgi_postfork(*_):
        """Dummy drop-in"""


from pyramid.settings import asbool

try:
    from zipkin.binding.sqlalchemy import bind as bind_sqla_zipkin
    from zipkin.models import Endpoint
except ImportError:
    bind_sqla_zipkin = None


log = logging.getLogger(__name__)


class ModelError(Exception):
    def __init__(self, errors):
        if isinstance(errors, str):
            errors = [errors]
        super(ModelError, self).__init__("\n".join(errors))
        self.errors = errors


class utcnow(FunctionElement):
    type = DateTime()


@compiles(utcnow, "postgresql")
def pg_utcnow(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(utcnow, "sqlite")
def sqlite_utcnow(element, compiler, **kw):
    return "current_timestamp"


@compiles(utcnow, "mysql")
def mysql_utcnow(element, compiler, **kw):
    return "utc_timestamp()"


class NaiveUTCDateTime(TypeDecorator):
    """
    A DateTime type for storing TIMESTAMP WITHOUT TIMEZONE PostgreSQL dates

    Incoming non-naive datetime objects are converted to UTC,
    then stripped of their tzinfo value,

    """

    impl = DateTime

    # bind/result is idempotent for this type
    cache_ok = True

    def __init__(self):
        TypeDecorator.__init__(self)

    def process_bind_param(self, value, dialect):
        if isinstance(value, datetime) and value.tzinfo is not None:
            return (value - value.utcoffset()).replace(tzinfo=None)
        return value

    def compare_values(self, a, b):
        if isinstance(a, datetime) and a.tzinfo is not None:
            a = (a - a.utcoffset()).replace(tzinfo=None)
        if isinstance(b, datetime) and b.tzinfo is not None:
            b = (b - b.utcoffset()).replace(tzinfo=None)
        return a == b

    def process_result_value(self, value, dialect):
        return value


class UUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """

    impl = CHAR

    # https://docs.sqlalchemy.org/en/14/core/custom_types.html
    cache_ok = True

    def __init__(self, as_uuid=False):
        """
        When `as_uuid` is True, attributes will be transformed to python UUID
        when fetched from database.
        Otherwise they will be returned as str
        """
        super(UUID, self).__init__()
        self.as_uuid = as_uuid

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(PgUUID(as_uuid=self.as_uuid))
        elif dialect.name == "mysql":
            # could have been BINARY(16) but it's not readable
            return dialect.type_descriptor(CHAR(36))
        else:
            return dialect.type_descriptor(CHAR(36))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value

        if isinstance(value, str):
            value = uuid.UUID(value)
        elif not isinstance(value, uuid.UUID):
            raise ValueError("value %s is not a valid uuid.UUId" % value)

        # mysql should be value.bytes if BINARY
        return str(value)

    def process_result_value(self, value, dialect):
        # psql will directly return UUIDs if as_uuid is True
        # mysql/sqlite will return strings, need to cast them
        if self.as_uuid and value is not None and not isinstance(value, uuid.UUID):
            return uuid.UUID(value)
        return value


class INET(TypeDecorator):
    """Platform-independent INET type.

    Uses Postgresql's INET type, otherwise uses
    CHAR(45).

    """

    impl = CHAR

    # bind/result is idempotent for this type
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(PgINET())
        else:
            return dialect.type_descriptor(String(45))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        return value

    def process_result_value(self, value, dialect):
        return value


class BaseCIText(Concatenable, UserDefinedType):
    # bind/result is idempotent for this type
    cache_ok = True

    def bind_processor(self, dialect):
        def process(value):
            return value

        return process

    def result_processor(self, dialect, coltype):
        def process(value):
            return value


class PgCIText(BaseCIText):
    def get_col_spec(self):
        return "CITEXT"


class SqliteCIText(BaseCIText):
    def get_col_spec(self):
        return "TEXT COLLATE NOCASE"


# Register CIText to SQLAlchemy's reflection subsystem.
pg_ischema_names["citext"] = PgCIText
sqlite_ischema_names["citext"] = SqliteCIText


class CIText(TypeDecorator):
    """Platform-independent Case Insensitive type.

    Uses Postgresql's CITEXT type, otherwise lower case every values.

    """

    impl = CHAR

    # bind/result is idempotent for this type
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(PgCIText())
        elif dialect.name == "sqlite":
            return dialect.type_descriptor(SqliteCIText())
        else:
            return dialect.type_descriptor(UnicodeText())

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name in ("postgresql", "sqlite"):
            return value
        return value.lower()

    def process_result_value(self, value, dialect):
        if value is None:
            return value

        if isinstance(value, bytes):
            value = value.decode("utf-8")
        return value


class BaseComplexType(TypeDecorator):
    _pg_type = None
    _pg_type_kwargs = {}
    impl = UnicodeText

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return self._pg_type(**self._pg_type_kwargs)
        else:
            return dialect.type_descriptor(UnicodeText())

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if dialect.name == "postgresql":
            return value
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        if not value:
            return None
        if dialect.name == "postgresql":
            return value
        return json.loads(value)

    """
    Found in documentation as recommended but it seems that it worked
    good so far without it. I don't want to change too much behaviour
    @ once.
    http://docs.sqlalchemy.org/en/latest/core/custom_types.html#sqlalchemy.types.TypeDecorator
    def coerce_compared_value(self, op, value):
        return self.impl.coerce_compared_value(op, value)

    """


class ARRAY(BaseComplexType):
    _pg_type = PgARRAY
    _pg_type_kwargs = {"item_type": CIText}  # To Be [re]defined in subclass


class JSONB(BaseComplexType):
    """Platform-independent JSON type.

    Uses Postgresql's JSONB type, otherwise use text.
    """

    _pg_type = PgJSONB
    _pg_type_kwargs = {"none_as_null": True}


class JSON(JSONB):
    """Platform-independent JSON type.

    Uses Postgresql's JSON type, otherwise use text.

    If you are using posgres 9.4+, you should consider using jsonb instead.

    The json type of postgresql is a text type with json function, and to query
    in it you are going to cast it to jsonb.

    When to use json:

     If you want to keep the input formating, such as whitespace and cariage
     return, or if you want to keep duplicate keys, which is json valid,
     but often not desired. jsonb keep the last duplicate value:

     +------------------------+------------+
     | json                   | jsonb      |
     +------------------------+------------+
     | `{ "a":1,\n   "a": 2}` | `{"a": 2}` |
     +------------------------+------------+

    """

    _pg_type = PgJSON


class _Base(object):
    __database__ = None  # XXX set in the "create_engine"

    @declared_attr
    def __tablename__(cls):
        # CamelCase to underscore cast
        s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", cls.__name__)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()

    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8"}

    @property
    def session(self):
        return SessionFactory.get_object_session(self)

    @classmethod
    def by_id(cls, id):
        return cls.first(where=(cls.id == id,))

    @classmethod
    def find(cls, **kwargs):
        """
        See method `build_query` for list of supported kwargs.

        count may be:
         True (will count rows, *including joins*)
         False (returns rows)
         a column to count (Model.column or Model.column.distinct())

        Warning: SQLA redefine boolean evaluation to build a SQL expression
                 DO NOT test 'count' directly because of this
        """
        count = kwargs.get("count")
        bcount = count is not False and count is not None
        if bcount:
            # XXX posgresql doesn't like it
            kwargs["offset"] = None
            kwargs["limit"] = None
            kwargs["order_by"] = None
        qry = cls.build_query(**kwargs)
        return qry.scalar() if bcount else qry.all()

    @classmethod
    def first(cls, *args, **kwargs):
        return cls.build_query(*args, **kwargs).first()

    @classmethod
    def all(cls, *args, page_size=1000, **kwargs):
        """
        See method `build_query` for list of supported kwargs.
        """
        session = SessionFactory.get(cls.__database__)()
        kwargs["offset"] = kwargs.get("offset", 0)
        kwargs["order_by"] = kwargs.get("order_by", cls.id)
        kwargs["limit"] = page_size
        while True:
            page = cls.find(*args, **kwargs)
            for m in page:
                yield m
            session.flush()
            if len(page) != page_size:
                return
            kwargs["offset"] += page_size

    @classmethod
    def exists(cls, *args, **kwargs):
        """
        See method `build_query` for list of supported kwargs.
        """
        session = SessionFactory.get(cls.__database__)()
        query = cls.build_query(*args, **kwargs)
        return session.query(query.exists()).scalar()

    @classmethod
    def build_order_by(cls, order_by):
        sa_order_by = []
        if not isinstance(order_by, (list, tuple)):
            order_by = [order_by]
        for field in order_by:
            if isinstance(field, str):
                desc_ = field.startswith("-")
                if desc_:
                    field = field[1:]

                # XXX Do we create own exception here ?
                attr = getattr(cls, field)  # raise AttributeError
                if not isinstance(attr, QueryableAttribute):
                    raise ValueError("%s is not a sortable field of %s" % (field, cls))
                field = attr.desc() if desc_ else attr
            sa_order_by.append(field)
        return sa_order_by

    @classmethod
    def build_query(
        cls,
        join=None,
        outerjoin=None,
        where=None,
        order_by=None,
        limit=None,
        offset=None,
        count=False,
        select=None,
    ):
        """
        Build a SQLAlchemy query using the given parameters.

        :param join: Build one or many `INNER JOIN` statement.
            It can be:

            * a model:
                SQLAlchemy will use Foreign Key if it is possible,
                otherwise it will fail.
            * a list of model:
                Like for a model, but will build many joins.
            * a list of (model, join condition):
                Use this syntax if you want to build a custom join condition,
                you must use this syntax in case SQLAlchemy can't build the
                join condition by Model reflexion.

        :param outerjoin: Build one or many `OUT JOIN` statement.
            Follow same rules as param 'join'.

        :param where: a list of SQLAlchemy query `filter`

        :param order_by: Build the `ORDER BY` SQL clause.
            It can be:

              * a string:
                In that case it has to be an attribute of the model, prefixed
                by `-` to build a `DESC` clause.

              * a list string:
                like a string, but in a list creating many columns in the
                `ORDER BY` clause

              * a SQLAlchemy field:
                Use it to sort by a field coming from a join condition.
                In that case could be `field.desc()`

              * a list of SQLAlchemy field:
                behave like SQLAlchemy field, but in a list creating many
                columns in the `ORDER BY` clause

        :param limit: `int` passed to `query.limit` condition for pagination
        :param offset: `int` passed to `query.offset` condition for pagination
        :param count: Generate a `count(*)` instead of a query
            Parameter is ignored, in case a `select` parameter is set.
        :param select: `list` of element passed to `session.query`,
            by default, the result return `cls` instances.
            Count is ignored, in case this parameter is set.
        """
        session = SessionFactory.get(cls.__database__)()
        if select:
            query = session.query(*select).select_from(cls)
        elif count is not False and count is not None:
            query = session.query(func.count(count)).select_from(cls)
        else:
            query = session.query(cls)

        if join:
            if isinstance(join, (list, tuple)):
                for j in join:
                    if isinstance(j, (list, tuple)):
                        query = query.join(*j)
                    else:
                        query = query.join(j)
            else:
                query = query.join(join)

        if outerjoin:
            if isinstance(outerjoin, (list, tuple)):
                for oj in outerjoin:
                    if isinstance(oj, (list, tuple)):
                        query = query.outerjoin(*oj)
                    else:
                        query = query.outerjoin(oj)
            else:
                query = query.outerjoin(outerjoin)

        if where is not None:
            for filter in where:
                query = query.filter(filter)

        if order_by is not None:
            order_by = cls.build_order_by(order_by)
            query = query.order_by(*order_by)
        if limit:
            query = query.limit(limit)
        if offset:
            query = query.offset(offset)

        return query

    def validate(self, session):
        """
        return True or raise a :class:`ModelError` Exception
        """
        return True

    def delete(self):
        self.session.delete(self)

    def __repr__(self):
        """String representation of a model."""
        cls = self.__class__
        repr_ = "%s.%s" % (cls.__module__, cls.__name__)

        if hasattr(self, "id"):
            repr_ += " #%s" % self.id

        return "<%s>" % repr_

    __str__ = __repr__

    def __iter__(self):
        attrs = sorted(dir(self))
        for attr in attrs:
            if attr.startswith("_"):
                continue
            val = getattr(self, attr)
            if val is None:
                continue
            yield attr, val

    def pop(self, name, *args):
        """
        Trick colander dict validator (which is using dict.pop(),
        values aren't actually removed (because they can be properties)

        The pop is useless in our case (unknown key is always set to ignore)

        """

        return getattr(self, name, *args)

    def items(self):
        return dict(self).items()

    def __contains__(self, item):
        return hasattr(self, item)

    def __getitem__(self, item):
        return getattr(self, item)


class Database(object):
    databases = {}

    @classmethod
    def register(cls, name):
        if name not in cls.databases:
            cls.databases[name] = declarative_base(cls=_Base)
            cls.databases[name].__database__ = name
        return cls.databases[name]

    @classmethod
    def get(cls, name):
        return cls.databases[name]

    @classmethod
    def close_all(cls):
        while cls.databases:
            database = cls.databases.pop(list(cls.databases.keys())[0])
            if database.metadata.bind:
                database.metadata.bind.dispose()


class SessionFactory(object):
    sessions = {}

    @classmethod
    def register(cls, name, keep_session=False):
        if not zqsla_register:
            extension = ZopeTransactionExtension(keep_session=keep_session)
            session = scoped_session(sessionmaker(extension=extension))
        else:
            session = scoped_session(sessionmaker())
            zqsla_register(session, keep_session=keep_session)

        cls.sessions[name] = session
        return cls.sessions[name]

    @classmethod
    def get(cls, name):
        return cls.sessions[name]

    @classmethod
    def get_object_session(cls, object_):
        try:
            return cls.sessions[object_.__database__].object_session(object_)
        except KeyError:
            return None


def register_pg_array_types(engine):
    """

    Register array types for deserializations.

    Currently it only register a serialization for the type `citext[]`
    which is currently not register by default in the psycopg2 driver.

    """

    if engine.dialect.name != "postgresql":
        return

    # The psycopg2 must be importable to get a postgres connection
    import psycopg2

    try:
        # Retrieve oids of citext arrays.
        oids = engine.execute("SELECT typarray FROM pg_type WHERE typname = 'citext'")

        oids = tuple(row[0] for row in oids)
        array_type = psycopg2.extensions.new_array_type(
            oids, "citext[]", psycopg2.STRING
        )
        psycopg2.extensions.register_type(array_type, None)
    except ProgrammingError:
        # citext is not available on the database.
        log.error("Unsupported type citext on %s", engine.url)


def create_engine(db_name, settings, prefix="sql."):
    """

    Create a SQLAlchemy database engine using the engine_from_config
    function, it concatenate prefix and dn_name to find has the
    engine_from_config prefix. This is useful if you have to connect
    to many SQL Server.

    If it is not a SQLAlchemy in memory ('sqlite://') it create
    a pool recycle and tests connections.

    """

    url_key = "%s%s.url" % (prefix, db_name)
    if settings[url_key] == "sqlite://":
        # for in memory database, for tests, we don't ping and reconnect
        engine = engine_from_config(settings, "%s%s." % (prefix, db_name))
    else:
        # By default we create a database that recycle connection
        # and tests the connection, reconnect, in order to avoid
        # the "MySQL has gone aways" syndrome
        recycle_key = "%s%s.pool_recycle" % (prefix, db_name)
        settings = settings.copy()  # Safe modification
        settings.setdefault(recycle_key, 3600)

        engine = engine_from_config(
            settings, "%s%s." % (prefix, db_name), pool_pre_ping=True
        )

    if (
        bind_sqla_zipkin
        and settings.get("zipkin.collector")
        and asbool(settings.get("logs.zipkin", "True"))
    ):
        zipkin_name = "SQL-%s-%s" % (
            db_name,
            settings.get("zipkin.env", "dev"),
        )
        bind_sqla_zipkin(engine, Endpoint(zipkin_name))

    ckey = "%s.keep_session" % db_name
    keep_session = settings.get(ckey, False)

    DBSession = SessionFactory.register(db_name, keep_session=keep_session)
    DBSession.configure(bind=engine)

    register_pg_array_types(engine)
    Database.get(db_name).metadata.bind = engine

    @uwsgi_postfork
    def dispose_after_fork():
        """
        Discard SQLAlchemy engine after uWSGI fork, since a connection
        may already have been established and will otherwise be shared
        between uWSGI workers
        """
        log.debug("disposing of engine %s after uwsgi fork" % db_name)
        engine.dispose()

    return engine


def dispose_engine(db_name):
    Database.get(db_name).metadata.bind.dispose()
