import logging
from typing import List, Tuple

import pkg_resources
import sentry_sdk
from pyramid.settings import asbool, aslist
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.pyramid import PyramidIntegration
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration

log = logging.getLogger(__name__)


def initialize_sentry_sdk(
    dsn: str,
    environment: str,
    version: str,
    enable_logger: bool = False,
    integrations=[],
    tags: List[Tuple[str, str]] = [],
):
    """Initialize the Sentry SDK client with the given parameters."""

    # Add the logging integration if it is set to True
    if enable_logger:
        integrations.append(LoggingIntegration(level=logging.WARNING))

    # Setup the Sentry client
    sentry_sdk.init(
        dsn=dsn,
        environment=environment,
        release=version,
        integrations=integrations,
        # -- Sentry SDK options
        traces_sample_rate=0,  # Disable tracing/perf monitoring
        send_default_pii=True,  # Trying to associate users to errors
    )

    # Setup tags
    for tag in tags:
        sentry_sdk.set_tag(tag[0], tag[1])

    log.info("Sentry client is configured!")


def initialize_sentry_for_backend(conf: dict, backend_name: str):
    """
    Initialize a specific Sentry SDK client for Celery backends by using the
    provided dict and the package name of the backend.
    """

    from sentry_sdk.integrations.celery import CeleryIntegration

    sentry_enabled = conf.get("enabled", False)
    dsn = conf.get("dsn", None)
    environment = conf.get("environment", None)
    sentry_logger = conf.get("enable_logger", False)
    sentry_integrations = [CeleryIntegration(), SqlalchemyIntegration()]
    service_tags = [("service_type", "backend")]

    # Don't continue if Sentry client is not enabled in the conf
    if not conf or not sentry_enabled:
        return

    # Early return if the DSN is not present in the conf
    if not dsn:
        log.error(
            "Unable to enable Sentry, the DSN is missing in the " "configuration file."
        )
        return

    # Get the version of the backend's package
    backend_version = pkg_resources.get_distribution(backend_name).version

    initialize_sentry_sdk(
        dsn,
        environment,
        backend_version,
        sentry_logger,
        sentry_integrations,
        tags=service_tags,
    )


def initialize_sentry_for_pyramid_app(settings):
    """
    Initialize a specific Sentry SDK client for Pyramid applications by using
    the provided Pyramid registry settings.
    """

    sentry_enabled = asbool(settings.get("sentry_sdk.enabled", False))
    dsn = settings.get("sentry_sdk.dsn", None)
    environment = settings.get("sentry_sdk.environment", None)
    sentry_logger = asbool(settings.get("sentry_sdk.enable_logger", False))
    sentry_integrations = [PyramidIntegration(), SqlalchemyIntegration()]
    service_tags = [("service_type", "API")]

    # Don't continue if Sentry client is not enabled in the conf
    if not sentry_enabled:
        return

    # Early return if the DSN is not present in the conf
    if not dsn:
        log.error(
            "Unable to enable Sentry, the DSN is missing in the " "configuration file."
        )
        return

    # Get the version of the first API module set in `sequoia.export` conf key
    sequoia_export = aslist(settings.get("sequoia.export", ""))

    if len(sequoia_export) == 0:
        log.error(
            "Unable to enable Sentry, no API are specified in the "
            "`sequoia.export` configuration key."
        )
        return

    main_api = sequoia_export[0].strip()
    main_api_version = pkg_resources.get_distribution(main_api).version

    initialize_sentry_sdk(
        dsn,
        environment,
        main_api_version,
        sentry_logger,
        sentry_integrations,
        tags=service_tags,
    )


def includeme(config):
    """Initialize Sentry for the Pyramid application."""

    initialize_sentry_for_pyramid_app(config.registry.settings)
