import pprint
import unittest
from unittest.mock import patch

from webtest import TestApp

from ..errors import internal_server_error


class PrettyDiffPrinter(pprint.PrettyPrinter):
    """
    A pprint subclass to hide some differences
    that usually don't matter in tests
    """

    _original_pformat = None

    def _repr(self, object, context, level):
        if isinstance(object, bytes):
            # If object is an unicode string, but ASCII only,
            # don't represent it with the leading 'b'
            try:
                object = object.decode("utf-8")
            except UnicodeDecodeError:
                pass

        return pprint.PrettyPrinter._repr(self, object, context, level)

    def ppformat(self, object, *unused):
        # method  replacing the pprint.pformat static one during patch
        return self.pformat(object)

    def __enter__(self):
        self._original_pformat = pprint.pformat
        pprint.pformat = self.ppformat

    def __exit__(self, *unused):
        pprint.pformat = self._original_pformat


class ApiTestCase(unittest.TestCase):
    fixtures = ()
    mocks = []
    app_settings = {
        "sequoia.export": "",
    }
    extra_environ = {}

    _printer = PrettyDiffPrinter()

    @classmethod
    def setUpClass(cls):
        from sequoia.base3 import main

        app = main({}, **cls.app_settings)
        cls.app = TestApp(app, extra_environ=cls.extra_environ)
        cls._fixtures = [fixture() for fixture in cls.fixtures]
        [fixture.setUp() for fixture in cls._fixtures]

        # Disable catching of internal server error during tests
        internal_server_error.disabled = True

    @classmethod
    def tearDownClass(cls):
        [fixture.tearDown() for fixture in reversed(cls._fixtures)]
        internal_server_error.disabled = False

    def setUp(self):
        super(ApiTestCase, self).setUp()
        self.maxDiff = None

        self._mocks = [patch(*mock_args) for mock_args in self.mocks]
        for dummy in self._mocks:
            dummy.start()

    def tearDown(self):
        super(ApiTestCase, self).tearDown()

        for dummy in reversed(self._mocks):
            dummy.stop()

    def pp(self, *args, **kwargs):
        return pprint.pprint(*args, **kwargs)

    def assertDictEqual(self, *args, **kwargs):
        with self._printer:
            return super(ApiTestCase, self).assertDictEqual(*args, **kwargs)

    def assertSequenceEqual(self, *args, **kwargs):
        with self._printer:
            return super(ApiTestCase, self).assertSequenceEqual(*args, **kwargs)


class ModelTestCase(unittest.TestCase):
    mocks = []

    @property
    def session(self):
        raise NotImplementedError

    @classmethod
    def setUpClass(cls):
        cls._fixtures = [fixture() for fixture in cls.fixtures]
        [fixture.setUp() for fixture in cls._fixtures]

    @classmethod
    def tearDownClass(cls):
        [fixture.tearDown() for fixture in reversed(cls._fixtures)]

    def setUp(self):
        super(ModelTestCase, self).setUp()

        self._mocks = [patch(*mock_args) for mock_args in self.mocks]
        for dummy in self._mocks:
            dummy.start()

    def tearDown(self):
        super(ModelTestCase, self).tearDown()

        for dummy in reversed(self._mocks):
            dummy.stop()

    def assertIsModel(self, item, attempt_class, attempt_id=None):
        self.assertIsInstance(item, attempt_class)
        if attempt_id:
            self.assertEqual(item.id, attempt_id)
