"""Base fixtures for testings."""
import errno
import inspect
import json
import logging
import os.path
import uuid
from unittest import TestCase  # noqa

import factory
import transaction
import yaml
from sqlalchemy.inspection import inspect as sqla_inspect
from zope.sqlalchemy import datamanager

from ...helpers.sqla import SessionFactory

YAMLLoader = getattr(yaml, "CSafeLoader", yaml.SafeLoader)

logger = logging.getLogger(__name__)


class FixtureBase(object):
    def setUp(self):
        raise NotImplementedError

    def tearDown(self):
        pass


class FixtureTableJSon(FixtureBase):
    # Actually, YAML now

    model = None  # Must be override
    database = None
    table = None  # To use if no model
    json_fields = None

    @property
    def db_session(self):
        return SessionFactory.get(self.database)()

    def setUp(self):
        tablename = self.model.__tablename__ if self.model else self.table
        directory = os.path.dirname(inspect.getsourcefile(self.__class__))

        for ftype in ("yaml", "json"):
            # xx/json/toto.json or xx/yaml/toto.yaml
            path = os.path.join(directory, ftype, "%s.%s" % (tablename, ftype))

            if os.path.exists(path):
                try:
                    with open(path, "r") as yfile:
                        data = yaml.load(yfile, YAMLLoader)
                        break
                except ValueError:
                    logger.error("Error in file %s" % path)
                    raise
        else:
            logger.error("unable to find fixture file for %s", tablename)
            raise FileNotFoundError(
                errno.ENOENT, "No such file or directory: %s" % path
            )

        # Needed even when no data, otherwise crash some tests
        # because this triggers the table creation
        session = self.db_session

        # Empty JSON list, we don't have anything to do
        if not data:
            return

        [self._insert(session, record) for record in data]
        datamanager.mark_changed(session)
        transaction.commit()

    def _insert(self, session, record):
        if "mysql" in session.connection().engine.driver:
            quote_char = "`"
        else:
            quote_char = '"'

        def quote(name):
            """Quote table/column names"""
            return "%s%s%s" % (quote_char, name, quote_char)

        if self.model:
            # Retrieve SQL columns from the model, they might not be the
            # same as the model attribute names
            tablename = self.model.__tablename__
            names = set(col.name for col in self.model.__mapper__.columns.values())

            insert = {}
            for key, value in list(record.items()):
                if key not in names:
                    key = getattr(self.model, key).name
                insert[key] = value
            record = insert
        else:
            tablename = self.table
        columns = list(record.keys())

        if self.json_fields and "psycopg2" not in session.connection().engine.driver:
            insert = {}
            for key, value in list(record.items()):
                if key in self.json_fields:
                    value = json.dumps(value)
                insert[key] = value
            record = insert

        sql = "INSERT INTO %s " % quote(tablename)
        sql += "(%s)" % ", ".join(map(quote, columns))
        sql += " VALUES (%s);" % ", ".join([":%s" % v for v in columns])
        session.execute(sql, record)
        return sql

    def tearDown(self):
        tablename = self.model.__tablename__ if self.model else self.table
        session = self.db_session
        session.execute("DELETE FROM `%s`" % tablename)
        datamanager.mark_changed(session)
        transaction.commit()


class SequoiaBaseFactoryOptions(factory.base.FactoryOptions):
    def _build_default_options(self):
        base = super(SequoiaBaseFactoryOptions, self)._build_default_options()
        return base + [
            factory.base.OptionDefault("session_name", None, inherit=True),
        ]


class SequoiaBaseFactory(factory.base.Factory):
    _options_class = SequoiaBaseFactoryOptions

    class Meta:
        abstract = True

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        # Create object, flush and commit using temporary session
        with transaction.manager:
            session_name = "fixtures-%s" % uuid.uuid4()
            session = SessionFactory.register(session_name)
            obj = model_class(*args, **kwargs)
            session.add(obj)
            session.flush()
            state = sqla_inspect(obj)
            identity = state.identity
        del SessionFactory.sessions[session_name]

        session = SessionFactory.get(cls._meta.session_name)
        return session.query(model_class).get(identity)
