import json
import shutil
from io import BytesIO
from json.decoder import JSONDecodeError

import colander
import translationstring
from cornice.resource import resource, view
from pyramid.httpexceptions import HTTPBadRequest, HTTPConflict, HTTPNotFound
from pyramid.response import Response
from webob.multidict import MultiDict

from sequoia.base3.api import APIBase
from sequoia.base3.api.base import paginated_view
from sequoia.base3.context import DefaultContext
from sequoia.base3.errors import colander_error_handler
from sequoia.base3.helpers.colander import (
    CIDR,
    IP,
    UUID,
    DateTime,
    FilteredPaginationSchema,
    ListReturn,
    PaginationSchema,
)
from sequoia.base3.helpers.cornice import bound_schema_validator

_ = translationstring.TranslationStringFactory("dummy_factory_domain")


class DummyNotFound(HTTPNotFound):
    object_name = "dummy"


class PostDummiesSchema(colander.MappingSchema):
    @colander.instantiate(name="body")
    class Body(colander.MappingSchema):
        name = colander.SchemaNode(
            colander.String(), validator=colander.Length(3, 7), location="body"
        )


class GetDummiesSchema(colander.MappingSchema):
    @colander.instantiate(name="path")
    class Path(colander.MappingSchema):
        name = colander.SchemaNode(colander.String(), validator=colander.Length(3, 7))


class ReturnDummiesSchema(colander.MappingSchema):
    name = colander.SchemaNode(
        colander.String(), validator=colander.Length(3, 7), location="body"
    )


class PostDummiesSchemaReturn(colander.MappingSchema):
    href = colander.SchemaNode(colander.String(), location="body")
    internal_href = colander.SchemaNode(
        colander.String(), location="body", missing=colander.drop
    )


class TypesReturn(colander.MappingSchema):
    string = colander.SchemaNode(
        colander.String(), missing=colander.drop, location="body"
    )
    float = colander.SchemaNode(
        colander.Float(), missing=colander.drop, location="body"
    )
    int = colander.SchemaNode(
        colander.Integer(), missing=colander.drop, location="body"
    )
    boolean = colander.SchemaNode(
        colander.Boolean(), missing=colander.drop, location="body"
    )
    uuid = colander.SchemaNode(UUID(), missing=colander.drop, location="body")
    cidr = colander.SchemaNode(CIDR(), missing=colander.drop, location="body")
    ip = colander.SchemaNode(IP(), missing=colander.drop, location="body")
    datetime = colander.SchemaNode(DateTime(), missing=colander.drop, location="body")

    @colander.instantiate(name="list", missing=colander.drop)
    class List(colander.SequenceSchema):
        string = colander.SchemaNode(colander.String(), location="body")

    payload = colander.SchemaNode(
        colander.Mapping(unknown="preserve"), missing=colander.drop, location="body"
    )


class TypesParams(colander.MappingSchema):
    @colander.instantiate(name="path")
    class TypesParamsPath(colander.MappingSchema):
        type = colander.SchemaNode(colander.String(), missing=colander.drop)

    body = TypesReturn()


def UnprocessableEntity(node, cstruct):
    raise colander.Invalid(node, "Invalide")


class GetUnprocessableEntity(colander.MappingSchema):
    @colander.instantiate(name="path")
    class Path(colander.MappingSchema):
        name = colander.SchemaNode(
            colander.String(),
            validator=UnprocessableEntity,
        )


class PageFilterSchema(colander.MappingSchema):
    @colander.instantiate(name="querystring")
    class Querystring(FilteredPaginationSchema):
        @colander.instantiate(name="dummyfilter")
        class DummyFilter(colander.SequenceSchema):
            _ = colander.SchemaNode(colander.String())


class OtherPageListReturn(ListReturn):
    @colander.instantiate(name="items")
    class DummyReturn(colander.SequenceSchema):
        @colander.instantiate()
        class DummyReturn(colander.MappingSchema):
            name = colander.SchemaNode(colander.String())


class PostEmptyBodySchema(colander.MappingSchema):
    @colander.instantiate(name="body")
    class Body(colander.MappingSchema):
        name = colander.SchemaNode(colander.String(), missing=colander.drop)


@resource(
    path="/dummies/{name}",
    collection_path="/dummies",
    name="Dummy",
    factory=DefaultContext,
)
class DummyAPI(APIBase):
    @view(
        renderer="json",
        schema=PostDummiesSchema(),
        return_schema=PostDummiesSchemaReturn,
        validators=[
            bound_schema_validator,
        ],
    )
    def collection_post(self):
        params = self.request.validated["body"]
        if params["name"] == "barbie":
            raise HTTPBadRequest(
                explanation=_("Dummy ${name} already exists"), mapping=params
            )
        if params["name"] == "get_url":
            return {"href": self.get_url("Dummy", name=params["name"])}
        if params["name"] == "ken":
            raise HTTPConflict(
                errors=[
                    {
                        "name": "t-name",
                        "location": "t-location",
                        "description": "t-desc",
                    }
                ]
            )

        if params["name"] == "intern":
            return {
                "href": self.get_url("Dummy", name=params["name"]),
                "internal_href": self.get_internal_url("Dummy", name=params["name"]),
            }

        return {"href": self.get_url("Dummy", name=params["name"])}

    @view(
        renderer="json",
        schema=GetDummiesSchema(),
        return_schema=ReturnDummiesSchema,
        http_cache=60,
        http_vary="X-Variant",
        validators=[
            bound_schema_validator,
        ],
    )
    def get(self):
        params = self.request.validated["path"]
        if params["name"] == "boom":
            # -> internal server error
            raise RuntimeError("Sometime it happens")
        elif params["name"] == "bad":
            return {"name": "too-big-to-be-returned"}
        elif params["name"] == "400":
            self.request.errors = [
                {
                    "description": "That is what you asked for",
                    "location": "path",
                    "name": "name",
                }
            ]
            raise HTTPBadRequest(explanation="400 is not a valide name")
        elif params["name"] == "400-lcz":
            self.request.errors = [
                {
                    "description": _("This should be localized"),
                    "location": "path",
                    "name": "name",
                },
                {
                    "description": _(
                        "With ${parma1} too", mapping={"parma1": "mapping"}
                    ),
                    "location": "path",
                    "name": "name",
                },
            ]
            raise HTTPBadRequest(explanation="400 is not a valide name")
        elif params["name"] == "nope":
            raise DummyNotFound(explanation="${name} was not found", mapping=params)
        return {"name": params["name"]}


@resource(
    path="/unprocessable_entity/{name}",
    collection_path="/unprocessable_entity",
    name="UnprocessableEntity",
    factory=DefaultContext,
    error_handler=colander_error_handler,
)
class UnprocessableEntityAPI(APIBase):
    @view(
        renderer="json",
        schema=GetUnprocessableEntity(),
        http_cache=(60, {"private": True}),
        validators=[
            bound_schema_validator,
        ],
    )
    def get(self):
        raise RuntimeError("Should never get here")


@resource(
    path="/pages/{name}",
    collection_path="/pages",
    name="Pages",
    factory=DefaultContext,
)
class PageAPI(APIBase):
    @view(
        renderer="string",
        schema=GetDummiesSchema(),
        http_cache=(60, {"public": True}),
        http_vary=("X-Variant-1", "X-Variant-2"),
        validators=[
            bound_schema_validator,
        ],
    )
    def get(self):
        return "Name is: %(name)s" % self.request.validated["path"]

    @paginated_view(renderer="string", schema=PaginationSchema)
    def collection_get(self, offset, limit, sort_by):
        return "We have pages"


@resource(
    path="/otherpages/{name}",
    collection_path="/otherpages",
    name="OhterPages",
    factory=DefaultContext,
)
class OtherPageAPI(APIBase):
    @paginated_view(
        renderer="json",
        schema=PageFilterSchema,
        validators=[
            bound_schema_validator,
        ],
        return_schema=OtherPageListReturn,
    )
    def collection_get(self, offset, limit, sort_by):
        acceptable = ("application/json", "text/json", "text/plain")
        content_type = acceptable[0]
        content_types = self.request.accept.acceptable_offers(acceptable)
        if content_types:
            content_type = content_types[0][0]

        if content_type == "text/plain":
            return "We have pages"

        data = [{"name": "item %d" % i} for i in range(1, 11)]
        rdata = data[offset:limit + offset]  # fmt: skip

        return {"X-Total-Count": len(data), "items": rdata}


@resource(
    path="/protected",
    factory=DefaultContext,
    name="ProtectedResources",
)
class ProtectedResourceAPI(APIBase):
    @view(
        renderer="string",
        permission="dummy:protected:read",
        http_cache=(60, {"private": True}),
    )
    def get(self):
        return "Confidential"


@resource(path="/documented", factory=DefaultContext, name="DocumentedViews")
class DocumentedResourceAPI(APIBase):
    @view(renderer="string", documentation={"status_code": 201})
    def get(self):
        return "created"


@resource(
    path="/types/{type}",
    collection_path="/types",
    factory=DefaultContext,
    name="ReturnTypes",
)
class TypesAPI(APIBase):
    @view(
        renderer="json",
        schema=TypesParams,
        return_schema=TypesReturn,
        validators=[
            bound_schema_validator,
        ],
    )
    def collection_post(self):
        return self.request.validated["body"]

    @view(
        renderer="json",
        schema=TypesParams,
        return_schema=TypesReturn,
        validators=[
            bound_schema_validator,
        ],
    )
    def get(self):
        if self.request.validated["path"]["type"] == "null":
            return {
                "string": None,
                "float": None,
                "int": None,
                "boolean": None,
                "uuid": None,
                "cidr": None,
                "ip": None,
                "datetime": None,
                "payload": None,
                "list": None,
            }


@resource(path="/empty-body", factory=DefaultContext, name="EmptyBody")
class EmptyBody(APIBase):
    @view(
        renderer="json",
        schema=PostEmptyBodySchema,
        return_schema=ReturnDummiesSchema,
        validators=[
            bound_schema_validator,
        ],
    )
    def post(self):
        return {"name": "1337"}


@resource(
    path="/",
    collection_path="/mimes",
    name="Mimes",
    factory=DefaultContext,
)
class MultiMimeAPI(APIBase):
    @view(renderer="string", accept_only="text/plain")
    @view(renderer="json")
    def collection_get(self):
        return [{"json": "stuff"}]


def extract_cstruct(request):
    body = {}
    if "query" in request.POST:
        input_file = request.POST["query"].file
        content = BytesIO()
        shutil.copyfileobj(input_file, content)
        try:
            body = json.loads(content.getvalue().decode("utf-8"))
        except (UnicodeDecodeError, JSONDecodeError):
            raise HTTPBadRequest(detail="Invalid Json query")

    cstruct = {
        "method": request.method,
        "url": request.url,
        "path": request.matchdict,
        "body": body,
    }

    for sub, attr in (
        ("querystring", "GET"),
        ("header", "headers"),
        ("cookies", "cookies"),
    ):
        data = getattr(request, attr)
        if isinstance(data, MultiDict):
            data = data.mixed()
        else:
            data = dict(data)
        cstruct[sub] = data
    return cstruct


class CreateNotification(colander.MappingSchema):
    @colander.instantiate(name="body")
    class Body(colander.MappingSchema):
        template = colander.SchemaNode(colander.String())
        sharing_id = colander.SchemaNode(UUID())


@resource(
    path="/test-multiparts",
    name="test_multiparts",
    factory=DefaultContext,
    error_handler=colander_error_handler,
)
class NotificationV2API(APIBase):
    @view(
        renderer="json",
        extract_cstruct=extract_cstruct,
        schema=CreateNotification,
        validators=[
            bound_schema_validator,
        ],
    )
    def post(self):
        params = self.request.validated
        return Response(params["body"]["sharing_id"])
