import json
import shutil
from datetime import datetime
from io import BytesIO
from json.decoder import JSONDecodeError
from typing import Any, Mapping, Optional, Sequence
from uuid import UUID

import translationstring
from cornice.resource import view as cornice_view
from pydantic import BaseModel, Field, IPvAnyAddress, IPvAnyNetwork, validator
from pyramid.httpexceptions import HTTPBadRequest, HTTPConflict, HTTPNotFound
from pyramid.response import Response
from webob.multidict import MultiDict

from sequoia.base3.api import APIBase
from sequoia.base3.helpers.pydantic import (
    PaginationSchema,
    PathInfoField,
    PostBodyField,
    QueryStringField,
    legacy_paginated_view,
    legacy_view,
    resource,
    view,
)
from sequoia.base3.helpers.pydantic.response_schema import (
    CollectionResponseSchema,
    ResponseSchema,
)

_ = translationstring.TranslationStringFactory("dummy_factory_domain")


class DummyNotFound(HTTPNotFound):
    object_name = "dummy"


class PostDummiesSchema(BaseModel):
    name: str = PostBodyField(min_length=3, max_length=7)


class GetDummiesSchema(BaseModel):
    name: str = PathInfoField(min_length=3, max_length=7)


t_float = float
t_int = int
t_datetime = datetime


class TypesParams(BaseModel):
    type: Optional[str] = PathInfoField()
    string: Optional[str] = PostBodyField()
    float: Optional[t_float] = PostBodyField()
    int: Optional[t_int] = PostBodyField()
    boolean: Optional[bool] = PostBodyField()
    uuid: Optional[UUID] = PostBodyField()
    cidr: Optional[IPvAnyNetwork] = PostBodyField()
    ip: Optional[IPvAnyAddress] = PostBodyField()
    datetime: Optional[t_datetime] = PostBodyField()
    list: Optional[Sequence[str]] = PostBodyField()
    payload: Optional[Mapping[str, Any]] = PostBodyField()


class GetUnprocessableEntity(BaseModel):
    name: str = PathInfoField()

    @validator("name")
    def validate_name(name: Any):
        raise ValueError("Invalide")


class PageFilterSchema(PaginationSchema):
    dummyfilter: Sequence[str] = QueryStringField()


class PostEmptyBodySchema(BaseModel):
    name: str = PostBodyField()


class PostDummiesResponse(BaseModel):
    href: str
    internal_href: Optional[str] = None


class Dummy(BaseModel):
    name: str = Field(min_length=3, max_length=7)


@resource(
    path="/dummies/{name}",
    collection_path="/dummies",
    name="Dummy",
)
class DummyAPI(APIBase):
    @legacy_view(
        schema=PostDummiesSchema,
        return_schema=ResponseSchema[PostDummiesResponse],
    )
    def collection_post(self):
        params = self.request.validated["body"]
        if params["name"] == "barbie":
            raise HTTPBadRequest(
                explanation=_("Dummy ${name} already exists"), mapping=params
            )
        if params["name"] == "get_url":
            return {"href": self.get_url("Dummy", name=params["name"])}
        if params["name"] == "ken":
            raise HTTPConflict(
                errors=[
                    {
                        "name": "t-name",
                        "location": "t-location",
                        "description": "t-desc",
                    }
                ]
            )

        if params["name"] == "intern":
            return {
                "href": self.get_url("Dummy", name=params["name"]),
                "internal_href": self.get_internal_url("Dummy", name=params["name"]),
            }

        return PostDummiesResponse(href=self.get_url("Dummy", name=params["name"]))

    @view(
        http_cache=60,
        http_vary="X-Variant",
    )
    def get(self, dummy_req: GetDummiesSchema) -> Dummy:
        if dummy_req.name == "boom":
            # -> internal server error
            raise RuntimeError("Sometime it happens")
        elif dummy_req.name == "bad":
            return Dummy(name="too-big-to-be-returned")
        elif dummy_req.name == "400":
            self.request.errors = [
                {
                    "description": "That is what you asked for",
                    "location": "path",
                    "name": "name",
                }
            ]
            raise HTTPBadRequest(explanation="400 is not a valide name")
        elif dummy_req.name == "400-lcz":
            self.request.errors = [
                {
                    "description": _("This should be localized"),
                    "location": "path",
                    "name": "name",
                },
                {
                    "description": _(
                        "With ${parma1} too", mapping={"parma1": "mapping"}
                    ),
                    "location": "path",
                    "name": "name",
                },
            ]
            raise HTTPBadRequest(explanation="400 is not a valide name")
        elif dummy_req.name == "nope":
            params = self.request.validated["path"]
            raise DummyNotFound(explanation="${name} was not found", mapping=params)
        return Dummy(name=dummy_req.name)


@resource(
    path="/unprocessable_entity/{name}",
    collection_path="/unprocessable_entity",
    name="UnprocessableEntity",
)
class UnprocessableEntityAPI(APIBase):
    @legacy_view(
        renderer="pydantic",
        schema=GetUnprocessableEntity,
        http_cache=(60, {"private": True}),
    )
    def get(self):
        raise RuntimeError("Should never get here")


@resource(
    path="/pages/{name}",
    collection_path="/pages",
    name="PagesWithPydantic",
)
class PageAPI(APIBase):
    @legacy_view(
        renderer="string",
        schema=GetDummiesSchema,
        http_cache=(60, {"public": True}),
        http_vary=("X-Variant-1", "X-Variant-2"),
    )
    def get(self):
        return "Name is: %(name)s" % self.request.validated["path"]

    @legacy_paginated_view(renderer="string", schema=PaginationSchema)
    def collection_get(self, offset, limit, sort_by):
        return "We have pages"


@resource(
    path="/otherpages/{name}",
    collection_path="/otherpages",
    name="OhterPages",
)
class OtherPageAPI(APIBase):
    @legacy_paginated_view(
        renderer="pydantic",
        schema=PageFilterSchema,
        return_schema=CollectionResponseSchema[Dummy],
    )
    def collection_get(self, offset, limit, sort_by):
        acceptable = ("application/json", "text/json", "text/plain")
        content_type = acceptable[0]
        content_types = self.request.accept.acceptable_offers(acceptable)
        if content_types:
            content_type = content_types[0][0]

        if content_type == "text/plain":
            return "We have pages"

        data = [Dummy(name="item %d" % i) for i in range(1, 11)]
        rdata = data[offset:limit + offset]  # fmt: skip
        return CollectionResponseSchema[Dummy](
            headers={"X-Total-Count": len(data)}, body=rdata
        )


@resource(
    path="/protected",
    name="ProtectedResources",
)
class ProtectedResourceAPI(APIBase):
    @legacy_view(
        renderer="string",
        permission="dummy:protected:read",
        http_cache=(60, {"private": True}),
    )
    def get(self):
        return "Confidential"


@resource(path="/documented", name="DocumentedViews")
class DocumentedResourceAPI(APIBase):
    @legacy_view(renderer="string", documentation={"status_code": 201})
    def get(self):
        return "created"


@resource(
    path="/types/{type}",
    collection_path="/types",
    name="ReturnTypes",
)
class TypesAPI(APIBase):
    @legacy_view(
        renderer="pydantic",
        schema=TypesParams,
        return_schema=ResponseSchema[TypesParams],
    )
    def collection_post(self):
        # XXX We user colander at the moment to validate
        # returns and colander do not know pydantic and will not.
        # Keep that until we don the return field here.
        data = self.request.validated["body"]
        if "ip" in data:
            data["ip"] = str(data["ip"])
        if "cidr" in data:
            data["cidr"] = str(data["cidr"])
        return {"body": data}

    @legacy_view(
        renderer="pydantic",
        schema=TypesParams,
        return_schema=ResponseSchema[TypesParams],
    )
    def get(self):
        if self.request.validated["path"]["type"] == "null":
            return {"body": {}}


@resource(path="/empty-body", name="EmptyBodyWithPydantic")
class EmptyBody(APIBase):
    @legacy_view(
        renderer="pydantic",
        schema=PostEmptyBodySchema,
        return_schema=ResponseSchema[Dummy],
    )
    def post(self):
        return {"name": "1337"}


class JasonStuff(BaseModel):
    json_: str = Field(alias="json")


@resource(
    path="/pydantic",
    collection_path="/mimes",
    name="MimesWithPydantic",
)
class MultiMimeAPI(APIBase):
    @cornice_view(renderer="string", accept_only="text/plain")
    @legacy_view(
        schema=BaseModel,
        return_schema=CollectionResponseSchema[JasonStuff],
    )
    def collection_get(self):
        return {"body": [{"json": "stuff"}]}


def extract_cstruct(request):
    body = {}
    if "query" in request.POST:
        input_file = request.POST["query"].file
        content = BytesIO()
        shutil.copyfileobj(input_file, content)
        try:
            body = json.loads(content.getvalue().decode("utf-8"))
        except (UnicodeDecodeError, JSONDecodeError):
            raise HTTPBadRequest(detail="Invalid Json query")

    cstruct = {
        "method": request.method,
        "url": request.url,
        "path": request.matchdict,
        "body": body,
    }

    for sub, attr in (
        ("querystring", "GET"),
        ("header", "headers"),
        ("cookies", "cookies"),
    ):
        data = getattr(request, attr)
        if isinstance(data, MultiDict):
            data = data.mixed()
        else:
            data = dict(data)
        cstruct[sub] = data
    return cstruct


class Body(BaseModel):
    template: str
    sharing_id: str


class CreateNotification(BaseModel):
    body: Body = PostBodyField()


@resource(
    path="/test-multiparts",
    name="test_multiparts_with_pydantic",
)
class NotificationV2API(APIBase):
    @legacy_view(
        renderer="pydantic",
        extract_cstruct=extract_cstruct,
        schema=CreateNotification,
    )
    def post(self):
        params = self.request.validated
        return Response(params["body"]["sharing_id"])
