import itertools
import re
import unittest
from urllib.parse import urlencode

import colander
from pyramid import testing

from sequoia.base3.api import APIBase, paginated_view
from sequoia.base3.helpers.colander import FilteredPaginationSchema
from sequoia.base3.testing.case import ApiTestCase


class PaginationTestCase(ApiTestCase):
    app_settings = {"sequoia.export": "sequoia.base3.tests.api.dummy"}

    def test_render_text(self):
        """GET item with string rendering"""

        resp = self.app.get("/pages/0x42")

        self.assertEqual(resp.content_type, "text/plain")
        self.assertEqual(resp.text, "Name is: 0x42")

    def test_render_text_paginated_view(self):
        """GET on collection with a @paginated_view with string rendering"""

        resp = self.app.get("/pages")

        self.assertEqual(resp.content_type, "text/plain")
        self.assertEqual(resp.text, "We have pages")

    def test_paginated_view_filter(self):
        # Test that when sending the same name 2 times,
        # both are shown in the Link: header
        res = self.app.get("/otherpages?per_page=2" "&dummyfilter=foo&dummyfilter=bar")
        self.assertEqual(res.content_type, "application/json")

        self.assertEqual(res.headers["X-Total-Count"], "10")

        for link in re.findall("<(.*?)>", res.headers["Link"]):
            self.assertIn("dummyfilter=foo", link)
            self.assertIn("dummyfilter=bar", link)

        self.assertEqual(len(res.json), 2)


class PaginationWithPydanticTestCase(PaginationTestCase):
    app_settings = {
        "pyramid.includes": "sequoia.base3.helpers.pydantic",
        "sequoia.export": "sequoia.base3.tests.api.dummy_pydantic",
    }


class BaseUnitTestCase(unittest.TestCase):
    def setUp(self):
        request = unittest.mock.Mock()
        request.context = unittest.mock.Mock()
        request.GET = {}

        def mock_current_route_url(_query=None):
            return "?" + urlencode(_query or {})

        request.current_route_url = mock_current_route_url

        self.api = APIBase(request.context, request)

    def _check_link(self, links, fragments):
        self.assertEqual(len(links), 1)
        self.assertIn("link", links)
        splitted_links = re.split("; |, ", links["link"])
        self.assertEqual(len(splitted_links), len(fragments) * 2)

        index = 0
        while index < len(splitted_links):
            # 'rel="last"' -> 'last'
            rel = splitted_links[index + 1][5:-1]
            values = fragments[rel]
            sub_possible = [
                "<?{}>".format("&".join(perm))
                for perm in itertools.permutations(values)
            ]
            self.assertIn(splitted_links[index], sub_possible)
            self.assertEqual(splitted_links[index + 1], 'rel="{}"'.format(rel))
            index += 2

    def test_build_list_headers_first_page_with_filtered_len(self):
        links = self.api.build_list_headers(
            1, per_page=5, sort_by="id", len_=15, filtered_len=8
        )
        fragments = {
            "next": ["per_page=5", "page=2", "sort_by=id"],
            "last": ["per_page=5", "page=2", "sort_by=id"],
        }
        self._check_link(links, fragments)

    def test_build_list_headers_sort_by_list(self):
        links = self.api.build_list_headers(
            1,
            per_page=5,
            sort_by="-type,name",
            len_=15,
            filtered_len=8,
        )
        fragments = {
            "next": ["per_page=5", "page=2", "sort_by=-type%2Cname"],
            "last": ["per_page=5", "page=2", "sort_by=-type%2Cname"],
        }
        self._check_link(links, fragments)

    def test_build_list_headers_first_page(self):
        links = self.api.build_list_headers(
            1,
            per_page=5,
            sort_by="id",
            len_=15,
        )
        fragments = {
            "next": ["per_page=5", "page=2", "sort_by=id"],
            "last": ["per_page=5", "page=3", "sort_by=id"],
        }
        self._check_link(links, fragments)

    def test_build_list_headers_second_page(self):
        links = self.api.build_list_headers(
            2,
            per_page=5,
            sort_by="id",
            len_=15,
        )
        fragments = {
            "next": ["per_page=5", "page=3", "sort_by=id"],
            "last": ["per_page=5", "page=3", "sort_by=id"],
        }
        self._check_link(links, fragments)

    def test_build_list_headers_last_page(self):
        links = self.api.build_list_headers(
            3,
            per_page=5,
            sort_by="id",
            len_=15,
        )
        self.assertEqual(
            links,
            {},
            "On last page, there is no next page, the last page point to the "
            "current url",
        )

    def test_build_list_headers_after_last_page(self):
        links = self.api.build_list_headers(4, per_page=5, sort_by="id", len_=15)
        self.assertEqual(
            links, {}, "If page is more that last page there is no link to return"
        )

    def test_build_list_headers_one_big_page(self):
        links = self.api.build_list_headers(1, per_page=500, sort_by="id", len_=15)
        self.assertEqual(
            links,
            {},
            "If only one page, there is no next page, and last page point to "
            "the current url",
        )

    def test_build_list_headers_no_records(self):
        links = self.api.build_list_headers(
            1,
            per_page=500,
            sort_by="id",
            len_=0,
        )
        self.assertEqual(
            links,
            {},
            "If there is not record in the collection, the only link point to"
            "the current url",
        )

    def test_build_list_headers_invalid_per_page(self):
        links = self.api.build_list_headers(
            1,
            per_page=0,
            sort_by="id",
            len_=50,
        )
        self.assertEqual(
            links,
            {},
            "When per_page is not set properly, don't crash",
        )

    def test_build_list_headers_invalid_length(self):
        links = self.api.build_list_headers(
            1,
            per_page=50,
            sort_by="id",
            len_=None,
        )
        self.assertEqual(
            links,
            {},
            "When None is not set properly, don't crash",
        )


def dummy_view(self, **kwargs):
    return len(kwargs), [
        kwargs,
    ]


class EmptyQuerystring(colander.MappingSchema):
    @colander.instantiate(name="querystring", missing=colander.drop)
    class Querystring(FilteredPaginationSchema):
        pass


class TestPaginatedView(unittest.TestCase):
    def test_no_schema(self):
        with self.assertRaises(RuntimeError) as error:
            paginated_view()(dummy_view)
        self.assertEqual(
            error.exception.args, ("A schema is mandatory to use @paginated_view",)
        )

    def test_bad_schema(self):
        with self.assertRaises(RuntimeError) as error:
            paginated_view(schema=FilteredPaginationSchema)(dummy_view)
        self.assertEqual(
            error.exception.args, ("querystring Node is mandatory in your schema",)
        )

        with self.assertRaises(RuntimeError) as error:
            paginated_view(schema=FilteredPaginationSchema())(dummy_view)
        self.assertEqual(
            error.exception.args, ("querystring Node is mandatory in your schema",)
        )

    def test_empty_querystring(self):
        dummy_class = unittest.mock.MagicMock()
        dummy_class.per_page = None
        dummy_class.request = testing.DummyRequest()
        dummy_class.request.validated = {}
        res = paginated_view(schema=EmptyQuerystring)(dummy_view)(dummy_class)
        self.assertEqual(len(res[0]), 3)
        self.assertEqual(res[0]["offset"], 0)
        self.assertEqual(res[0]["limit"], 1)
        self.assertEqual(res[0]["sort_by"], "id")
