import json
from unittest import TestCase

import colander
from cornice.errors import Errors
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.testing import DummyRequest as _DummyRequest

from ...helpers.colander import NullableSchemaNode
from ...helpers.cornice import bound_schema_validator


class DummyRequest(_DummyRequest):
    def __init__(self, *args, **kwargs):
        kwargs["errors"] = Errors()
        kwargs["validated"] = {}
        super(DummyRequest, self).__init__(*args, **kwargs)
        self.content_type = "application/json"

    @property
    def json_body(self):
        return json.loads(self.body)


class UnflattenQueryStringSchema(colander.MappingSchema):
    @colander.instantiate(name="querystring")
    class Query(colander.MappingSchema):
        id = colander.SchemaNode(colander.String())

        @colander.instantiate(name="meta", missing=colander.drop)
        class Meta(colander.MappingSchema):
            name = colander.SchemaNode(colander.String())

            @colander.instantiate(name="tags", missing=colander.drop)
            class Tags(colander.SequenceSchema):
                name = colander.SchemaNode(
                    colander.String(),
                    validator=colander.Length(min=2),
                )


class RequiredUnflattenQueryStringSchema(colander.MappingSchema):
    @colander.instantiate(name="querystring")
    class Query(colander.MappingSchema):
        id = colander.SchemaNode(colander.String())

        @colander.instantiate(name="meta")
        class Meta(colander.MappingSchema):
            name = colander.SchemaNode(
                colander.String(),
                validator=colander.Length(min=2),
            )


class BoundSchemaTestCase(TestCase):
    def test_instance_or_class(self):
        request = DummyRequest()
        bound_schema_validator(request, schema=colander.MappingSchema)
        self.assertEqual(request.errors, [])

        request = DummyRequest()
        bound_schema_validator(request, schema=colander.MappingSchema())
        self.assertEqual(request.errors, [])

    def test_bad_body(self):
        expected_errors = [
            {
                "description": "Invalid JSON: Expecting property name "
                "enclosed in double quotes: line 1 column "
                "2 (char 1)",
                "location": "body",
                "name": "",
            },
        ]

        request = DummyRequest(body="{bad json}")
        with self.assertRaises(HTTPBadRequest):
            bound_schema_validator(request, schema=colander.MappingSchema)
        self.assertEqual(request.errors, expected_errors)

    def test_dict_not_mapping_type(self):
        class Dict(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.MappingSchema):
                first_name = colander.SchemaNode(colander.String())

        expected_data = {
            "body": {},
            "extra": {},
        }

        expected_errors = [
            {
                "description": '"${val}" is not a mapping type: ${err}',
                "location": "body",
                "name": "",
            },
        ]

        request = DummyRequest(body=json.dumps([]))

        bound_schema_validator(request, schema=Dict())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)
        self.assertEqual(
            request.errors[0]["description"].mapping,
            {"val": [], "err": "Does not implement " "dict-like functionality."},
        )

    def test_dict_ok(self):
        class PostBodyDict(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.MappingSchema):
                missing = colander.SchemaNode(colander.String(), missing=colander.drop)
                default = colander.SchemaNode(colander.String(), missing="default")
                defined = colander.SchemaNode(colander.String())
                mandatory = colander.SchemaNode(colander.String())

        expected_data = {
            "body": {
                "defined": "value",
                "default": "default",
            },
            "extra": {},
        }
        expected_errors = [
            {
                "description": "Required",
                "location": "body",
                "name": "mandatory",
            },
        ]

        request = DummyRequest(
            body=json.dumps(
                {
                    "defined": "value",
                    "default": "default",
                }
            ),
        )
        bound_schema_validator(request, schema=PostBodyDict())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)

    def test_list_not_iterable(self):
        class List(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.SequenceSchema):
                name = colander.SchemaNode(colander.String())

        expected_data = {
            "body": [],
            "extra": {},
        }

        expected_errors = [
            {
                "description": '"${val}" is not iterable',
                "location": "body",
                "name": "",
            },
        ]

        request = DummyRequest(body=json.dumps({}))

        bound_schema_validator(request, schema=List())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)
        self.assertEqual(request.errors[0]["description"].mapping, {"val": {}})

    def test_post_list_simple_type_ok(self):
        class PostBodyList(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.SequenceSchema):
                name = colander.SchemaNode(colander.String())

        expected_data = {
            "body": [
                "0",
                colander.null,
                "2",
            ],
            "extra": {},
        }
        expected_errors = [
            {
                "description": "Required",
                "location": "body",
                "name": "1",
            },
        ]

        request = DummyRequest(
            body=json.dumps(
                [
                    "0",
                    "",
                    "2",
                ]
            ),
        )
        bound_schema_validator(request, schema=PostBodyList())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)

    def test_post_body_list_complex_type_ok(self):
        class PostBodyList(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.SequenceSchema):
                @colander.instantiate()
                class Item(colander.MappingSchema):
                    name = colander.SchemaNode(colander.String())

        expected_data = {
            "body": [
                {"name": "0"},
                colander.null,
                {"name": "2"},
            ],
            "extra": {},
        }
        expected_errors = [
            {
                "description": "Required",
                "location": "body",
                "name": "1.name",
            },
        ]

        request = DummyRequest(
            body=json.dumps(
                [
                    {"name": "0"},
                    {"name": ""},
                    {"name": "2"},
                ]
            ),
        )
        bound_schema_validator(request, schema=PostBodyList())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)

    def test_validator_ok(self):
        class PostBodyDict(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.MappingSchema):
                name = colander.SchemaNode(colander.String())

                def validator(self, node, cstruct):
                    if cstruct["name"] == "raise":
                        raise colander.Invalid(node, "Using raise will Raise")

        expected_data = {
            "body": {
                "name": "no raise",
            },
            "extra": {},
        }
        expected_errors = []

        request = DummyRequest(
            body=json.dumps(
                {
                    "name": "no raise",
                }
            ),
        )
        bound_schema_validator(request, schema=PostBodyDict())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)

    def test_validator_ok_drop_null(self):
        class PostBodyDict(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.MappingSchema):
                firstname = colander.SchemaNode(
                    colander.String(), missing=colander.drop
                )
                lastname = colander.SchemaNode(colander.String(), missing=colander.drop)
                orgname = NullableSchemaNode(colander.String(), missing=colander.drop)
                streetaddr = NullableSchemaNode(
                    colander.String(), missing=colander.drop
                )

        expected_data = {
            "body": {"orgname": None},
            "extra": {},
        }
        expected_errors = []

        request = DummyRequest(
            body=json.dumps(
                {
                    "firstname": None,
                    "lastname": "",
                    "orgname": None,
                    "streetaddr": "",
                }
            ),
        )
        bound_schema_validator(request, schema=PostBodyDict())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)

    def test_validator_ko(self):
        class PostBodyDict(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.MappingSchema):
                name = colander.SchemaNode(colander.String())

                def validator(self, node, cstruct):
                    if cstruct["name"] == "raise":
                        raise colander.Invalid(node, "Using raise will Raise")

        expected_data = {
            "body": {
                "name": "raise",
            },
            "extra": {},
        }
        expected_errors = [
            {
                "description": "Using raise will Raise",
                "location": "body",
                "name": "",
            },
        ]

        request = DummyRequest(
            body=json.dumps(
                {
                    "name": "raise",
                }
            ),
        )
        bound_schema_validator(request, schema=PostBodyDict())
        self.assertEqual(request.validated, expected_data)
        self.assertEqual(request.errors, expected_errors)

    def test_mapping_unknown_preserve(self):
        class PlusOne(colander.Integer):
            # check that custom nodes are used
            def deserialize(self, node, value):
                return super().deserialize(node, value) + 1

        class PostBodyDict(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.MappingSchema):
                @staticmethod
                def schema_type():
                    return colander.MappingSchema.schema_type("preserve")

                id = colander.SchemaNode(PlusOne())

        post_body = {
            "id": 78,
            "string": "meh",
            "int": 11,
            "list": [1, "foo"],
            "dict": {"titi": "toto"},
            "null": None,
            "bool": True,
        }

        request = DummyRequest(body=json.dumps(post_body))
        bound_schema_validator(request, schema=PostBodyDict())

        expected = post_body.copy()
        expected["id"] = 79

        self.assertEqual(request.validated["body"], expected)

    def test_mapping_unknown_raise(self):
        class PostBodyDict(colander.MappingSchema):
            @colander.instantiate(name="body")
            class Body(colander.MappingSchema):
                @staticmethod
                def schema_type():
                    return colander.MappingSchema.schema_type("raise")

                id = colander.SchemaNode(colander.Integer())

        post_body = {"id": 1, "error": "refusé"}

        request = DummyRequest(body=json.dumps(post_body))

        bound_schema_validator(request, schema=PostBodyDict())

        expected_errors = [
            {
                "description": 'Unrecognized keys in mapping: "${val}"',
                "location": "body",
                "name": "",
            },
        ]

        self.assertEqual(request.errors, expected_errors)
        self.assertEqual(
            request.errors[0]["description"].mapping, {"val": {"error": "refusé"}}
        )
        self.assertEqual(request.validated["body"], {"id": 1})

    def test_mapping_in_querystring_ok_with_missing(self):
        request = DummyRequest(
            params={
                "id": "tada",
                "trailing": "should be ignored",
                "meta.name": "myname",
            }
        )

        expected = {
            "extra": {},
            "querystring": {"id": "tada", "meta": {"name": "myname"}},
        }

        bound_schema_validator(request, schema=UnflattenQueryStringSchema())
        self.assertEqual(request.validated, expected)
        self.assertEqual(request.errors, [])

    def test_mapping_in_querystring_ok_complete(self):
        request = DummyRequest(
            params={
                "id": "tada",
                "trailing": "should be ignored",
                "meta.name": "myname",
                "meta.tags.0": "tag0",
                "meta.tags.1": "tag1",
            }
        )

        expected = {
            "extra": {},
            "querystring": {
                "id": "tada",
                "meta": {"name": "myname", "tags": ["tag0", "tag1"]},
            },
        }

        bound_schema_validator(request, schema=UnflattenQueryStringSchema())
        self.assertEqual(request.validated, expected)
        self.assertEqual(request.errors, [])

    def test_mapping_in_querystring_invalid(self):
        request = DummyRequest(
            params={
                "id": "tada",
                "trailing": "should be ignored",
                "meta.tags.0": "tag0",
                "meta.tags.1": "tag1",
            }
        )

        expected = {
            "extra": {},
            "querystring": {
                "id": "tada"
                # FIXME: the tags should be deserialized in order
                # to be validated.
                # 'meta': {'tags': ['tag0', 'tag1']}
            },
        }
        bound_schema_validator(request, schema=UnflattenQueryStringSchema())
        self.assertEqual(request.validated, expected)
        self.assertEqual(
            request.errors,
            [
                {
                    "location": "querystring",
                    "name": "meta.name",
                    "description": "Required",
                }
            ],
        )

    def test_mapping_in_querystring_invalid_2(self):
        request = DummyRequest(
            params={
                "id": "tada",
                "trailing": "should be ignored",
                "meta.tags.0": "tag0",
                "meta.tags.1": "t",
                "meta.tags.2": "tag2",
            }
        )

        expected = {
            "extra": {},
            "querystring": {
                "id": "tada"
                # FIXME: the tags should be deserialized in order
                # to be validated.
                # 'meta': {'tags': ['tag0', colander.null ,'tag2']}
            },
        }
        bound_schema_validator(request, schema=UnflattenQueryStringSchema())
        self.assertEqual(request.validated, expected)
        self.assertEqual(
            sorted(request.errors, key=lambda val: val["name"]),
            [
                {
                    "location": "querystring",
                    "name": "meta.name",
                    "description": "Required",
                },
                {
                    "description": "Shorter than minimum length 2",
                    "location": "querystring",
                    "name": "meta.tags.1",
                },
            ],
        )

    def test_mapping_in_querystring_missing(self):
        request = DummyRequest(
            params={
                "id": "tada",
                "trailing": "should be ignored",
            }
        )

        expected = {"extra": {}, "querystring": {"id": "tada"}}
        errors = [
            {
                "location": "querystring",
                "name": "meta.name",
                "description": "Required",
            }
        ]
        bound_schema_validator(request, schema=RequiredUnflattenQueryStringSchema())
        self.assertEqual(request.validated, expected)
        self.assertEqual(request.errors, errors)
