from unittest import mock

import colander
from cornice.errors import Errors
from pyramid.testing import DummyRequest

from sequoia.base3.api.dry_run import (
    DryRunReturn,
    dry_run_decorator,
    dry_run_handler,
    dry_run_view,
    is_dry_run,
)
from sequoia.base3.helpers.cornice import bound_schema_validator
from sequoia.base3.testing.case import ApiTestCase


class DummyCorniceRequest(DummyRequest):
    def __init__(self, *args, **kwargs):
        kwargs["errors"] = Errors()
        kwargs["validated"] = {}
        super(DummyCorniceRequest, self).__init__(*args, **kwargs)
        self.content_type = "application/json"


class DryRunTestCase(ApiTestCase):
    def test_DryRunReturn(self):
        data = {"status": "success", "filtered": "filter"}
        res = DryRunReturn().bind(request=DummyRequest()).deserialize(data)
        self.assertEqual(res, {"status": "success"})

        data = {
            "status": "error",
            "errors": [
                {"location": "querystring", "name": "foo", "description": "bar"}
            ],
        }
        res = DryRunReturn().bind(request=DummyRequest()).deserialize(data)
        self.assertEqual(res, data)

    def test_is_dry_run(self):
        self.assertFalse(is_dry_run(DummyRequest()))
        self.assertFalse(is_dry_run(DummyRequest(headers={"Dry-Run": "0"})))
        self.assertFalse(is_dry_run(DummyRequest(headers={"DRY-RUN": "1"})))

        self.assertTrue(is_dry_run(DummyRequest(headers={"Dry-Run": "1"})))
        self.assertTrue(is_dry_run(DummyRequest(headers={"Dry-Run": "true"})))
        self.assertTrue(is_dry_run(DummyRequest(headers={"Dry-Run": "True"})))

    def test_dry_run_handler(self):
        req = DummyRequest(errors=[])
        self.assertEqual(dry_run_handler(req), {"status": "success"})

        req = DummyRequest(
            errors=[{"location": "querystring", "name": "foo", "description": "bar"}]
        )
        self.assertEqual(
            dry_run_handler(req),
            {
                "errors": [
                    {"description": "bar", "location": "querystring", "name": "foo"}
                ],
                "status": "error",
            },
        )

    def test_dry_run_view(self):
        decorator = mock.Mock()
        dry_run_decorator = mock.Mock(return_value=decorator)
        cornice_view = mock.Mock()

        with mock.patch("sequoia.base3.api.dry_run.cornice_view", cornice_view):
            with mock.patch(
                "sequoia.base3.api.dry_run.dry_run_decorator", dry_run_decorator
            ):
                dry_run_view()
        cornice_view.assert_called_once_with(decorator=[decorator])

        previous = mock.Mock()
        cornice_view.reset_mock()
        with mock.patch("sequoia.base3.api.dry_run.cornice_view", cornice_view):
            with mock.patch(
                "sequoia.base3.api.dry_run.dry_run_decorator", dry_run_decorator
            ):
                dry_run_view(decorator=[previous])
        cornice_view.assert_called_once_with(
            decorator=[previous, dry_run_decorator({})]
        )

        cornice_view.reset_mock()
        with mock.patch("sequoia.base3.api.dry_run.cornice_view", cornice_view):
            with mock.patch(
                "sequoia.base3.api.dry_run.dry_run_decorator", dry_run_decorator
            ):
                dry_run_view(foo="bar")

        cornice_view.assert_called_once_with(
            foo="bar", decorator=[dry_run_decorator({})]
        )

    def test_dry_run_decorator(self):
        class DummySchema(colander.MappingSchema):
            @colander.instantiate(name="querystring")
            class Querystring(colander.MappingSchema):
                name = colander.SchemaNode(colander.String())

        class Context(object):
            return_schema = DummySchema

        def unlucky(request, **kw):
            request.errors.add("querystring", "dummy", "unlucky")

        view = mock.Mock(return_value="Intacto")
        view_config = {
            "schema": DummySchema(),
            "validators": [bound_schema_validator, unlucky],
        }
        func = dry_run_decorator(view_config)

        request = DummyCorniceRequest(
            headers={"Dry-Run": "1"},
            context=Context,
        )
        res = func(view)(request.context, request)
        self.assertEqual(res.status, "200 OK")
        self.assertEqual(res.headers["Content-Type"], "application/json")
        self.assertEqual(
            res.json,
            {
                "errors": [
                    {
                        "description": "Required",
                        "location": "querystring",
                        "name": "name",
                    },
                    {
                        "description": "unlucky",
                        "location": "querystring",
                        "name": "dummy",
                    },
                ],
                "status": "error",
            },
        )

        request = DummyCorniceRequest(
            params={"name": "Federico"},
            headers={"Dry-Run": "1"},
            context=Context,
        )
        res = func(view)(request.context, request)
        self.assertEqual(res.status, "200 OK")
        self.assertEqual(res.headers["Content-Type"], "application/json")
        self.assertEqual(
            res.json,
            {
                "errors": [
                    {
                        "description": "unlucky",
                        "location": "querystring",
                        "name": "dummy",
                    }
                ],
                "status": "error",
            },
        )

        self.assertEqual(view.called, 0, "Underlying view should not have been called")

        # test success

        def lucky(request, **kw):
            request.validated["foo"] = "bar"

        view_config = {
            "schema": DummySchema(),
            "validators": [bound_schema_validator, lucky],
        }
        func = dry_run_decorator(view_config)

        request = DummyCorniceRequest(
            params={"name": "Samuel"},
            headers={"Dry-Run": "1"},
            context=Context,
        )
        res = func(view)(request.context, request)
        self.assertEqual(res.status, "200 OK")
        self.assertEqual(res.headers["Content-Type"], "application/json")
        self.assertEqual(res.json, {"status": "success"})

        self.assertEqual(view.called, 0, "Underlying view should not have been called")

        # test that non dry run call the underlying method
        request = DummyCorniceRequest(
            params={"name": "Samuel"},
            context=Context,
        )
        res = func(view)(request.context, request)
        view.assert_called_once_with(request.context, request)
        self.assertEqual(res, "Intacto")

    def test_dry_run_decorator_contract(self):
        # Test that adding 'contracts' to validated inside a validator
        # will return that list when calling the API with dry-run
        # Should work with request.validated and request.validated['extra']
        view = mock.Mock(return_value="Intacto")
        contract = {"id": "void", "name": "Dummy contract", "href": "http://void/"}

        class Context(object):
            return_schema = colander.MappingSchema

        def contract_validator_legacy(request, **kwargs):
            request.validated["contracts"] = [contract]

        def contract_validator_extra(request, **kwargs):
            request.validated["extra"]["contracts"] = [contract]

        def make_func(validator):
            view_config = {
                "schema": colander.MappingSchema(),
                "validators": [bound_schema_validator, validator],
            }
            return dry_run_decorator(view_config)

        request = DummyCorniceRequest(
            headers={"Dry-Run": "1"},
            context=Context,
        )
        func = make_func(contract_validator_legacy)
        res = func(view)(request.context, request)

        self.assertEqual(res.json, {"status": "success", "contracts": [contract]})

        request = DummyCorniceRequest(
            headers={"Dry-Run": "1"},
            context=Context,
        )

        func = make_func(contract_validator_extra)
        res = func(view)(request.context, request)
        self.assertEqual(res.json, {"status": "success", "contracts": [contract]})
