import os.path
from unittest.mock import patch

from sequoia.base3.errors import internal_server_error
from sequoia.base3.testing.case import ApiTestCase


class ErrorTestCase(ApiTestCase):
    app_settings = {
        "sequoia.export": "sequoia.base3.tests.api.dummy",
    }

    def test_explanation_mapping_translated(self):
        resp = self.app.post_json("/dummies", {"name": "barbie"}, status=400)
        self.assertEqual(
            resp.json,
            {
                "message": "Dummy barbie already exists",
                "code": 400,
                "cause": "Bad Request",
                "object": "HTTPBadRequest",
            },
        )

    def test_error_custom_object(self):
        resp = self.app.get("/dummies/nope", status=404)

        self.assertEqual(
            resp.json,
            {
                "message": "nope was not found",
                "code": 404,
                "cause": "Not Found",
                "object": "dummy",
            },
        )

    def test_colander_invalid(self):
        # XXX this test doesn't test `sequoia.api.base.error.colander_invalid`
        resp = self.app.post_json("/dummies", {"name": "bb"}, status=400)
        self.assertEqual(
            resp.json,
            {
                "errors": [
                    {
                        "description": "Shorter than minimum length 3",
                        "location": "body",
                        "name": "name",
                    }
                ],
                "status": "error",
            },
        )

    @patch.object(internal_server_error, "disabled", False)
    def test_internal_server_error(self):
        resp = self.app.get("/dummies/boom", status=500)
        self.assertEqual(
            resp.json,
            {
                "code": 500,
                "message": "The server has either erred or is incapable of "
                "performing the requested operation.",
                "object": "Dummy",
                "cause": "Internal Server Error",
            },
        )

    def test_return_colander_ok(self):
        resp = self.app.get("/dummies/good", status=200)
        assert resp.json == {"name": "good"}

    def test_return_colander_ko(self):
        resp = self.app.get("/dummies/bad", status=500)
        self.assertEqual(
            resp.json,
            {
                "message": "The server has either erred or is incapable of "
                "performing the requested operation.",
                "code": 500,
                "cause": "Internal Server Error",
                "object": "HTTPInternalServerError",
            },
        )

    def test_colander_error_handler(self):
        resp = self.app.get("/unprocessable_entity/none", status=422)
        self.assertEqual(
            resp.json,
            {
                "message": "Unable to process the contained instructions",
                "code": 422,
                "cause": "Unprocessable Entity",
                "errors": [
                    {"description": "Invalide", "location": "path", "name": "name"},
                ],
                "object": "HTTPUnprocessableEntity",
            },
        )

    def test_bad_request(self):
        resp = self.app.get("/dummies/400", status=400)
        self.assertEqual(
            resp.json,
            {
                "message": "400 is not a valide name",
                "code": 400,
                "cause": "Bad Request",
                "errors": [
                    {
                        "description": "That is what you asked for",
                        "location": "path",
                        "name": "name",
                    },
                ],
                "object": "HTTPBadRequest",
            },
        )

    def test_conflict(self):
        resp = self.app.post_json("/dummies", {"name": "ken"}, status=409)
        self.assertEqual(
            resp.json,
            {
                "message": "There was a conflict when trying to "
                "complete your request.",
                "code": 409,
                "cause": "Conflict",
                "errors": [
                    {
                        "description": "t-desc",
                        "location": "t-location",
                        "name": "t-name",
                    },
                ],
                "object": "HTTPConflict",
            },
        )

    def test_localized_error(self):
        resp = self.app.get("/dummies/400-lcz", status=400)
        self.assertEqual(
            resp.json,
            {
                "message": "400 is not a valide name",
                "code": 400,
                "cause": "Bad Request",
                "errors": [
                    {
                        "description": "This should be localized",
                        "location": "path",
                        "name": "name",
                    },
                    {
                        "description": "With mapping too",
                        "location": "path",
                        "name": "name",
                    },
                ],
                "object": "HTTPBadRequest",
            },
        )


class ErrorWithPydanticTestCase(ErrorTestCase):
    app_settings = {
        "pyramid.includes": "sequoia.base3.helpers.pydantic",
        "sequoia.export": "sequoia.base3.tests.api.dummy_pydantic",
    }

    def test_colander_invalid(self):  # because the description changed
        # XXX this test doesn't test `sequoia.api.base.error.colander_invalid`
        resp = self.app.post_json("/dummies", {"name": "bb"}, status=422)
        self.assertEqual(
            resp.json,
            {
                "cause": "Unprocessable Entity",
                "code": 422,
                "errors": [
                    {
                        "description": "ensure this value has at least 3 characters",
                        "location": "body",
                        "name": "name",
                    }
                ],
                "message": "Unable to process the contained instructions",
                "object": "HTTPUnprocessableEntity",
            },
        )


class ForbiddenTestCase(ApiTestCase):
    here = os.path.dirname(__file__)

    app_settings = {
        "sequoia.export": "sequoia.base3.tests.api.dummy",
        "pyramid.includes": "\n  ".join(
            [
                "sequoia.auth.testing",
            ]
        ),
        "sequoia.auth.testing.data": "%s/authenticated_users.json" % here,
    }

    def test_unauthorized_error(self):
        resp = self.app.get("/protected", status=401)
        self.assertEqual(
            resp.json,
            {
                "code": 401,
                "message": "The server could not verify that you authorized to "
                "access the document you requested. Either you "
                "supplied the wrong credentials (e.g., bad api key), "
                "or your access token has expired",
                "object": "HTTPUnauthorized",
                "cause": "Unauthorized",
            },
        )

    def test_forbidden_access(self):
        resp = self.app.get(
            "/protected", status=403, headers={"X-Auth-Testing": "noperm"}
        )
        self.assertEqual(
            resp.json,
            {
                "message": "Access was denied to this resource.",
                "code": 403,
                "cause": "Forbidden",
                "object": "HTTPForbidden",
            },
        )


class ForbiddenWithPydanticTestCase(ForbiddenTestCase):
    here = os.path.dirname(__file__)

    app_settings = {
        "sequoia.export": "sequoia.base3.tests.api.dummy_pydantic",
        "pyramid.includes": "\n  ".join(
            [
                "sequoia.base3.helpers.pydantic",
                "sequoia.auth.testing",
            ]
        ),
        "sequoia.auth.testing.data": "%s/authenticated_users.json" % here,
    }
