# -*- coding: utf-8 -*-

import datetime
import uuid
from unittest import TestCase, mock

import colander
import isodate
import pytz
import transaction
from pyramid.testing import DummyRequest

from sequoia.base3.helpers.colander import (
    CIDR,
    IP,
    UUID,
    ContactType,
    Date,
    DateTime,
    Duration,
    DurationReturn,
    Fqdn,
    FqdnType,
    HTTPAcceptedReturn,
    HTTPCreatedReturn,
    HTTPNoContentReturn,
    IDHTTPAcceptedReturn,
    IDHTTPCreatedReturn,
    ListReturn,
    LocalizedString,
    ManyOf,
    ManyOfSortBy,
    MappingSchema,
    Multiple,
    NullableSchemaNode,
    ScalarSequenceSchema,
    SchemaNode,
    SequenceSchema,
    SplittedString,
    String,
    StrippedString,
    UUIDHTTPCreatedReturn,
    WildcardFqdn,
    cast_idna,
    email_validator,
    nested_missing,
    sort_by_preparer,
    sql_like_preparer,
    validate_email,
)

from .sqlalchemy import Dummy, MiniDummy, TestCommon


class DummyReturn(MappingSchema):
    name = SchemaNode(String())
    lname = SchemaNode(String())


class DummySequenceReturn(SequenceSchema):
    _ = DummyReturn()


class ColanderTestCase(TestCase):
    def test_sql_like_preparer(self):
        self.assertEqual(sql_like_preparer("usernam*"), "usernam%")
        self.assertEqual(sql_like_preparer("usernam?"), "usernam_")
        self.assertEqual(sql_like_preparer("usernam%"), "usernam\\%")
        self.assertEqual(sql_like_preparer("usernam_"), "usernam\\_")
        self.assertEqual(sql_like_preparer("usernam*s"), "usernam%s")
        self.assertEqual(sql_like_preparer("usernam\\*s"), "usernam*s")
        self.assertEqual(sql_like_preparer("usernam\\?s"), "usernam?s")
        self.assertEqual(sql_like_preparer("usernam?s"), "usernam_s")
        self.assertEqual(sql_like_preparer("usernam%s"), "usernam\\%s")
        self.assertEqual(sql_like_preparer("usernam_s"), "usernam\\_s")
        self.assertEqual(sql_like_preparer("*s"), "%s")
        self.assertEqual(sql_like_preparer("?s"), "_s")
        self.assertEqual(sql_like_preparer("%s"), "\\%s")
        self.assertEqual(sql_like_preparer("_s"), "\\_s")
        self.assertEqual(
            sql_like_preparer("\\* is it _work*_ t\\*g?ther\\?"),
            "* is it \\_work%\\_ t*g_ther?",
        )

    def test_ManyOf(self):
        valid = ManyOf(["a", "b"])
        for param in (["a"], ["a", "b"], []):
            self.assertEqual(valid(None, param), param)

        self.assertEqual(valid(None, "a"), ["a"])

        for param in (["c"], ["a", "c"], ["-b"]):
            self.assertRaises(colander.Invalid, valid, None, param)

        self.assertRaises(colander.Invalid, valid, None, ["a", "a"])
        self.assertRaises(colander.Invalid, valid, None, ["a", "-a"])

    def test_ManyOfSortBy(self):
        valid = ManyOfSortBy(["a", "b"])
        for param in (["a"], ["a", "b"], [], ["a", "-b"]):
            self.assertEqual(valid(None, param), param)

        self.assertEqual(valid(None, "a"), ["a"])

        for param in (["c"], ["a", "c"]):
            self.assertRaises(colander.Invalid, valid, None, param)

        self.assertRaises(colander.Invalid, valid, None, ["a", "a"])
        self.assertRaises(colander.Invalid, valid, None, ["a", "-a"])

    def test_SplittedString(self):
        sort_by = SplittedString()

        self.assertEqual(sort_by.deserialize(None, "a"), ["a"])
        self.assertEqual(sort_by.deserialize(None, "a,b"), ["a", "b"])

    def test_CIDR(self):
        cidr = CIDR()
        self.assertEqual(cidr.deserialize(None, "127.0.0.1"), "127.0.0.1")
        self.assertEqual(cidr.deserialize(None, "0.0.0.0/0"), "0.0.0.0/0")
        self.assertEqual(cidr.deserialize(None, "::1"), "::1")
        self.assertEqual(cidr.deserialize(None, "fe80:e47c::/64"), "fe80:e47c::/64")

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "subnet"
            cidr.deserialize(node, "choucroute")

        self.assertEqual(
            invalid.exception.asdict(), {"subnet": "choucroute is not a valid CIDR"}
        )

    def test_Date(self):
        date = Date()
        self.assertEqual(
            date.deserialize(None, datetime.datetime(2015, 0o1, 16)),
            datetime.date(2015, 0o1, 16),
        )

        self.assertEqual(
            date.deserialize(None, datetime.date(2015, 0o1, 16)),
            datetime.date(2015, 0o1, 16),
        )

        self.assertEqual(
            date.deserialize(None, "2015-01-16"), datetime.date(2015, 0o1, 16)
        )

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "dt"
            date.deserialize(node, "First Day July Ten years BC")

        self.assertEqual(invalid.exception.asdict(), {"dt": "Invalid date"})

    def test_DateTime(self):
        date = DateTime()
        self.assertEqual(
            date.deserialize(None, datetime.datetime(2015, 0o1, 16)),
            datetime.datetime(2015, 0o1, 16),
        )

        self.assertEqual(
            date.deserialize(None, "2015-01-16 00:00:00Z"),
            datetime.datetime(2015, 0o1, 16, tzinfo=pytz.UTC),
        )

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "dt"
            date.deserialize(node, "First Day July Ten years BC")

        self.assertEqual(invalid.exception.asdict(), {"dt": "Invalid date"})

    def test_Duration(self):
        duration = Duration()
        self.assertEqual(
            duration.deserialize(None, "P7Y5M1DT6H4M2S"),
            isodate.duration.Duration(1, 21842, 0, years=7, months=5),
        )

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "dt"
            duration.deserialize(node, "not valid")

        self.assertEqual(
            invalid.exception.asdict(),
            {"dt": "not valid is not a valid ISO8601 duration"},
        )

    def test_DurationReturn(self):
        duration_return = DurationReturn()
        self.assertEqual(
            duration_return.deserialize(
                None,
                datetime.timedelta(days=3, hours=5, seconds=18),
            ),
            "P3DT5H18S",
        )

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "dt"
            duration_return.deserialize(node, "not valid")

        self.assertEqual(
            invalid.exception.asdict(),
            {"dt": "not valid is not a valid datetime.timedelta"},
        )

    def test_IP(self):
        ip = IP()
        self.assertEqual(ip.deserialize(None, "127.0.0.1"), "127.0.0.1")
        self.assertEqual(ip.deserialize(None, "::1"), "::1")

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "ip"
            ip.deserialize(node, "10.0.0.0/16")

        self.assertEqual(
            invalid.exception.asdict(), {"ip": "10.0.0.0/16 is not a valid IP"}
        )

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "ip"
            ip.deserialize(node, "fe80:e47c::/64")

        self.assertEqual(
            invalid.exception.asdict(), {"ip": "fe80:e47c::/64 is not a valid IP"}
        )

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "ip"
            ip.deserialize(node, "choucroute")

        self.assertEqual(
            invalid.exception.asdict(), {"ip": "choucroute is not a valid IP"}
        )

    def test_UUID(self):
        uuid_ = uuid.uuid1()
        uuid_co = UUID()

        uuid_str = "cf1ebeb7-9c26-4130-9d49-f5aa4a5cf696"
        uuid_str2 = "cf1ebeb79c2641309d49f5aa4a5cf696"
        self.assertEqual(uuid_co.deserialize(None, uuid_str), uuid_str)
        self.assertEqual(uuid_co.deserialize(None, uuid_str2), uuid_str2)
        self.assertEqual(uuid_co.deserialize(None, uuid_), str(uuid_))

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "uuid"
            uuid_co.deserialize(node, 21312312414141)

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "uuid"
            uuid_co.serialize(node, 21312312414141)

        self.assertEqual(
            invalid.exception.asdict(), {"uuid": '"21312312414141" is not an UUID'}
        )

        with self.assertRaises(colander.Invalid) as invalid:
            node = mock.Mock()
            node.name = "uuid"
            uuid_co.deserialize(node, "choucroute")

        self.assertEqual(
            invalid.exception.asdict(), {"uuid": '"choucroute" is not an UUID'}
        )

    def test_Multiple(self):
        validator = Multiple(2)

        self.assertIsNone(validator(None, 2))
        self.assertIsNone(validator(None, 4))

        self.assertRaises(colander.Invalid, validator, None, 1)
        self.assertRaises(colander.Invalid, validator, None, 3)

    def test_ListReturn_default(self):
        class DummyListReturn(ListReturn):
            @colander.instantiate(name="body")
            class Sequence(colander.SequenceSchema):
                @colander.instantiate()
                class Content(colander.MappingSchema):
                    name = colander.SchemaNode(colander.String())

        data = {
            "X-Total-Count": 20,
            "body": [{"name": "foo", "age": 42}, {"name": "bar", "age": 42.5}],
            "xtra": "filtered",
        }
        expected = {"X-Total-Count": 20, "body": [{"name": "foo"}, {"name": "bar"}]}
        res = DummyListReturn().deserialize(data)
        self.assertEqual(res, expected)

    def test_HTTPCreatedReturn_default(self):
        uuid_ = str(uuid.uuid1())
        location = "http://localhost/testing/%s" % uuid_
        data = {"location": location}
        expected = {
            "status_code": 201,
            "message": "Resource Created",
            "location": location,
        }
        res = HTTPCreatedReturn().deserialize(data)
        self.assertEqual(res, expected)

    def test_HTTPCreatedReturn_missing(self):
        with self.assertRaises(colander.Invalid) as invalid:
            HTTPCreatedReturn().deserialize({})
        self.assertEqual(invalid.exception.asdict(), {"location": "Required"})

    def test_IDHTTPCreatedReturn_default(self):
        uuid_ = str(uuid.uuid1())
        location = "http://localhost/testing/%s" % uuid_
        data = {"id": uuid_, "location": location}
        expected = {
            "status_code": 201,
            "message": "Resource Created",
            "location": location,
            "id": uuid_,
        }
        res = IDHTTPCreatedReturn().deserialize(data)
        self.assertEqual(res, expected)

    def test_IDHTTPCreatedReturn_errors_missing(self):
        with self.assertRaises(colander.Invalid) as invalid:
            IDHTTPCreatedReturn().deserialize({})
        self.assertEqual(
            invalid.exception.asdict(), {"id": "Required", "location": "Required"}
        )

    def test_IDHTTPCreatedReturn_errors_invalid(self):
        with self.assertRaises(colander.Invalid) as invalid:
            IDHTTPCreatedReturn().deserialize({"id": "choucroute"})
        self.assertEqual(
            invalid.exception.asdict(),
            {"id": '"choucroute" is not an UUID', "location": "Required"},
        )

    def test_IDHTTPCreatedReturn_optional_params(self):
        # XXX tested but status_code should not be overridable in fact...
        uuid_ = str(uuid.uuid1())
        location = "http://localhost/testing/%s" % uuid_
        data = {
            "id": uuid_,
            "href": "/testing/%s" % uuid_,
            "status_code": 201,
            "message": "Dummy Created",
            "xtra": "should not raised",
            "location": location,
        }
        expected = {
            "status_code": 201,
            "message": "Dummy Created",
            "location": location,
            "href": "/testing/%s" % uuid_,
            "id": uuid_,
        }
        res = IDHTTPCreatedReturn().deserialize(data)
        self.assertEqual(res, expected)

    def test_HTTPAcceptedParams_default(self):
        expected = {"message": "Request Accepted", "status_code": 202}
        res = HTTPAcceptedReturn().deserialize({})
        self.assertEqual(res, expected)

    def test_HTTPAcceptedReturn_optional_params(self):
        uuid_ = str(uuid.uuid1())
        location = "http://localhost/testing/%s" % uuid_
        data = {
            "id": uuid_,
            "status_code": 200,
            "message": "Dummy Accepted",
            "xtra": "should not raised",
            "location": location,
        }
        expected = {
            "status_code": 200,
            "message": "Dummy Accepted",
            "location": location,
        }
        res = HTTPAcceptedReturn().deserialize(data)
        self.assertEqual(res, expected)

    def test_IDHTTPAcceptedReturn_default(self):
        uuid_ = str(uuid.uuid1())
        expected = {"message": "Request Accepted", "status_code": 202, "id": uuid_}
        res = IDHTTPAcceptedReturn().deserialize({"id": uuid_})

        self.assertEqual(res, expected)

    def test_IDHTTPAcceptedReturn_params(self):
        uuid_ = str(uuid.uuid1())
        data = {
            "id": uuid_,
            "status_code": 200,
            "href": "/testing/%s" % uuid_,
            "message": "Request Running",
        }
        expected = {
            "message": "Request Running",
            "status_code": 200,
            "href": "/testing/%s" % uuid_,
            "id": uuid_,
        }
        res = IDHTTPAcceptedReturn().deserialize(data)
        self.assertEqual(res, expected)

    def test_IDHTTPAcceptedReturn_errors_missing(self):
        with self.assertRaises(colander.Invalid) as invalid:
            IDHTTPAcceptedReturn().deserialize({})
        self.assertEqual(invalid.exception.asdict(), {"id": "Required"})

    def test_IDHTTPAcceptedReturn_errors_invalid(self):
        with self.assertRaises(colander.Invalid) as invalid:
            IDHTTPAcceptedReturn().deserialize({"id": "choucroute"})
        self.assertEqual(
            invalid.exception.asdict(), {"id": '"choucroute" is not an UUID'}
        )

    def test_HTTPNoContentReturn(self):
        res = HTTPNoContentReturn().deserialize({})
        self.assertEqual(res, {"status_code": 204})

    def test_HTTPNoContentReturn_optional(self):
        res = HTTPNoContentReturn().deserialize(
            {"status_code": 200, "wathever": "no impact"}
        )
        self.assertEqual(res, {"status_code": 200})

    def test_UUIDHTTPCreatedReturn_default(self):
        # XXX deprecated type, use IDHTTPCreatedReturn instead
        uuid_ = str(uuid.uuid1())
        location = "http://localhost/testing/%s" % uuid_
        data = {"uuid": uuid_, "location": location}
        expected = {
            "status_code": 201,
            "message": "Resource Created",
            "location": "http://localhost/testing/%s" % uuid_,
            "uuid": uuid_,
        }
        res = UUIDHTTPCreatedReturn().deserialize(data)
        self.assertEqual(res, expected)

    def test_Fqdn(self):
        good_fqdn = "good-fqdn-like-we-like.tech"
        bad_fqdn = "-bad-fqdn-with-leading-dash.com"
        bad_fqdn2 = "*.good-fqdn-like-we-like.tech"

        self.assertEqual(Fqdn().validator(None, good_fqdn), good_fqdn)
        with self.assertRaises(colander.Invalid):
            Fqdn().validator(None, bad_fqdn)

        with self.assertRaises(colander.Invalid):
            Fqdn().validator(None, bad_fqdn2)

    def test_WildcardFqdn(self):
        good_fqdn = "*.fun.org"
        good_fqdn2 = "good-fqdn-like-we-like.tech"
        bad_fqdn = "*.-bad-fqdn-with-dash.com"
        bad_fqdn2 = "*.*.bad-fqdn-with-dash.com"

        self.assertEqual(WildcardFqdn().validator(None, good_fqdn), good_fqdn)
        self.assertEqual(WildcardFqdn().validator(None, good_fqdn2), good_fqdn2)

        with self.assertRaises(colander.Invalid):
            Fqdn().validator(None, bad_fqdn)

        with self.assertRaises(colander.Invalid):
            Fqdn().validator(None, bad_fqdn2)

    def test_WilcardFqdnIDNA(self):
        deserialize = WildcardFqdn().typ.deserialize

        fqdn = "*.bières.org"
        self.assertEqual(deserialize(None, fqdn), "*.xn--bires-5ra.org")

        fqdn = "*.mañana.es"
        self.assertEqual(deserialize(None, fqdn), "*.xn--maana-pta.es")

        fqdn = "*.weiß.de"
        self.assertEqual(deserialize(None, fqdn), "*.xn--wei-7ka.de")

    def test_ScalarSequenceSchema(self):
        class FakeSchema(ScalarSequenceSchema):
            _ = colander.SchemaNode(colander.String())

        result = FakeSchema().deserialize(["a", "b", "c"])
        self.assertEqual(result, ["a", "b", "c"])

        result = FakeSchema().deserialize("a")
        self.assertEqual(result, ["a"])

    def test_cast_idna(self):
        value = "変態.co.jp"
        self.assertEqual(cast_idna(None, value), "xn--urs05q.co.jp")

    def test_cast_idna_xn(self):
        value = "xn--urs05q.co.jp"
        self.assertEqual(cast_idna(None, value), "xn--urs05q.co.jp")

    def test_cast_idna_ip(self):
        value = "192.168.0.1"
        with self.assertRaises(colander.Invalid):
            cast_idna(None, value)

    def test_cast_idna_upper(self):
        value = "MYNEWDOMAIN.FR"
        self.assertEqual(cast_idna(None, value), "mynewdomain.fr")

    def test_cast_idna_2008(self):
        value = "weiße-zähne.tips"
        self.assertEqual(cast_idna(None, value), "xn--weie-zhne-i1a2d.tips")

    def test_cast_idna_2008_tld(self):
        value = "whois.\u5728\u7ebf"
        self.assertEqual(cast_idna(None, value), "whois.xn--3ds443g")

    def test_cast_idna_2008_2(self):
        value = "xn--weie-zhne-i1a2d.tips"
        self.assertEqual(cast_idna(None, value), "xn--weie-zhne-i1a2d.tips")

    def test_cast_idna_2008_3(self):
        value = "xn--weisse-zhne-s8a.tips"
        self.assertEqual(cast_idna(None, value), "xn--weisse-zhne-s8a.tips")

    def test_cast_idna_2003_only(self):
        value = "xn--ichmag-nz34e.ws"
        self.assertEqual(cast_idna(None, value), value)

    def test_cast_idna_2003_only_2(self):
        value = "xn--21sud-kia.com"
        self.assertEqual(cast_idna(None, value), value)

    def test_cast_idna_2003_missing(self):
        value = "😀.ws"
        self.assertEqual(cast_idna(None, value), "xn--e28h.ws")

    def test_cast_idna_mixed(self):
        # regression test
        value = "ちんぽこ.xn--tckwe"
        self.assertEqual(cast_idna(None, value), "xn--28j1ayfyd.xn--tckwe")

    def test_cast_idna_robase(self):
        # idna >= 2.7 introduced a regression for which we have a workaround
        self.assertEqual(cast_idna(None, "@"), "@")

    def test_cast_idna_small_letter(self):
        value = "gand\u0131.net"  # gandı.net
        self.assertEqual(cast_idna(None, value), "xn--gand-oza.net")

    def test_cast_idna_invalid(self):
        # idna >= 2.7 introduced a regression for which we have a workaround
        value = "invalid##name"
        self.assertEqual(cast_idna(None, value), value)

    def test_email_validator(self):
        validator = email_validator()

        request = DummyRequest()
        node = colander.SchemaNode(String()).bind(request=request)

        for email in ("john@doe.engineering", "john@変態.co.jp"):
            self.assertIsNone(validator(node, email))

        for email in ("john@doe", "johndoe.engineering", 1234):
            self.assertRaises(colander.Invalid, validator, node, email)

    def test_validate_email_integer(self):
        self.assertFalse(validate_email(12345))

    def test_validate_email_no_at(self):
        self.assertFalse(validate_email("blabla.truc"))

    def test_validate_email_not_fqdn(self):
        self.assertFalse(validate_email("local@bla"))

    def test_validate_email_unicode(self):
        self.assertTrue(validate_email("john@変態.co.jp"))

    def test_validate_email_punycode(self):
        self.assertTrue(validate_email("local@xn--atsuk-8ra.fr"))

    def test_validate_email_normal(self):
        self.assertTrue(validate_email("truc@gandi.net"))

    def test_validate_email_tld_gt_8(self):
        self.assertTrue(validate_email("john@doe.engineering"))

    def test_validate_idna_validation(self):
        self.assertFalse(validate_email("yo@xn--assemble.fr"))
        self.assertTrue(validate_email("yo@xn--hllo-bpa.com"))
        self.assertTrue(validate_email("yo@héllo.com"))
        self.assertTrue(validate_email("yo@🍺.com"))

    def test_validate_email_spaces(self):
        self.assertFalse(validate_email("plop@bip.bop "))
        self.assertFalse(validate_email(" plop@bip.bop"))
        self.assertFalse(validate_email("plop@bip.b op"))
        self.assertFalse(validate_email("plo p@bip.bop"))

    def test_validate_email_localpart_special_characters(self):
        self.assertTrue(validate_email("ping#pong@bip.bop"))
        self.assertTrue(validate_email("ping!pong@bip.bop"))
        self.assertTrue(validate_email("ping%pong@bip.bop"))
        self.assertTrue(validate_email("ping~pong@bip.bop"))
        self.assertTrue(validate_email("ping.pong@bip.bop"))
        self.assertTrue(validate_email("ping|pong@bip.bop"))
        self.assertTrue(validate_email("ping?=pong@bip.bop"))
        self.assertTrue(validate_email("ping'pong'@bip.bop"))
        self.assertTrue(validate_email("ping+pong@bip.bop"))

    def test_StrippedString(self):
        deserialize = StrippedString().deserialize

        self.assertEqual(deserialize(None, "a"), "a")
        self.assertEqual(deserialize(None, "  a , b \t\n"), "a , b")

        self.assertEqual(
            deserialize(None, "le café\u00a0du\ncoin\n "), "le café du\ncoin"
        )

        self.assertEqual(
            deserialize(None, "le café\u202cdu\ncoin\n "), "le café du\ncoin"
        )

        self.assertEqual(deserialize(None, ""), colander.null)

        self.assertEqual(
            deserialize(None, "le café\u202cdu coin\u200b"), "le café du coin"
        )

        self.assertEqual(
            deserialize(None, " le café\u200bdu coin\u200b"), "le café\u200bdu coin"
        )

        self.assertEqual(
            deserialize(None, "\x00le café\x00du coin\x00\x00\x04"), "le café du coin"
        )

    def test_NullableSchemaNode(self):
        fake_type = mock.MagicMock()
        fake_type.deserialize.return_value = "apple"
        schema_node = NullableSchemaNode(fake_type)

        result = schema_node.deserialize(None)
        self.assertIsNone(result)
        self.assertIsNone(fake_type.deserialize.call_args)

        fake_type.reset()
        result = schema_node.deserialize(42)
        self.assertEqual(result, "apple")
        self.assertEqual(fake_type.deserialize.call_args, [(schema_node, 42)])

        fake_type.reset()
        result = schema_node.deserialize(colander.null)
        self.assertEqual(result, "apple")
        self.assertEqual(
            fake_type.deserialize.call_args,
            [(schema_node, colander.null)],
        )

        fake_type.reset()
        result = schema_node.deserialize()
        self.assertEqual(result, "apple")
        self.assertEqual(
            fake_type.deserialize.call_args,
            [(schema_node, colander.null)],
        )

    def test_sort_by_preparer(self):
        self.assertEqual(
            sort_by_preparer("field,field,-field"), ["field", "field", "-field"]
        )

    def test_sort_by_preparer_null(self):
        self.assertEqual(sort_by_preparer(""), "")

    def test_validate_email_localpart_ge_255(self):
        hostpart = "t" * 255
        self.assertFalse(validate_email("ping#pong@%s.bop" % hostpart))

    def test_cast_idna_ip_allow_numbers(self):
        value = "192.168.0.1"
        self.assertEqual(cast_idna(None, value, allow_numbers=True), value)

    def test_host_numbers(self):
        fqdn = FqdnType()

        good_host = "192"

        with self.assertRaises(colander.Invalid):
            fqdn.deserialize(None, good_host)

    def test_host_numbers_allow_number_true(self):
        fqdn = FqdnType()
        fqdn.allow_numbers = True

        good_host = "192"

        self.assertEqual(fqdn.deserialize(None, good_host), good_host)

    def test_ContactType(self):
        deserialize = ContactType().deserialize

        for val in ("a", 0, "0", "blabla"):
            self.assertEqual(deserialize(None, val), val)

        self.assertRaises(colander.Invalid, deserialize, None, 4)
        self.assertRaises(colander.Invalid, deserialize, None, "reseller")


class TestSQLAlchemyMapping(TestCommon, TestCase):
    def test_render_do_not_evaluate_relations(self):
        with transaction.manager:
            dummy = Dummy(name="Xian Ping")
            mini = MiniDummy(name="Fou Mi", dummy=dummy)
            self.session.add(dummy)
            self.session.add(mini)
            self.session.flush()

            dummy_id = dummy.id

        # Identity map should be empty after transaction has been closed
        self.assertEqual(len(self.session.identity_map), 0)

        # Double check that the relation is honored
        with transaction.manager:
            dummy = Dummy.by_id(dummy_id)

            # Just make sure relations aren't eager loaded
            self.assertEqual(len(self.session.identity_map), 1)

            minis = dummy.minis
            self.assertEqual(len(minis), 1)
            self.assertEqual(len(self.session.identity_map), 2)

        with transaction.manager:
            data = [Dummy.by_id(dummy_id)]

            res = DummySequenceReturn().deserialize(data)

            self.assertEqual(res, [{"name": "Xian Ping", "lname": "xian ping"}])

            # The identity map should only have one object (the dummy), and
            # *not* it's MiniDummy relation
            # Otherwise, that means the renderer did evaluate the relation, and
            # will trigger useless selects in production
            self.assertEqual(
                len(self.session.identity_map),
                1,
                "Identity map should only contain one object",
            )


class TestLocalizedString(TestCase):
    def test_ok(self):
        def translate(tstring, domain=None, mapping=None):
            return "Localized String"

        with mock.patch("pyramid.i18n.Localizer.translate", translate):
            request = DummyRequest()
            node = colander.SchemaNode(LocalizedString()).bind(request=request)

            msg = "to localized"
            self.assertEqual(translate(msg), node.deserialize(msg))

    def test_not_a_string(self):
        request = DummyRequest()
        node = colander.SchemaNode(LocalizedString()).bind(request=request)

        with self.assertRaises(colander.Invalid):
            node.deserialize(42)


class PostSchemaOk(colander.MappingSchema):
    @colander.instantiate(name="optional", missing=nested_missing)
    class Optional(colander.MappingSchema):
        sharing_id = colander.SchemaNode(UUID(), missing=colander.drop)
        currency = colander.SchemaNode(colander.String(), missing="EUR")

        @colander.instantiate(name="duration", missing=nested_missing)
        class Duration(colander.MappingSchema):
            unit = colander.SchemaNode(colander.String(), missing="d")
            value = colander.SchemaNode(colander.Integer(), missing=42)


class PostSchemaError(colander.MappingSchema):
    @colander.instantiate(name="optional", missing=nested_missing)
    class Optional(colander.SequenceSchema):
        currency = colander.SchemaNode(colander.String(), missing="EUR")


class TestAutoMissing(TestCase):
    def test_missing(self):
        expected = {
            "optional": {
                "currency": "EUR",
                "duration": {
                    "unit": "d",
                    "value": 42,
                },
            },
        }
        res = PostSchemaOk().bind(request=DummyRequest()).deserialize({})
        self.assertEqual(res, expected)

    def test_full(self):
        data = {
            "optional": {
                "currency": "USD",
                "duration": {
                    "unit": "m",
                    "value": 24,
                },
            },
        }
        expected = data
        res = PostSchemaOk().bind(request=DummyRequest()).deserialize(data)
        self.assertEqual(res, expected)

    def test_sub_node(self):
        data = {
            "optional": {
                "currency": "USD",
            },
        }
        expected = {
            "optional": {
                "currency": "USD",
                "duration": {
                    "unit": "d",
                    "value": 42,
                },
            },
        }
        res = PostSchemaOk().bind(request=DummyRequest()).deserialize(data)
        self.assertEqual(res, expected)

    def test_root_node(self):
        data = {
            "optional": {
                "duration": {
                    "unit": "d",
                    "value": 42,
                }
            },
        }
        expected = {
            "optional": {
                "currency": "EUR",
                "duration": {
                    "unit": "d",
                    "value": 42,
                },
            },
        }
        res = PostSchemaOk().bind(request=DummyRequest()).deserialize(data)
        self.assertEqual(res, expected)

    def test_runtime_error(self):
        with self.assertRaises(RuntimeError) as err:
            PostSchemaError().bind(request=DummyRequest()).deserialize({})
        self.assertEqual(err.exception.args, ("node has to be a MappingSchema",))
