# coding: utf-8

from unittest import TestCase

from pyramid.testing import DummyRequest

from sequoia.base3.helpers import cornice


class Request(object):
    def __init__(self, team_id):
        self.validated = {"body": {"team_id": team_id}, "extra": {}}


# Do not call check function here, done by colander.
def load_team(request, **kwargs):
    params = request.validated["body"]
    n = params.get("team_id", None)
    team = list(range(10 * n)) if n else None
    if team:
        params["team"] = team


@cornice.required_params("team")
def load_members2(request, **kwargs):
    params = request.validated["body"]
    params["members"] = ["%d" % d for d in params["team"]]


@cornice.fetch_required()
def load_team_members(request, team, **kwargs):
    members = ["%d" % d for d in team]
    request.validated["body"]["members"] = members


@cornice.fetch_request_params()
def load_a_default_team(request, team, **kwargs):
    if team is None:
        request.validated["extra"]["no_team"] = True
        return
    members = ["%d" % d for d in team]
    request.validated["body"]["members"] = members


@cornice.fetch_location
def update_extra(request, extra, **kwargs):
    extra["is_called"] = True


@cornice.fetch_location
def update_extra_missing_body(request, body, extra, **kwargs):
    extra["is_called"] = True


@cornice.fetch_required()
def params_with_defaults(request, a_non_required_parameter=None, **kwargs):
    request.validated["extra"]["val"] = a_non_required_parameter


@cornice.fetch_required()
def params_with_false_value(request, falsee, **kwargs):
    request.validated["extra"]["val"] = falsee


class ValidatorTestCase(TestCase):
    def test_demo(self):
        request_ok = Request(1)
        request_noteamid = Request(0)

        validators_ = [load_team, load_members2]

        for validator in validators_:
            validator(request_ok)
            validator(request_noteamid)

        self.assertIn("members", request_ok.validated["body"])
        self.assertNotIn("members", request_noteamid.validated["body"])

    def test_fetch_required(self):
        request_ok = Request(1)
        request_noteamid = Request(0)

        validators_ = [load_team, load_team_members]

        for validator in validators_:
            validator(request_ok)
            validator(request_noteamid)

        self.assertIn("members", request_ok.validated["body"])
        self.assertNotIn("members", request_noteamid.validated["body"])

    def test_fetch_request_params(self):
        request_ok = Request(1)
        request_noteamid = Request(0)
        validators_ = [load_team, load_a_default_team]

        for validator in validators_:
            validator(request_ok)
            validator(request_noteamid)

        self.assertEqual(
            request_ok.validated,
            {
                "body": {
                    "members": [
                        "0",
                        "1",
                        "2",
                        "3",
                        "4",
                        "5",
                        "6",
                        "7",
                        "8",
                        "9",
                    ],
                    "team": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                    "team_id": 1,
                },
                "extra": {},
            },
        )
        self.assertEqual(
            request_noteamid.validated,
            {"body": {"team_id": 0}, "extra": {"no_team": True}},
        )

    def test_fetch_required_params_with_false_value(self):
        """
        Default value can be a falsee value (False, 0, ...)

        In that case, it must be consider provided
        """
        request = Request(1)
        request.validated = {"body": {"falsee": False}, "extra": {}}
        params_with_false_value(request)  # type: ignore
        self.assertEqual(request.validated["extra"], {"val": False})

        # if none, do not call the decorated validator
        request.validated = {"body": {}, "extra": {}}
        params_with_false_value(request)  # type: ignore
        self.assertEqual(request.validated["extra"], {})

    def test_fetch_required_default_non_manadatory_is_none(self):
        """
        fetch_required can have not mandatory parameters,

        using a default value
        """
        request = Request(1)
        request.validated = {"body": {}, "extra": {}}
        params_with_defaults(request)
        self.assertEqual(request.validated["extra"], {"val": None})

    def test_fetch_required_default_non_manadatory_take_given_value(self):
        """
        fetch_required can have not mandatory parameters,

        using a default value
        """
        request = Request(1)
        request.validated = {"body": {"a_non_required_parameter": "value"}, "extra": {}}
        params_with_defaults(request)
        self.assertEqual(request.validated["extra"], {"val": "value"})

    def test_default_not_required(self):
        request = Request(1)
        request.validated = {"body": {}, "extra": {}}
        params_with_defaults(request)
        self.assertEqual(request.validated["extra"], {"val": None})

    def test_missing_location(self):
        request = Request(1)
        request.validated = {}  # drop the 'body' then ensure validators
        # will not raise KeyError, decorator will skip them

        validators_ = [
            load_members2,
            load_team_members,
        ]

        for validator in validators_:
            validator(request)

        self.assertEqual(request.validated, {})

    def test_fetch_location(self):
        request = DummyRequest()
        request.validated = {"extra": {}}  # type: ignore
        update_extra(request)  # type: ignore
        self.assertEqual(
            request.validated,  # type: ignore
            {"extra": {"is_called": True}},
        )

    def test_fetch_location_missing_body(self):
        request = DummyRequest()
        request.validated = {"extra": {}}  # type: ignore
        update_extra_missing_body(request)  # type: ignore
        self.assertEqual(
            request.validated,  # type: ignore
            {"extra": {"is_called": True}},
        )
