from typing import Any, Mapping, Optional, Sequence

import pytest
from pydantic import (
    BaseModel,
    EnumMemberError,
    IPvAnyAddress,
    IPvAnyAddressError,
    IPvAnyNetwork,
    MissingError,
    validator,
)
from pydantic.error_wrappers import ErrorWrapper
from webob.multidict import MultiDict

from sequoia.base3.helpers.pydantic.pagination import PaginationSchema
from sequoia.base3.helpers.pydantic.request_validator import (
    extract_val_from_request,
    translate_err,
    validate_pydantic_schema,
)
from sequoia.base3.helpers.pydantic.types import UUID
from sequoia.base3.helpers.pydantic.typing import (
    HeaderField,
    PathInfoField,
    PostBodyField,
    QueryStringField,
)


class MyRequestSchema(BaseModel):
    name: str = PathInfoField()
    x_country_code: Optional[str] = HeaderField(alias="x-country-code")
    gt_age: int = QueryStringField(alias=">age", ge=18, le=100)
    super_powers: Optional[Sequence[str]] = PostBodyField(default_factory=list)

    @validator("name")
    def validate_name(cls, v: Any) -> str:
        v = v.lower()
        if v == "admin":
            raise ValueError("admin is a reserved name")
        return v


class CustomTypesSchema(BaseModel):
    id: UUID = QueryStringField()
    ip_network: IPvAnyNetwork = QueryStringField()
    ip: IPvAnyAddress = QueryStringField()


request = {
    "matchdict": {"nick": "bob", "name": "marley"},
    "params": {"nick": "alice"},
    "headers": {"nick": "peter"},
    "body": {"nick": "bonny"},
}


@pytest.mark.parametrize(
    "params",
    [
        pytest.param(
            {
                "request": request,
                "location": "path",
                "name": "nick",
                "is_scalar": False,
                "expected": ("bob", False),
            },
            id="from path",
        ),
        pytest.param(
            {
                "request": request,
                "location": "querystring",
                "name": "nick",
                "is_scalar": False,
                "expected": ("alice", False),
            },
            id="from querystring",
        ),
        pytest.param(
            {
                "request": {
                    "GET": MultiDict([("nick", "alice"), ("nick", "bob")]),
                },
                "location": "querystring",
                "name": "nick",
                "is_scalar": True,
                "expected": (["alice", "bob"], False),
            },
            id="from querystring",
        ),
        pytest.param(
            {
                "request": request,
                "location": "header",
                "name": "nick",
                "is_scalar": False,
                "expected": ("peter", False),
            },
            id="from header",
        ),
        pytest.param(
            {
                "request": request,
                "location": "body",
                "name": "nick",
                "is_scalar": False,
                "expected": ("bonny", False),
            },
            id="from body",
        ),
        pytest.param(
            {
                "request": {"body": {}},
                "location": "body",
                "name": "nick",
                "is_scalar": False,
                "expected": (None, True),
            },
            id="null value product a missing value",
        ),
        pytest.param(
            {
                "request": {"body": {"nick": None}},
                "location": "body",
                "name": "nick",
                "is_scalar": False,
                "expected": (None, False),
            },
            id="None value is a missing value",
        ),
        pytest.param(
            {
                "request": {"body": None},
                "location": "body",
                "name": "nick",
                "is_scalar": False,
                "expected": (None, False),
            },
            id="None json document is a missing value",
        ),
    ],
)
def test_extract_val_from_request(params, dummy_request):
    val = extract_val_from_request(
        dummy_request, params["location"], params["name"], params["is_scalar"]
    )
    assert val == params["expected"]


@pytest.mark.parametrize("params", [{"request": {"body": {}}}])
def test_extract_val_from_request_raise_explicit_error(dummy_request):
    with pytest.raises(ValueError) as ctx:
        extract_val_from_request(dummy_request, "cookie", "~", False)
    assert str(ctx.value) == "Invalid location: cookie"


trans = {
    "field required": "Champs requis",
    "value is not a valid enumeration member; permitted: ${permitted}": ""
    "Valeur invalide; valeurs permises: {enum_values!r}",
}


@pytest.mark.parametrize(
    "params",
    [
        pytest.param(
            {
                "translations": trans,
                "expected": "Champs requis",
                "error": MissingError(),
            },
            id="type error translation",
        ),
        pytest.param(
            {
                "translations": trans,
                "expected": "Valeur invalide; valeurs permises: ['foo', 'bar']",
                "error": EnumMemberError(enum_values=["foo", "bar"]),
            },
            id="type error translation",
        ),
        pytest.param(
            {
                "translations": trans,
                "expected": "value is not a valid IPv4 or IPv6 address",
                "error": IPvAnyAddressError(),
            },
            id="test msg_template error",
        ),
        pytest.param(
            {
                "translations": trans,
                "expected": "Fréro is offensing, read the term of use",
                "error": ValueError("Fréro is offensing, read the term of use"),
            },
            id="test fallback on missing translation",
        ),
    ],
)
def test_translate_err(params, localizer):
    err = ErrorWrapper(params["error"], ())
    l10n_err = translate_err(localizer, err)
    assert l10n_err == params["expected"]


@pytest.mark.parametrize(
    "params",
    [
        pytest.param(
            {
                "schema": MyRequestSchema,
                "request": {
                    "matchdict": {"name": "bob"},
                    "params": {">age": "dix-huit"},
                    "body": {"super_powers": ["singer"]},
                },
                "expected_validated": {
                    "body": {"super_powers": ["singer"]},
                    "header": {},
                    "path": {"name": "bob"},
                    "querystring": {},
                },
                "expected_errors": [
                    {
                        "description": "value is not a valid integer",
                        "location": "querystring",
                        "name": ">age",
                    }
                ],
            },
            id="partial schema error, type error",
        ),
        pytest.param(
            {
                "schema": MyRequestSchema,
                "request": {
                    "matchdict": {"name": "bob"},
                    "params": {">age": 8},
                    "body": {"super_powers": ["singer"]},
                },
                "expected_errors": [
                    {
                        "description": "ensure this value is greater than or equal to "
                        "18",
                        "location": "querystring",
                        "name": ">age",
                    }
                ],
                "expected_validated": {
                    "body": {"super_powers": ["singer"]},
                    "header": {},
                    "path": {"name": "bob"},
                    "querystring": {},
                },
            },
            id="range error",
        ),
        pytest.param(
            {
                "schema": MyRequestSchema,
                "request": {"body": {}},
                "expected_errors": [
                    {
                        "description": "field required",
                        "location": "path",
                        "name": "name",
                    },
                    {
                        "description": "field required",
                        "location": "querystring",
                        "name": ">age",
                    },
                ],
                "expected_validated": {
                    "body": {"super_powers": []},
                    "header": {},
                    "path": {},
                    "querystring": {},
                },
            },
            id="missing error, with default_factory",
        ),
        pytest.param(
            {
                "schema": MyRequestSchema,
                "request": {
                    "matchdict": {"name": "ADMIN"},
                    "params": {">age": 34},
                    "body": {},
                },
                "expected_errors": [
                    {
                        "description": "admin is a reserved name",
                        "location": "path",
                        "name": "name",
                    },
                ],
                "expected_validated": {
                    "body": {"super_powers": []},
                    "header": {},
                    "path": {},
                    "querystring": {"gt_age": 34},
                },
            },
            id="missing error, with default_factory",
        ),
        pytest.param(
            {
                "schema": CustomTypesSchema,
                "request": {
                    "params": {"id": "yolo", "ip_network": "1.2.3.4/5", "ip": "ff"}
                },
                "expected_errors": [
                    {
                        "description": "uuid format expected",
                        "location": "querystring",
                        "name": "id",
                    },
                    {
                        "description": "value is not a valid IPv4 or IPv6 network",
                        "location": "querystring",
                        "name": "ip_network",
                    },
                    {
                        "description": "value is not a valid IPv4 or IPv6 address",
                        "location": "querystring",
                        "name": "ip",
                    },
                ],
                "expected_validated": {
                    "body": {},
                    "header": {},
                    "path": {},
                    "querystring": {},
                },
            },
            id="Custosm schema,",
        ),
    ],
)
def test_validate_pydantic_schema(params: Mapping[str, Any], dummy_request):
    validate_pydantic_schema(dummy_request, params["schema"])

    assert dummy_request.errors == params["expected_errors"]
    assert dummy_request.validated == params["expected_validated"]


@pytest.mark.parametrize(
    "params",
    [
        pytest.param(
            {
                "schema": MyRequestSchema,
                "request": {
                    "matchdict": {"name": "bob"},
                    "params": {">age": 34},
                    "body": {"super_powers": ["singer"]},
                },
                "expected_validated": {
                    "body": {"super_powers": ["singer"]},
                    "header": {},
                    "path": {"name": "bob"},
                    "querystring": {"gt_age": 34},
                },
            },
            id="partial type (x-country-code is not present in the request)",
        ),
        pytest.param(
            {
                "schema": MyRequestSchema,
                "request": {
                    "matchdict": {"name": "BOB"},
                    "params": {">age": 34},
                    "body": {"super_powers": ["singer"]},
                },
                "expected_validated": {
                    "body": {"super_powers": ["singer"]},
                    "header": {},
                    "path": {"name": "bob"},
                    "querystring": {"gt_age": 34},
                },
            },
            id="validator is also a preparer",
        ),
        pytest.param(
            {
                "schema": PaginationSchema,
                "request": {},
                "expected_validated": {
                    "body": {},
                    "header": {},
                    "path": {},
                    "querystring": {"page": 1, "sort_by": "id", "per_page": 50},
                },
            }
        ),
    ],
)
def test_validate_pydantic_schema_ok(params: Mapping[str, Any], dummy_request):
    validate_pydantic_schema(dummy_request, params["schema"])
    assert dummy_request.errors == []
    assert dummy_request.validated == params["expected_validated"]
