# -*- coding: utf-8 -*-
import os.path
import uuid
from unittest.mock import patch

from freezegun import freeze_time
from pyramid.response import Response

from sequoia.base3.testing.case import ApiTestCase


def view_extra_route(request):
    return Response(request.route_url("another_app", path="to", _query={"the": "next"}))


def view_remote_addr(request):
    # echo client address back at him
    return Response(text=str(request.client_addr))


def includeme(config):
    config.add_route("another_app", "http://show.me/the/way/{path}")
    config.add_route("view_extra_route", "/view_extra_route")
    config.add_view(view_extra_route, route_name="view_extra_route")
    config.add_route("view_remote_addr", "/view_remote_addr")
    config.add_view(view_remote_addr, route_name="view_remote_addr")


class TestClientAddress(ApiTestCase):
    """
    Test that the request.clientadd
    """

    app_settings = {
        "sequoia.export": [],
        "pyramid.includes": "sequoia.base3.tests.helpers.test_pyramid",
    }

    def patch_environ(self, new_values):
        """
        Patch the environ variables used for testing

        Hopefully the variable is shared between self and the app

        """
        # Fail test if any of those values are already defined
        forbidden = {"REMOTE_ADDR", "HTTP_X_REAL_IP", "HTTP_X_FORWARDED_FOR"}
        self.assertFalse(forbidden & set(self.extra_environ))

        return patch.dict(self.extra_environ, new_values)

    def test_remote_only(self):
        with self.patch_environ({"REMOTE_ADDR": "1.1.1.1"}):
            res = self.app.get("/view_remote_addr")
        self.assertEqual(res.text, "1.1.1.1")

    def test_no_real_ip(self):
        value = {"REMOTE_ADDR": "1.1.1.1", "HTTP_X_FORWARDED_FOR": "2.2.2.2"}
        with self.patch_environ(value):
            res = self.app.get("/view_remote_addr")
        self.assertEqual(res.text, "2.2.2.2")

    def test_use_real_ip(self):
        value = {
            "REMOTE_ADDR": "1.1.1.1",
            "HTTP_X_FORWARDED_FOR": "2.2.2.2",
            "HTTP_X_REAL_IP": "3.3.3.3",
        }
        with self.patch_environ(value):
            res = self.app.get("/view_remote_addr")
        self.assertEqual(res.text, "3.3.3.3")


class HateoasRouteUrlTestCase(ApiTestCase):
    app_settings = {
        "sequoia.export": "sequoia.base3.tests.api.dummy",
        "sequoia.hosts.Dummy": "dummies.local.gandi.net",
        "pyramid.includes": "sequoia.base3.tests.helpers.test_pyramid",
    }

    def test_render_default(self):
        resp = self.app.post_json("/dummies", params={"name": "Nkishi"})

        self.assertEqual(
            resp.json["href"], "http://dummies.local.gandi.net/dummies/Nkishi"
        )

    def test_render_x_forwarded_proto(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Nkishi"},
            headers={"X-Forwarded-Proto": "https"},
        )

        self.assertEqual(
            resp.json["href"], "https://dummies.local.gandi.net/dummies/Nkishi"
        )

    def test_render_hateoas_prefix(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={"Hateoas-Prefix": "https://fetish.gandi.net/api"},
        )
        self.assertEqual(
            resp.json["href"], "https://fetish.gandi.net/api/dummies/Songye"
        )

    def test_render_hateoas_prefix_no_path(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={"Hateoas-Prefix": "https://fetish.gandi.net"},
        )
        self.assertEqual(resp.json["href"], "https://fetish.gandi.net/dummies/Songye")

    def test_render_hateoas_prefix_slash_path(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={"Hateoas-Prefix": "https://fetish.gandi.net/"},
        )
        self.assertEqual(resp.json["href"], "https://fetish.gandi.net/dummies/Songye")

    def test_render_hateoas_with_port(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={"Hateoas-Prefix": "http://fetish.gandi.net:8888/api"},
        )
        self.assertEqual(
            resp.json["href"], "http://fetish.gandi.net:8888/api/dummies/Songye"
        )

    def test_render_hateoas_prefix_replace(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={
                "Hateoas-Prefix": "https://dummies.gandi.net/",
                "Hateoas-Replace": "/dummies",
            },
        )
        self.assertEqual(resp.json["href"], "https://dummies.gandi.net/Songye")

    def test_render_hateoas_prefix_replace_invalid(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={
                "Hateoas-Prefix": "https://dummies.gandi.net/",
                "Hateoas-Replace": "not-in-route",
            },
        )
        self.assertEqual(resp.json["href"], "https://dummies.gandi.net/dummies/Songye")

    def test_render_hateoas_prefix_replace_once(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "dummies"},
            headers={
                "Hateoas-Prefix": "https://dummies.gandi.net/",
                "Hateoas-Replace": "/dummies",
            },
        )
        self.assertEqual(resp.json["href"], "https://dummies.gandi.net/dummies")

    def test_render_with_get_url(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "get_url"},
            extra_environ={"HTTP_HOST": "banana"},
        )
        self.assertEqual(
            resp.json["href"], "http://dummies.local.gandi.net/dummies/get_url"
        )

    def test_extra_route(self):
        resp = self.app.get("/view_extra_route")
        self.assertEqual(resp.text, "http://show.me/the/way/to?the=next")


class HostSchemeForcedSettingRouteUrlTestCase(ApiTestCase):
    app_settings = {
        "sequoia.export": "sequoia.base3.tests.api.dummy",
        "sequoia.hosts.Dummy": "dummies.local.gandi.net",
        "sequoia.hosts.Dummy.scheme": "https",
    }

    def test_render_default(self):
        resp = self.app.post_json("/dummies", params={"name": "Nkishi"})

        self.assertEqual(
            resp.json["href"], "https://dummies.local.gandi.net/dummies/Nkishi"
        )


class NoHostsSettingsRouteUrlTestCase(ApiTestCase):
    app_settings = {"sequoia.export": "sequoia.base3.tests.api.dummy"}

    def test_render_default(self):
        resp = self.app.post_json(
            "/dummies", params={"name": "Nkishi"}, extra_environ={"HTTP_HOST": "banana"}
        )

        self.assertEqual(resp.json["href"], "http://banana/dummies/Nkishi")

    def test_render_default_port(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Nkishi"},
            extra_environ={"HTTP_HOST": "banana:5000"},
        )

        self.assertEqual(resp.json["href"], "http://banana:5000/dummies/Nkishi")

    def test_render_hateoas_prefix(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={"Hateoas-Prefix": "https://fetish.gandi.net/api"},
        )

        self.assertEqual(
            resp.json["href"], "https://fetish.gandi.net/api/dummies/Songye"
        )

    def test_render_hateoas_prefix_path_trailing_slash(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "Songye"},
            headers={
                "Hateoas-Prefix": "https://fetish.gandi.net/foo/",
                "Hateoas-Replace": "/dummies/",
            },
        )
        self.assertEqual(resp.json["href"], "https://fetish.gandi.net/foo/Songye")

    def test_render_internal_with_hateoas_prefix(self):
        resp = self.app.post_json(
            "/dummies",
            params={"name": "intern"},
            headers={
                "Hateoas-Prefix": "https://fetish.gandi.net/foo/",
                "Hateoas-Replace": "/dummies/",
            },
        )

        self.assertEqual(resp.json["href"], "https://fetish.gandi.net/foo/intern")
        self.assertEqual(resp.json["internal_href"], "http://localhost/dummies/intern")


class TestHttpCacheHeaders(ApiTestCase):
    here = os.path.dirname(__file__)

    app_settings = {
        "sequoia.export": "sequoia.base3.tests.api.dummy",
        "pyramid.includes": "\n  ".join(
            [
                "sequoia.auth.testing",
            ]
        ),
        "sequoia.auth.testing.data": "%s/../api/authenticated_users.json" % here,
    }

    @freeze_time("1999-06-01 13:37:00")
    def test_no_http_cache_view_predicate(self):
        resp = self.app.get("/pages")
        self.assertEqual(
            resp.headers["Pragma"],
            "no-cache",
            "Missing HTTP/1.0 Pragma header to disable caching",
        )
        self.assertEqual(
            resp.headers["Cache-Control"],
            "max-age=0, must-revalidate, no-cache, no-store",
            "Missing HTTP/1.1 Cache-Control header to disable caching",
        )

        self.assertEqual(resp.headers["Expires"], "Tue, 01 Jun 1999 13:37:00 GMT")

        self.assertEqual(resp.headers["Expires"], resp.headers["Last-Modified"])

    def test_http_cache_header_not_applied_on_400_errors(self):
        resp = self.app.get("/unprocessable_entity/400", expect_errors=True)

        self.assertEqual(
            resp.headers["Cache-Control"],
            "max-age=0, must-revalidate, no-cache, no-store",
        )

    def test_http_cache_header_applied_on_500_errors(self):
        resp = self.app.get("/unprocessable_entity/boom", expect_errors=True)

        self.assertEqual(
            resp.headers["Cache-Control"],
            "max-age=0, must-revalidate, no-cache, no-store",
        )

    def test_http_cache_header_not_applied_on_403_errors(self):
        resp = self.app.get(
            "/protected", status=403, headers={"X-Auth-Testing": "noperm"}
        )
        self.assertEqual(
            resp.headers["Cache-Control"],
            "max-age=0, must-revalidate, no-cache, no-store",
        )

    def test_http_cache_header_not_applied_on_401_errors(self):
        resp = self.app.get("/protected", status=401)
        self.assertEqual(
            resp.headers["Cache-Control"],
            "max-age=0, must-revalidate, no-cache, no-store",
        )

    def test_http_cache_header_applied_on_public_with_vary(self):
        resp = self.app.get("/pages/007")
        self.assertEqual(resp.headers["Cache-Control"], "max-age=60, public")
        self.assertEqual(resp.headers["Vary"], "X-Variant-1, X-Variant-2")

    def test_vary_as_str(self):
        resp = self.app.get("/dummies/teddy")
        self.assertEqual(resp.headers["Cache-Control"], "max-age=60")
        self.assertEqual(resp.headers["Vary"], "X-Variant")

    def test_http_cache_header_applied_on_protected_no_vary(self):
        resp = self.app.get("/protected", headers={"X-Auth-Testing": "permitted"})
        self.assertEqual(resp.headers["Cache-Control"], "max-age=60, private")
        self.assertNotIn("Vary", resp.headers)


class TestDocumentationPredicate(ApiTestCase):
    app_settings = {"sequoia.export": "sequoia.base3.tests.api.dummy"}

    def test_documentation_predicate(self):
        srv = self.app.app.registry.cornice_services["/documented"]
        definition = [x for x in srv.definitions if x[0] == "GET"][0]
        self.assertEqual(definition[2]["documentation"], {"status_code": 201})


class TestCustomExtractCStruct(ApiTestCase):
    here = os.path.dirname(__file__)

    app_settings = {
        "sequoia.export": "sequoia.base3.tests.api.dummy",
        "pyramid.includes": "\n  ".join(
            [
                "sequoia.auth.testing",
            ]
        ),
        "sequoia.auth.testing.data": "%s/../api/authenticated_users.json" % here,
    }

    def test_extract_cstruct_multipart(self):
        sharing_id = str(uuid.uuid4())
        query = '{"sharing_id": "%s", "template": "YYY"}' % sharing_id
        resp = self.app.post(
            "/test-multiparts",
            content_type="multipart/form-data",
            upload_files=[
                (
                    "query",
                    "whatever.json",
                    query.encode("utf-8"),
                ),
            ],
        )
        self.assertEqual(resp.text, sharing_id)

    def test_extract_cstruct_multipart_422(self):
        resp = self.app.post(
            "/test-multiparts",
            content_type="multipart/form-data",
            upload_files=[
                (
                    "query",
                    "whatever.json",
                    b'{"sharing_id": "XXX"}',
                ),
            ],
            status=422,
        )

        self.assertEqual(
            resp.json,
            {
                "cause": "Unprocessable Entity",
                "code": 422,
                "errors": [
                    {"description": "Required", "location": "body", "name": "template"},
                    {
                        "description": '"XXX" is not a UUID',
                        "location": "body",
                        "name": "sharing_id",
                    },
                ],
                "message": "Unable to process the contained instructions",
                "object": "HTTPUnprocessableEntity",
            },
        )

    def test_extract_cstruct_multipart_400_bad_json(self):
        resp = self.app.post(
            "/test-multiparts",
            content_type="multipart/form-data",
            upload_files=[
                (
                    "query",
                    "whatever.json",
                    b'{"{',
                ),
            ],
            status=400,
        )
        self.assertEqual(
            resp.json,
            {
                "cause": "Bad Request",
                "code": 400,
                "message": (
                    "The server could not comply with the request since it is either "
                    "malformed or otherwise incorrect."
                ),
                "object": "HTTPBadRequest",
            },
        )
