import datetime
import decimal
import enum
import json
import uuid
from collections import OrderedDict
from unittest import TestCase, mock

import colander
import transaction

from sequoia.base3.helpers.colander import ListReturn, UUIDHTTPCreatedReturn
from sequoia.base3.helpers.renderer import JSONEncoder, JsonRenderer, ResponseBuilder

from .sqlalchemy import AddOnDummy, Dummy, MiniDummy, TestCommon


class DummyStrEnum(enum.Enum):
    red = "red"
    green = "green"
    blue = "blue"


class SimpleTypeSchema(colander.MappingSchema):
    int_ = colander.SchemaNode(colander.Integer(), name="int")
    float_ = colander.SchemaNode(colander.Float(), name="float")
    decimal_ = colander.SchemaNode(colander.Decimal(), name="decimal")
    string = colander.SchemaNode(colander.String())
    bool_ = colander.SchemaNode(colander.Boolean(), name="bool")
    date = colander.SchemaNode(colander.Date())
    datetime_ = colander.SchemaNode(colander.DateTime(), name="datetime")
    enum_str = colander.SchemaNode(colander.Enum(DummyStrEnum, attr="value"))


class ComplexTypeSchema(colander.MappingSchema):
    name = colander.SchemaNode(colander.String())
    age = colander.SchemaNode(colander.Integer(), missing=colander.drop)

    simples = SimpleTypeSchema(missing={"bool": True})

    @colander.instantiate(name="expertises", missing=colander.drop)
    class Expertise(colander.SequenceSchema):
        _ = colander.SchemaNode(colander.String())

    @colander.instantiate(name="hobbies", missing=colander.drop)
    class Hobbies(colander.SequenceSchema):
        @colander.instantiate()
        class Hobby(colander.MappingSchema):
            name = colander.SchemaNode(colander.String())
            level = colander.SchemaNode(colander.String(), missing=colander.drop)


class DeseserializedOverride(colander.MappingSchema):
    name = colander.SchemaNode(colander.String(), missing="egg")
    happy = colander.SchemaNode(colander.Boolean(), missing=False)

    def deserialize(self, cstruct):
        return {"happy": True}


class DeseserializedSequenceOverride(ListReturn):
    @colander.instantiate(name="body")
    class Items(colander.SequenceSchema):
        @colander.instantiate(name="body")
        class Item(colander.MappingSchema):
            name = colander.SchemaNode(colander.String(), missing=colander.drop)
            happy = colander.SchemaNode(colander.Boolean(), missing=False)

        def deserialize(self, cstruct):
            return [{"happy": True}]


class ColanderMissingCompat(colander.MappingSchema):
    name = colander.SchemaNode(colander.String(), missing="egg")
    happy = colander.SchemaNode(colander.Boolean(), missing=colander.drop)

    @colander.instantiate(name="preferred_item", missing={"name": "shuriken"})
    class Item(colander.MappingSchema):
        name = colander.SchemaNode(colander.String(), missing=colander.drop)

    @colander.instantiate(
        name="items", missing=[{"name": "shuriken"}, {"name": "nunchaku"}]
    )
    class Items(colander.SequenceSchema):
        @colander.instantiate()
        class Item(colander.MappingSchema):
            name = colander.SchemaNode(colander.String(), missing=colander.drop)

    @colander.instantiate(name="dropped_field", missing=colander.drop)
    class DroppedItems(colander.MappingSchema):
        name = colander.SchemaNode(colander.String())


class ColanderTitlelizeHeaders(colander.MappingSchema):
    foo = colander.SchemaNode(colander.String(), location="header")
    bar = colander.SchemaNode(colander.String(), name="foo-bar-corp", location="header")
    plouf = colander.SchemaNode(
        colander.String(), name="Plouf-Plouf", location="header"
    )


class JsonSerializerTestCase(TestCase):
    maxDiff = None

    def test_build_simple_types(self):
        builder = ResponseBuilder(SimpleTypeSchema())
        data = {
            "int": 1,
            "float": 2.5,
            "decimal": decimal.Decimal("1.3"),
            "string": "1",
            "bool": True,
            # WRONG should be datetime here!
            "date": "2015-01-15",
            "datetime": "2015-04-15T11:10:00Z",
            # 'date': datetime.date(2015, 1, 15),
            # 'datetime': datetime.datetime(2015, 4, 15, 11, 10, 0),
            "enum_str": DummyStrEnum.green,
        }

        resp = builder.build(data)
        self.assertEqual(builder.errors, [])
        self.assertEqual(
            resp,
            {
                "body": {
                    "bool": True,
                    "float": 2.5,
                    "decimal": decimal.Decimal("1.3"),
                    "int": 1,
                    "string": "1",
                    # WRONG, should be string here
                    "date": datetime.date(2015, 1, 15),
                    "datetime": datetime.datetime(
                        2015, 4, 15, 11, 10, 0, tzinfo=datetime.timezone.utc
                    ),
                    # 'date': '2015-01-15',
                    # 'datetime': '2015-04-15T11:10:00+00:00',
                    "enum_str": "green",
                },
                "header": {},
                "status_code": 200,
            },
        )

    def test_build_complex_type(self):
        self.maxDiff = None
        builder = ResponseBuilder(ComplexTypeSchema())
        data = {
            "name": "bob",
            "expertises": ["guitar"],
            "hobbies": [{"name": "footbal"}],
        }
        resp = builder.build(data)
        data["simples"] = {"bool": True}
        self.assertEqual(builder.errors, [])
        self.assertEqual(resp, {"body": data, "header": {}, "status_code": 200})

    def test_build_complex_type_with_errors(self):
        self.maxDiff = None
        builder = ResponseBuilder(ComplexTypeSchema())
        data = {
            "name": "bob",
            "hobbies": [{"name": "guitar", "level": "noob"}, {"level": "expert"}],
            "simples": {},
        }
        resp = builder.build(data)
        self.assertEqual(
            resp,
            {
                "body": {
                    "hobbies": [{"level": "noob", "name": "guitar"}],
                    "name": "bob",
                    "simples": {},
                },
                "header": {},
                "status_code": 200,
            },
        )
        self.assertEqual(
            builder.errors,
            [
                ("simples.int", "Required"),
                ("simples.float", "Required"),
                ("simples.decimal", "Required"),
                ("simples.string", "Required"),
                ("simples.bool", "Required"),
                ("simples.date", "Required"),
                ("simples.datetime", "Required"),
                ("simples.enum_str", "Required"),
                ("hobbies.1.name", "Required"),
            ],
        )

    def test_deserialize_mappingschema_override(self):
        builder = ResponseBuilder(DeseserializedOverride())
        resp = builder.build({"foo": "bar"})
        self.assertEqual(
            resp, {"body": {"happy": True}, "header": {}, "status_code": 200}
        )

    def test_deserialize_sequence_override(self):
        builder = ResponseBuilder(DeseserializedSequenceOverride())
        resp = builder.build({"body": [{"foo": "bar"}], "X-Total-Count": 1})
        self.assertEqual(
            resp,
            {
                "body": [{"happy": True}],
                "header": {"X-Total-Count": "1"},
                "status_code": 200,
            },
        )

    def test_colander_deserialize_compat(self):
        builder = ResponseBuilder(ColanderMissingCompat())
        resp = builder.build({})
        attempt = ColanderMissingCompat().deserialize({})
        self.assertEqual(resp["body"], attempt)

    def test_serialize_response(self):
        uuid_ = str(uuid.uuid1())
        location = "http://localhost/testing/%s" % uuid_
        data = {
            "message": "Test Created",
            "uuid": uuid_,
            "food": "choucroute",
            "location": location,
        }

        expected = {
            "message": "Test Created",
            "uuid": uuid_,
        }

        context = {}
        context["request"] = mock.Mock()
        context["request"].response.headers = {}
        context["request"].context.return_schema = UUIDHTTPCreatedReturn
        context["request"].accept.acceptable_offers = lambda x: [
            ("application/json", 1.0)
        ]

        res = JsonRenderer(None)(data, context)
        self.assertEqual(res, json.dumps(expected))
        self.assertEqual(context["request"].response.headers, {"Location": location})
        self.assertEqual(context["request"].response.status_code, 201)

    def test_json_encoder(self):
        data = OrderedDict()
        data["int"] = 1
        data["bool"] = True
        data["float"] = 1.0
        data["decimal"] = decimal.Decimal("1.33")
        data["date"] = datetime.date(2015, 11, 18)
        data["datetime"] = datetime.datetime(2015, 11, 18, 10, 12, 54)
        resp = json.dumps(data, cls=JSONEncoder)
        self.assertEqual(
            resp,
            '{"int": 1, '
            '"bool": true, '
            '"float": 1.0, '
            '"decimal": 1.33, '
            '"date": "2015-11-18T00:00:00Z", '
            '"datetime": "2015-11-18T10:12:54Z"}',
        )

    def test_titlelize_headers(self):
        data = {"foo": "bar", "foo-bar-corp": "foo", "Plouf-Plouf": "tiplouf"}

        context = {}
        context["request"] = mock.Mock()
        context["request"].response.headers = {}
        context["request"].context.return_schema = ColanderTitlelizeHeaders
        context["request"].accept.acceptable_offers = lambda x: [
            ("application/json", 1.0)
        ]

        JsonRenderer(None)(data, context)
        self.assertEqual(
            context["request"].response.headers,
            {"Foo": "bar", "Foo-Bar-Corp": "foo", "Plouf-Plouf": "tiplouf"},
        )


class DummyListReturn(ListReturn):
    @colander.instantiate(name="body")
    class DummySequenceReturn(colander.SequenceSchema):
        @colander.instantiate()
        class DummyReturn(colander.MappingSchema):
            name = colander.SchemaNode(colander.String())

            @colander.instantiate(name="addons")
            class AddOns(colander.SequenceSchema):
                @colander.instantiate()
                class AddOn(colander.MappingSchema):
                    name = colander.SchemaNode(colander.String())


class SQLAlchemySerializationTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        super(SQLAlchemySerializationTestCase, cls).setUpClass()
        cls._db = TestCommon()
        cls._db.setUp()

        with transaction.manager:
            dummy = Dummy(name="Xian Ping")
            mini = MiniDummy(name="Fou Mi", dummy=dummy)
            addon = AddOnDummy(name="shuriken", dummy=dummy)
            cls._db.session.add(dummy)
            cls._db.session.add(mini)
            cls._db.session.add(addon)
            cls._db.session.flush()

    @classmethod
    def tearDownClass(cls):
        cls._db.tearDown()
        super(SQLAlchemySerializationTestCase, cls).tearDownClass()

    def test_serialization_leaks(self):
        dummies = Dummy.find()
        builder = ResponseBuilder(DummyListReturn())
        data = builder.build({"body": dummies, "X-Total-Count": 1})

        self.assertEqual(builder.errors, [])
        self.assertEqual(
            data,
            {
                "status_code": 200,
                "header": {"X-Total-Count": "1"},
                "body": [{"addons": [{"name": "shuriken"}], "name": "Xian Ping"}],
            },
        )
