import json
from typing import Optional

import colander
import pytest
from pydantic import BaseModel, Field, ValidationError
from pyramid.interfaces import IRendererFactory

from sequoia.base3.helpers.pydantic.renderer import (
    JsonRenderer,
    includeme,
    render_response,
)
from sequoia.base3.helpers.pydantic.response_schema import (
    CollectionResponseSchema,
    NoContentResponseSchema,
    ResponseSchema,
)


def test_includeme(dummy_configurator):
    includeme(dummy_configurator)
    renderer = dummy_configurator.registry.getUtility(IRendererFactory, "pydantic")
    assert renderer is JsonRenderer


class Dummy(BaseModel):
    name: str = Field()
    size: Optional[int] = Field(default=None)


DummyCollectionResponseSchema = CollectionResponseSchema[Dummy]


@pytest.mark.parametrize(
    "params",
    [
        pytest.param(
            {
                "data": {"body": {"name": "foo", "ignored": "value"}},
                "schema": ResponseSchema[Dummy],
                "expected_status": 200,
                "expected_headers": {},
                "expected_body": {"name": "foo"},
            },
            id="A body response in a dict",
        ),
        pytest.param(
            {
                "data": ResponseSchema[Dummy](body=Dummy(name="foo")),
                "schema": ResponseSchema[Dummy],
                "expected_status": 200,
                "expected_headers": {},
                "expected_body": {"name": "foo"},
            },
            id="A wrapped response",
        ),
        pytest.param(
            {
                "data": Dummy(name="foo"),
                "schema": ResponseSchema[Dummy],
                "expected_status": 200,
                "expected_headers": {},
                "expected_body": {"name": "foo"},
            },
            id="A bared response",
        ),
        pytest.param(
            {
                "data": {},
                "schema": NoContentResponseSchema,
                "expected_status": 204,
                "expected_headers": {},
                "expected_body": "",
            },
            id="An empty response in HTTP",
        ),
        pytest.param(
            {
                "data": {
                    "headers": {
                        "X-Total-Count": 42,
                    },
                    "body": [{"name": "foo", "size": 3}, {"name": "bar"}],
                },
                "schema": CollectionResponseSchema[Dummy],
                "expected_status": 200,
                "expected_headers": {"X-Total-Count": "42"},
                "expected_body": [{"name": "foo", "size": 3}, {"name": "bar"}],
            },
            id="A collection",
        ),
        pytest.param(
            {
                "data": DummyCollectionResponseSchema(
                    headers={"X-Total-Count": 1}, body=[Dummy(name="foo")]
                ),
                "schema": CollectionResponseSchema[Dummy],
                "expected_status": 200,
                "expected_headers": {"X-Total-Count": "1"},
                "expected_body": [{"name": "foo"}],
            },
            id="A serialized collection",
        ),
    ],
)
def test_render_response(params, dummy_response):
    body = render_response(params["schema"], params["data"], dummy_response)
    if not isinstance(params["expected_body"], str):
        body = json.loads(body)
    assert body == params["expected_body"]
    headers = dict(dummy_response.headers)
    del headers["Content-Length"]
    headers.pop("Content-Type", None)
    assert headers == params["expected_headers"]
    assert dummy_response.status_code == params["expected_status"]


@pytest.mark.parametrize(
    "params",
    [
        {
            "exception": RuntimeError,
            "expected_message": "Missing return_schema predicate",
        },
        {
            "context": {"return_schema": colander.MappingSchema},
            "exception": TypeError,
            "expected_message": "colander.Schema: not a ResponseSchema",
        },
    ],
)
def test_bad_view_configuration(params, render_with_pydantic, dummy_context):
    with pytest.raises(params["exception"]) as ctx:
        render_with_pydantic({}, dummy_context)
    assert str(ctx.value) == params["expected_message"]


@pytest.mark.parametrize(
    "params",
    [
        pytest.param(
            {
                "data": ResponseSchema[Dummy](body=Dummy(name="foo")),
                "context": {"return_schema": ResponseSchema[Dummy]},
                "expected_status": 200,
                "expected_headers": {"Content-Type": "application/json"},
                "expected_body": {"name": "foo"},
            },
            id="A wrapped response",
        ),
    ],
)
def test_render_with_pydantic(
    params, render_with_pydantic, dummy_context, dummy_request
):
    response = render_with_pydantic(params["data"], dummy_context)

    assert json.loads(response) == params["expected_body"]
    headers = dict(dummy_request.response.headers)
    del headers["Content-Length"]
    assert headers == params["expected_headers"]
    assert dummy_request.response.status_code == params["expected_status"]


@pytest.mark.parametrize(
    "params",
    [
        {"headers": {}, "expected_msg": "Missing header X-Total-Count"},
        {
            "headers": {"X-Total-Count": "deux"},
            "expected_msg": "X-Total-Count is not a number",
        },
    ],
)
def test_header_x_total_count(params):
    with pytest.raises(ValidationError) as ctx:
        DummyCollectionResponseSchema(headers=params["headers"])
    assert ctx.value.errors() == [
        {"loc": ("headers",), "msg": params["expected_msg"], "type": "value_error"}
    ]
