"""
Some currently useless sqlalchemy tests
"""
import uuid
from unittest import TestCase

import transaction

from sequoia.base3.helpers.sqla import ModelError

from .sqlalchemy import AddOnDummy, Dummy, MiniDummy, TestCommon


class ModelErrorTestCase(TestCase):
    def test_string_parameter(self):
        err = ModelError("Something Wrong Happen")
        self.assertEqual(err.errors, ["Something Wrong Happen"])
        self.assertEqual(str(err), "Something Wrong Happen")

    def test_list_parameter(self):
        err = ModelError(["Error 1", "Error 2"])
        self.assertEqual(err.errors, ["Error 1", "Error 2"])
        self.assertEqual(str(err), "Error 1\nError 2")


class TestSQLAlchemyHelper(TestCommon, TestCase):
    def test_base(self):
        dummy_id = None
        with transaction.manager:
            dummy = Dummy(name="Toto")
            self.session.add(dummy)
            self.session.flush()

            dummy_id = dummy.id
            self.assertTrue(dummy_id)

        with transaction.manager:
            lst = Dummy.find()
            self.assertEqual(len(lst), 1)

            found = Dummy.by_id(dummy_id)

            self.assertEqual(found.minis, [])

    def test_build_where_clause(self):
        with transaction.manager:
            self.session.add(Dummy(name="Ping"))
            self.session.add(Dummy(name="Pong"))
            self.session.add(Dummy(name="Plouf"))

        # Test that an exception is raised when where clause is not iterable
        with self.assertRaises(NotImplementedError) as ctx:
            Dummy.find(where=(Dummy.name == "Ping"))
        self.assertEqual(
            str(ctx.exception),
            "Operator 'getitem' is not supported on this " "expression",
        )

        # Test the normal case
        dummies = Dummy.find(where=(Dummy.name == "Pong",))
        dummies = [dummy.name for dummy in dummies]
        self.assertEqual(["Pong"], dummies)

    def test_exists(self):
        with transaction.manager:
            self.session.add(Dummy(name="exists"))
        self.assertTrue(Dummy.exists(where=[Dummy.name == "exists"]))
        self.assertFalse(Dummy.exists(where=[Dummy.name == "does not exists"]))

    def test_uuid_type(self):
        u1 = uuid.UUID("00000001-0000-4000-0000-000000000001")
        u2 = uuid.UUID("00000002-0000-4000-0000-000000000002")

        with transaction.manager:
            self.session.add(Dummy(name="uuid", uuid=u1, uuid_ng=u2))

        with transaction.manager:
            dummy = self.session.query(Dummy).filter(Dummy.uuid == u1).one()

            self.assertIsInstance(dummy.uuid, str)
            self.assertIsInstance(dummy.uuid_ng, uuid.UUID)
            self.assertEqual(dummy.uuid, str(u1))
            self.assertEqual(dummy.uuid_ng, u2)

    def test_build_query(self):
        with transaction.manager:
            dum = Dummy(name="TestBuildQuery")
            self.session.add(dum)
            self.session.add(AddOnDummy(dummy=dum, name="arm0"))
            self.session.add(AddOnDummy(dummy=dum, name="arm1"))
            dum = Dummy(name="TestBuildQuery2")
            self.session.add(dum)
            self.session.add(AddOnDummy(dummy=dum, name="leg0"))
            self.session.add(AddOnDummy(dummy=dum, name="leg1"))
            dum = Dummy(name="TestBuildQuery3")
            self.session.add(dum)
            self.session.add(AddOnDummy(dummy=dum, name="body0"))
            self.session.add(AddOnDummy(dummy=dum, name="body1"))
            self.session.add(AddOnDummy(name="body2"))

        dummies = Dummy.build_query(
            select=(Dummy, AddOnDummy),
            join=[AddOnDummy],
            order_by=[Dummy.name, AddOnDummy.name],
        ).all()

        dummies = [(dum.name, addon.name) for dum, addon in dummies]
        self.assertEqual(
            dummies,
            [
                ("TestBuildQuery", "arm0"),
                ("TestBuildQuery", "arm1"),
                ("TestBuildQuery2", "leg0"),
                ("TestBuildQuery2", "leg1"),
                ("TestBuildQuery3", "body0"),
                ("TestBuildQuery3", "body1"),
            ],
        )

        dummies = Dummy.build_query(
            select=(Dummy.name, AddOnDummy.name),
            join=[AddOnDummy],
            order_by=[Dummy.name, AddOnDummy.name],
        ).all()

        self.assertEqual(
            dummies,
            [
                ("TestBuildQuery", "arm0"),
                ("TestBuildQuery", "arm1"),
                ("TestBuildQuery2", "leg0"),
                ("TestBuildQuery2", "leg1"),
                ("TestBuildQuery3", "body0"),
                ("TestBuildQuery3", "body1"),
            ],
        )

        dummies = AddOnDummy.build_query(
            select=(Dummy.name, AddOnDummy.name),
            outerjoin=[Dummy],
        ).all()

        self.assertEqual(
            dummies,
            [
                ("TestBuildQuery", "arm0"),
                ("TestBuildQuery", "arm1"),
                ("TestBuildQuery2", "leg0"),
                ("TestBuildQuery2", "leg1"),
                ("TestBuildQuery3", "body0"),
                ("TestBuildQuery3", "body1"),
                (None, "body2"),
            ],
        )

    def test_count_join(self):
        with transaction.manager:
            d1 = Dummy(name="First")
            d2 = Dummy(name="Second")

            d1m1 = MiniDummy(name="1", dummy=d1)
            d1m1 = MiniDummy(name="2", dummy=d1)
            d2m1 = MiniDummy(name="1", dummy=d2)
            d2m2 = MiniDummy(name="2", dummy=d2)

            self.session.add_all((d1, d2, d1m1, d1m1, d2m1, d2m2))
            self.session.flush()

            self.assertEqual(Dummy.find(count=True), 2)
            self.assertEqual(MiniDummy.find(count=True), 4)

            # relationships are correct
            dummies = Dummy.find(where=(Dummy.name == "First",))
            self.assertEqual(len(dummies), 1)
            self.assertEqual(len(dummies[0].minis), 2)

            # count=True will actually count the number of rows
            # (**including joins**)
            count = Dummy.find(
                where=(Dummy.name == "First",), join=(MiniDummy,), count=True
            )
            self.assertEqual(count, 2)

            # Using a column + distinct will count the number of objects
            count = Dummy.find(
                where=(Dummy.name == "First",),
                join=(MiniDummy,),
                count=Dummy.id.distinct(),
            )
            self.assertEqual(count, 1)
